#include <assert.h>

#include <opcerr.h>
#include <opcdebug.h>

#include <opcbufferpool.h>

static inline OPCBufferContext opcBufferPoolCreateBuffer(OPCBufferPoolContext ctx, unsigned long buffer_size) {
    OPCBufferContext buffer;
    
    assert(ctx);
    assert(buffer_size);
    
    buffer = (OPCBufferContext)opcMalloc(ctx->malloc, sizeof(struct OPCBufferContextT));
    if (!buffer) return buffer;
    
    buffer->buffer = (char*)opcMalloc(ctx->malloc, buffer_size);
    if (!buffer->buffer) {
	opcFree(ctx->malloc, buffer);
	return NULL;
    }
    buffer->buffer_size = buffer_size;

    
    if (opcBufferInitContext(buffer)) {
	opcFree(ctx->malloc, buffer);
	opcFree(ctx->malloc, buffer->buffer);
	return NULL;
    }
    
    buffer->pool = ctx;
    
    return buffer;
}

static inline void opcBufferPoolDestroyBuffer(OPCBufferPoolContext ctx, OPCBufferContext buffer) {

    assert(ctx);
    assert(buffer);
    
    opcBufferFreeContext(buffer);
    if (buffer->buffer) opcFree(ctx->malloc, buffer->buffer);
    opcFree(ctx->malloc, buffer);
}


static void opcBufferPoolZeroContext(OPCBufferPoolContext ctx) {
    assert(ctx);

    ctx->first = NULL;
    ctx->last = NULL;
    ctx->lock = NULL;
    ctx->allocated = NULL;
}


OPCBufferPoolContext opcBufferPoolCreateContext(OPCMallocContext mctx) {
    OPCBufferPoolContext ctx;

    opcDebugServer(OPC_DEBUG_NOTICE, "opcBufferPoolCreateContext");    

    ctx = (OPCBufferPoolContext)opcMalloc(mctx,sizeof(struct OPCBufferPoolContextT));
    if (ctx) {
	ctx->malloc=mctx;
	opcBufferPoolZeroContext(ctx);
    }
    return ctx;
}

int opcBufferPoolInitContext(OPCBufferPoolContext ctx, unsigned long first_bin, unsigned char bins) {
    int err;
    unsigned char i;
    OPCBufferContext *ptr1, *ptr2;
    unsigned long *ptr3;

    opcDebugServer(OPC_DEBUG_NOTICE, "opcBufferPoolInitContext");    

    assert(ctx);
    assert(bins);
    assert(first_bin>8);

    ptr1 = (OPCBufferContext*)opcMalloc(ctx->malloc, bins*sizeof(OPCBufferContext));
    ptr2 = (OPCBufferContext*)opcMalloc(ctx->malloc, bins*sizeof(OPCBufferContext));
    ptr3 = (unsigned long*)opcMalloc(ctx->malloc, (bins+1)*sizeof(unsigned long));
    
    if ((!ptr1)||(!ptr2)||(!ptr3)) {
	if (ptr1) opcFree(ctx->malloc, ptr1);
	if (ptr2) opcFree(ctx->malloc, ptr2);
	if (ptr3) opcFree(ctx->malloc, ptr3);
	return ERR_OPC_MEM;
    }
    
    for (i=0;i<bins;i++) {
	ptr1[i]=NULL;
	ptr2[i]=NULL;
	ptr3[i]=0;
    }
    ptr3[bins]=0;

    ctx->lock = LockCreateContext(ctx->malloc);
    if (!ctx->lock) {
	opcBufferPoolFreeContext(ctx);
	return ERR_OPC_MEM;
    }
    err = LockInitContext(ctx->lock);
    if (err) {
	opcBufferPoolFreeContext(ctx);
	return err;
    }

    ctx->bins = bins;

    ctx->first = ptr1;
    ctx->last = ptr2;
    ctx->allocated = ptr3;
    
    for (i=0;first_bin;first_bin>>=1) i++;
    ctx->lsr = i;
    
    return 0;
}

void opcBufferPoolFreeContext(OPCBufferPoolContext ctx) {
    unsigned char i;
    OPCBufferContext ptr1, ptr2;

    opcDebugServer(OPC_DEBUG_NOTICE, "opcBufferPoolFreeContext");    

    assert(ctx);
    
    if (ctx->lock) LockDestroyContext(ctx->lock);
    
    // check allocated zeroes
    if (ctx->first) {
	for (i=0;i<ctx->bins;i++) {
	    for (ptr1=ctx->first[i];ptr1;ptr1=ptr2) {
		ptr2 = ptr1->next;
		opcBufferPoolDestroyBuffer(ctx, ptr1);
	    }
	}
	opcFree(ctx->malloc, ctx->first);
    }

    if (ctx->last) opcFree(ctx->malloc, ctx->last);

    if (ctx->allocated) {
	for (i=0;i<=ctx->bins;i++) {
	    if (ctx->allocated[i]) {
		if (i==ctx->bins) 
		    opcDebugServer(OPC_DEBUG_ERROR, "opcBufferPoolFreeContext: We lost somethere %lu huge buffers", ctx->allocated[i]);
		else
		    opcDebugServer(OPC_DEBUG_ERROR, "opcBufferPoolFreeContext: We lost somethere %lu buffers of size %lu", ctx->allocated[i], 1<<(ctx->lsr+i-1));
	    }
	}
	opcFree(ctx->malloc, ctx->allocated);
    }

    opcBufferPoolZeroContext(ctx);    
}

void opcBufferPoolDestroyContext(OPCBufferPoolContext ctx) {
    opcDebugServer(OPC_DEBUG_NOTICE, "opcBufferPoolDestroyContext");    
    opcBufferPoolFreeContext(ctx);
    opcFree(ctx->malloc,ctx);
}

OPCBufferContext opcBufferPoolGetBuffer(OPCBufferPoolContext ctx, unsigned long size) {
    int err;
    OPCBufferContext buffer;
    unsigned long int i;
    unsigned char j;
    
    assert(ctx);

    for (j=0,i=size>>ctx->lsr;i;i>>=1) j++;
    i = (1<<(j+ctx->lsr-1));
    if (i<size) {
	size=i<<1;
	j++;
    } else size=i;

    
    if (j>=ctx->bins) {
	opcDebugServer(OPC_DEBUG_NOTICE, "opcBufferPoolGetBuffer: Allocating new BigBuffer: %lu", size);

	buffer = opcBufferPoolCreateBuffer(ctx,size);
	if (!buffer) return buffer;

	LockWrite(ctx->lock);
	ctx->allocated[ctx->bins]++;
	LockReturn(ctx->lock);
	return buffer;
    }

    
    LockWrite(ctx->lock);
    buffer = ctx->first[j];
    if (buffer) {
	opcDebugServer(OPC_DEBUG_DEBUG, "opcBufferPoolGetBuffer: GetBuffer: %lu", size);
	if (buffer->next) {
	    ctx->first[j]=buffer->next;
	} else {
	    ctx->first[j]=NULL;
	    ctx->last[j]=NULL;
	}
    } else {
	opcDebugServer(OPC_DEBUG_NOTICE, "opcBufferPoolGetBuffer: Allocating new buffer: %lu", size);
	buffer = opcBufferPoolCreateBuffer(ctx, size);
	i=1;
    }
    if (buffer) ctx->allocated[j]++;
    LockReturn(ctx->lock);
    
    if (!i) opcBufferRestoreContext(buffer);

    return buffer;
}

void opcBufferPoolReturnBuffer(OPCBufferPoolContext ctx, OPCBufferContext buffer) {
    unsigned long size;
    unsigned long int i;
    unsigned char j;

    assert(ctx);
    assert(buffer);
    
    size = buffer->buffer_size;
    
    for (j=0, i=(size>>ctx->lsr);i;i>>=1) j++;
    

    if (j>=ctx->bins) {
	assert(ctx->allocated[ctx->bins]>0);

	opcDebugServer(OPC_DEBUG_NOTICE, "opcBufferPoolReturnBuffer: Destroying BigBuffer: %lu", size);

	opcBufferPoolDestroyBuffer(ctx, buffer);

	LockWrite(ctx->lock);
	ctx->allocated[ctx->bins]--;
	LockReturn(ctx->lock);
	return;
    }

    opcDebugServer(OPC_DEBUG_DEBUG, "opcBufferPoolReturnBuffer: ReturnBuffer: %lu", size);


    LockWrite(ctx->lock);
    assert(ctx->allocated[j]>0);

    if (ctx->last[j]) {
	ctx->last[j]->next = buffer;
	ctx->last[j] = buffer;
    } else {
	ctx->first[j] = buffer;
	ctx->last[j] = buffer;
    }    
    buffer->next = NULL;

    ctx->allocated[j]--;
    LockReturn(ctx->lock);

}


int opcBufferPoolResizeBuffer(OPCBufferPoolContext ctx, OPCBufferContext buffer, unsigned long size) {
    unsigned char *buff;
    unsigned long old_size;
    unsigned long int i;
    unsigned char j, old_j;

    assert(ctx);
    assert(buffer);

    old_size = buffer->buffer_size;
    for (old_j=0, i=(old_size>>ctx->lsr);i;i>>=1) old_j++;


    for (j=0,i=size>>ctx->lsr;i;i>>=1) j++;
    i = (1<<(j+ctx->lsr-1));
    if (i<size) {
	size=i<<1;
	j++;
    } else size=i;

    opcDebugServer(OPC_DEBUG_DEBUG, "opcBufferPoolGetBuffer: Resizing Buffer: %p. Old Size: %lu, New Size: %lu", buffer, buffer->buffer_size, size);

    buff = (char*)opcRealloc(ctx->malloc, buffer->buffer, size);
    if (!buff) return ERR_OPC_MEM;
    buffer->buffer = buff;
    buffer->buffer_size = size;

    if (j>ctx->bins) {
	j = ctx->bins;
	if (old_j>ctx->bins) old_j = ctx->bins;
    }

    LockWrite(ctx->lock);
    assert(ctx->allocated[old_j]>0);
    ctx->allocated[old_j]--;
    ctx->allocated[j]++;
    LockReturn(ctx->lock);

    return 0;
}

int opcBufferPoolReplaceBuffer(OPCBufferPoolContext ctx, OPCBufferContext buffer, unsigned long size) {
    unsigned char *buff;
    unsigned long old_size;
    unsigned long int i;
    unsigned char j, old_j;

    assert(ctx);
    assert(buffer);

    old_size = buffer->buffer_size;
    for (old_j=0, i=(old_size>>ctx->lsr);i;i>>=1) old_j++;

    for (j=0,i=size>>ctx->lsr;i;i>>=1) j++;
    i = (1<<(j+ctx->lsr-1));

    if (i<size) {
	size=i<<1;
	j++;
    } else size=i;

    buff = (char*)opcMalloc(ctx->malloc, size);
    if (!buff) return ERR_OPC_MEM;
    opcFree(ctx->malloc, buffer->buffer);

    opcDebugServer(OPC_DEBUG_DEBUG, "opcBufferPoolGetBuffer: Replacing Buffer: %p. Old Size: %lu, New Size: %lu", buffer, buffer->buffer_size, size);
    
    buffer->buffer = buff;
    buffer->buffer_size = size;

    if (j>ctx->bins) {
	j = ctx->bins;
	if (old_j>ctx->bins) old_j = ctx->bins;
    }
    LockWrite(ctx->lock);
    assert(ctx->allocated[old_j]>0);
    ctx->allocated[old_j]--;
    ctx->allocated[j]++;
    LockReturn(ctx->lock);
    return 0;
}


void opcBufferPoolMantain(OPCBufferPoolContext ctx) {
    // ToDo: Free unnececary buffers
}
