#define _OPC_DATA_BUFFER_C

#include <opcassert.h>
#include <opcerr.h>

#include <opcbinaryconfiguration.h>
#include <opcdatabuffer.h>

#define HEADER_SIZE (sizeof(void*)+sizeof(unsigned long))
#define ALIGN(pos,alignment) \
    if (pos&(alignment-1)) pos = (pos&~(alignment-1))+alignment;

#define OPC_DATA_BUFFER_APPROXIMATION_INTERVAL 20
#define OPC_DATA_BUFFER_APPROXIMATION_INTERVAL_05 10
#define OPC_DATA_BUFFER_APPROXIMATION_INTERVAL_15 30
#define OPC_DATA_BUFFER_APPROXIMATION_RISE 2

OPCAllocator opcAllocatorDataBuffer() {
    return &opc_allocator_databuffer;
}

OPCDataBufferOptions opcDataBufferGetDefaultOptions() {
    int err;
    OPCDataBufferOptions options;

    options = (OPCDataBufferOptions)malloc(sizeof(struct OPCDataBufferOptionsT));
    if (!options) return options;
    
    err = opcDataBufferInitDefaultOptions(options);
    if (err) {
	opcDataBufferDestroyOptions(options);
	return NULL;
    }
    
    return options;
}

int opcDataBufferInitDefaultOptions(OPCDataBufferOptions options) {
    assert(options);
        
    if (!options) return ERR_OPC_PARAM;
    
    memcpy(options, &opc_databuffer_options, sizeof(struct OPCDataBufferOptionsT));
    return 0;
}

void opcDataBufferDestroyOptions(OPCDataBufferOptions options) {
    if (options) free(options);
}

int opcDataBufferOptionsSetSize(OPCDataBufferOptions options, unsigned long minimal_size, unsigned long maximal_size) {
    assert(options);
    
    options->minimal_size = minimal_size;
    options->maximal_size = maximal_size;
    return 0;
}

int opcDataBufferOptionsSetBlockSize(OPCDataBufferOptions options, unsigned long block_size) {
    assert(options);

    options->block_size = block_size;
        
    return 0;
}

int opcDataBufferOptionsSetAllocationSize(OPCDataBufferOptions options, unsigned long fixed_allocation_size) {
    assert(options);
    
    options->fixed_allocation_size = fixed_allocation_size;
    
    return 0;
}

int opcDataBufferOptionsSetBufferPool(OPCDataBufferOptions options, OPCBufferPoolContext pool) {
    assert(options);
    
    options->pool = pool;
    
    return 0;
}


static inline void opcDataBufferZeroContext(OPCDataBufferContext ctx) {
    assert(ctx);
    ctx->buffers = NULL;
    ctx->current = NULL;
    ctx->pool = NULL;
    
    ctx->n = 0;
    ctx->max1 = 0;
    ctx->max2 = 0;
    ctx->save = 0;
    ctx->maxfill = 0;
    ctx->allocation_size = 0;
    ctx->preallocated = 0;
}

OPCDataBufferContext opcDataBufferCreateContext(OPCMallocContext mctx) {
    OPCDataBufferContext ctx;

    ctx = (OPCDataBufferContext)opcMalloc(mctx,sizeof(struct OPCDataBufferContextT));
    if (ctx) {
	ctx->base.malloc=mctx;
	/*ctx->base.allocator=&opc_allocator_databuffer;*/
	memcpy(&ctx->base.allocator, &opc_allocator_databuffer, sizeof(struct OPCAllocatorT));
	opcDataBufferZeroContext(ctx);
    }
    return ctx;
}

int opcDataBufferInitContext(OPCDataBufferContext ctx, OPCDataBufferOptions options) {
    unsigned long i, size;
    OPCBufferContext buffer;
    
    assert(ctx);

    if (!options) return ERR_OPC_DATABUFFER_OPTIONS;
    if (!options->pool) return ERR_OPC_DATABUFFER_OPTIONS_BUFFERPOOL;

    opcDataBufferZeroContext(ctx);

    ctx->pool = options->pool;

    size = options->minimal_size;
    for (i=1;i<size;i<<=1);
    ctx->minimal_size = i;

    size = options->maximal_size;
    if (size) {
	for (;i<size;i<<=1);
	ctx->maximal_size = i;
    } else ctx->maximal_size = 0;
    
    if (options->block_size) {
        size = options->block_size;
	for (i=1;i<size;i<<=1);
	ctx->block_size=i;
    } else ctx->block_size=1>>(options->pool->lsr-2); /* lsr>3 */
    
    if (options->fixed_allocation_size) { 
/*	if (options->fixed_allocation_size<ctx->block_size) ctx->fixed_allocation_size=ctx->block_size;
	else {
    	    size = options->fixed_allocation_size;
	    for (i=1;i<size;i<<=1);
	    ctx->fixed_allocation_size=i;
	}*/
	ctx->fixed_allocation_size = options->fixed_allocation_size;
    }
    else ctx->fixed_allocation_size = 0;
    
    /* ctx->fixed_allocation_size = 0; For benchmarking DS */

    buffer = opcBufferPoolGetBuffer(options->pool, ctx->minimal_size /* 0 is OK */);
    if (!buffer) return ERR_OPC_BUFFERPOOL_GET_BUFFER;

    ctx->current = buffer;
    ctx->buffers = buffer;
    
    return 0;
}

void opcDataBufferFreeContext(OPCDataBufferContext ctx) {
    OPCBufferContext buffer, nbuffer;

    assert(ctx);

    for (buffer=ctx->buffers;buffer;buffer=nbuffer) {
	nbuffer = buffer->next;
	opcBufferPoolReturnBuffer(ctx->pool, buffer);
    }
    opcDataBufferZeroContext(ctx);    
}

void opcDataBufferDestroyContext(OPCDataBufferContext ctx) {
    opcDataBufferFreeContext(ctx);
    opcFree(ctx->base.malloc,ctx);
}

static inline OPCBufferContext opcDataBufferClear(OPCDataBufferContext ctx, OPCBufferContext buffer) {
    OPCBufferContext nbuffer;
    unsigned char destroy = 0;
    
    if ((ctx->buffers == buffer)&&(buffer->buffer_start == buffer->buffer_end)) destroy = 1;

    for (nbuffer=buffer->next;nbuffer;nbuffer=buffer->next) {
	if (destroy) opcBufferPoolReturnBuffer(ctx->pool, buffer);
	buffer = nbuffer;
    }
    if (destroy) ctx->buffers = buffer;
    return buffer;
}

static inline int opcDataBufferStart(OPCDataBufferContext ctx) {
    assert(ctx);
    
    if (ctx->fixed_allocation_size) {
	if (ctx->preallocated > ctx->fixed_allocation_size) return 0;
	ctx->preallocated = opcBufferReserve(ctx->current, ctx->maximal_size, ctx->fixed_allocation_size);
    } else {    
	if (ctx->n<OPC_DATA_BUFFER_APPROXIMATION_INTERVAL) {
    	    if (ctx->preallocated > ctx->max1) return 0;
	    if (ctx->max1>ctx->block_size) ctx->preallocated = opcBufferReserve(ctx->current, ctx->maximal_size, ctx->max1);
	    else ctx->preallocated = opcBufferReserve(ctx->current, ctx->maximal_size, ctx->block_size);
	} else {
	    if (ctx->preallocated > ctx->max2) return 0;
	    if (ctx->max2>ctx->block_size) ctx->preallocated = opcBufferReserve(ctx->current, ctx->maximal_size, ctx->max2);
	    else ctx->preallocated = opcBufferReserve(ctx->current, ctx->maximal_size, ctx->block_size);
	}
    }

    if (ctx->current->next) ctx->current=opcDataBufferClear(ctx, ctx->current);
//    for (buffer=ctx->current;buffer->next;buffer=buffer->next);
//    ctx->current=buffer;
    
    return 0;
}


unsigned long opcDataBufferReserve(OPCDataBufferContext ctx, OPCAllocation notrequired, unsigned long size) {
//    OPCBufferContext buffer;
    
    if (ctx->preallocated<size) {
	if (size>ctx->block_size) ctx->preallocated = opcBufferReserve(ctx->current, ctx->maximal_size, size);
	else ctx->preallocated = opcBufferReserve(ctx->current, ctx->maximal_size, ctx->block_size);

	if (ctx->current->next) ctx->current=opcDataBufferClear(ctx, ctx->current);
//	for (buffer=ctx->current;buffer->next;buffer=buffer->next);
//	ctx->current=buffer;
    } 
    return ctx->preallocated;
}

void opcDataBufferTrim(OPCDataBufferContext ctx, OPCAllocation notrequired, unsigned long size) {
    opcBufferTrim(ctx->current, size);
}


void *opcDataBufferAllocate(OPCDataBufferContext ctx, OPCAllocation notrequired, unsigned long size, unsigned char alignment) {
    OPCBufferContext buffer;
    unsigned long preallocate;

    assert(ctx);
    buffer = ctx->current;
    assert(buffer);

    preallocate = size + opc_binary_configuration->align_allocation + HEADER_SIZE;
    if (preallocate>ctx->preallocated) {
	if (preallocate>ctx->block_size) ctx->preallocated = opcBufferReserve(buffer, ctx->maximal_size, preallocate);
	else ctx->preallocated = opcBufferReserve(buffer, ctx->maximal_size, ctx->block_size);
	if (!ctx->preallocated) return NULL;
	
	if (buffer->next) {
	    buffer=opcDataBufferClear(ctx, buffer);
	    ctx->current=buffer;
	}
    }

    return opcBufferFastAllocate(buffer, size, alignment, &ctx->preallocated);
}

void opcDataBufferFree(OPCDataBufferContext ctx, OPCAllocation buffer) {
    OPCBufferContext buff;
    
    assert(ctx);
    assert(buffer);
    
    ctx->buffers = opcBufferFree(ctx->buffers, (void*)buffer);
}

void *opcDataBufferAcceptChunk(OPCDataBufferContext ctx, OPCAllocation notrequired, unsigned long *req_size) {
    void *buffer;
    unsigned long size;

    assert(ctx);
    
    if (ctx->fixed_allocation_size) {
	buffer=opcBufferChunkAccept(ctx->current, req_size);
    } else {
	if (!req_size) req_size = &size;
	buffer = (void*)opcBufferChunkAccept(ctx->current, req_size);
	if (buffer) ctx->allocation_size += *req_size;
    }

    if (!ctx->first_chunk) {
	ctx->first_chunk = ctx->current;
	ctx->allocation = (OPCAllocation)buffer;
    }
    
    return buffer;
}

void opcDataBufferCancelChunk(OPCDataBufferContext ctx, OPCAllocation notrequired) {
    assert(ctx);
    opcBufferChunkCancel(ctx->current);
    /* Ignoring rise of preallocated space */
}

void *opcDataBufferSimpleChunk(OPCDataBufferContext ctx, OPCAllocation notrequired, unsigned long size) {
    OPCBufferContext buffer;
    unsigned long preallocate;

    assert(ctx);
    
    buffer = ctx->current;
    
    preallocate = size + opc_binary_configuration->align_allocation + HEADER_SIZE;
    if (preallocate>ctx->preallocated) {
	if (preallocate>ctx->block_size) ctx->preallocated = opcBufferReserve(buffer, ctx->maximal_size, preallocate);
	else ctx->preallocated = opcBufferReserve(buffer, ctx->maximal_size, ctx->block_size);
	if (!ctx->preallocated) return NULL;	
        if (buffer->next) {
	    buffer=opcDataBufferClear(ctx, ctx->current);
	    ctx->current = buffer;
	}
/*	for (;buffer->next;buffer=buffer->next);
	ctx->current=buffer;*/
    }
    
    if (!ctx->fixed_allocation_size) ctx->allocation_size += size;
    
    return opcBufferChunkSimple(buffer, size, &ctx->preallocated);
}

OPCAllocation opcDataBufferSimpleAllocate(OPCDataBufferContext ctx, unsigned long size) {
    OPCBufferContext buffer;
    void *buff;
    unsigned long used_space;
    unsigned long preallocate;
    
    assert(ctx);
    assert(!ctx->first_chunk);

    buffer = ctx->current;

    preallocate = size + opc_binary_configuration->align_allocation + HEADER_SIZE;
    if (preallocate>ctx->preallocated) {
	if (preallocate>ctx->block_size) ctx->preallocated = opcBufferReserve(buffer, ctx->maximal_size,  preallocate);
	else ctx->preallocated = opcBufferReserve(buffer, ctx->maximal_size, ctx->block_size);
	if (!ctx->preallocated) return NULL;	
        if (buffer->next) {
	    buffer=opcDataBufferClear(ctx, ctx->current);
	    ctx->current = buffer;
	}
/*
	for (;buffer->next;buffer=buffer->next);
	ctx->current=buffer;*/
    }

    if (ctx->buffers == buffer) {
	buff=opcBufferSimpleAllocate(buffer, size, &ctx->preallocated, &used_space);
	if (used_space > ctx->maxfill) ctx->maxfill = used_space;
    } else {
	buff=opcBufferSimpleAllocate(buffer, size, &ctx->preallocated, NULL);
	ctx->maxfill = buffer->buffer_size;
    }

/*
    In case of single call to this function instead of common way, this data can be
    considerate non-standard, and therefore shouldn't be used in approximation.
    
    if (!ctx->fixed_allocation_size) {
	if ((++(ctx->n))==OPC_DATA_BUFFER_APPROXIMATION_INTERVAL) ctx->max1 = 0;
	else if (ctx->n==OPC_DATA_BUFFER_APPROXIMATION_INTERVAL_15) {
	    ctx->max2 = 0;
	    ctx->n = OPC_DATA_BUFFER_APPROXIMATION_INTERVAL_05;
	}
        
	used_space = ctx->save;

	if ((size>OPC_DATA_BUFFER_APPROXIMATION_RISE*ctx->max1)&&(size>OPC_DATA_BUFFER_APPROXIMATION_RISE*ctx->max2)) {
	    if (used_space) {
		if (size>ctx->max1) ctx->max1 = size;
	        if (size>ctx->max2) ctx->max2 = size;
		
		if (used_space>ctx->max1) ctx->max1 = used_space;
	        if (used_space>ctx->max2) ctx->max2 = used_space;
		
		ctx->save = 0;
	    } else ctx->save = size;
	} else {
	    if (size>ctx->max1) ctx->max1=size;
	    if (size>ctx->max2) ctx->max2=size;

	    if (used_space) {
		if (used_space<OPC_DATA_BUFFER_APPROXIMATION_RISE*size) {
		    if (used_space>ctx->max1) ctx->max1 = used_space;
		    if (used_space>ctx->max2) ctx->max2 = used_space;
	        }
		ctx->save=0;
	    }
	}
    }
*/

/*
     Preparing for next round.
    ToDo: Find better place? (Manager?)
    
    Commented just for same case as above, - in case of good place to allocate memory
    is found, - enable.
    opcDataBufferStart(ctx);
*/

    return (OPCAllocation)buff;
}

OPCAllocation opcDataBufferAccept(OPCDataBufferContext ctx, OPCAllocation notrequired) {
    unsigned long save;
    unsigned long size;
    unsigned long used_space;


    if (ctx->buffers == ctx->current) {
	if (opcBufferAccept(ctx->current, &used_space)) return NULL;
	if (used_space > ctx->maxfill) ctx->maxfill = used_space;
    } else {
	if (opcBufferAccept(ctx->current, NULL)) return NULL;
	ctx->maxfill = ctx->current->buffer_size;
    }
    
    if (!ctx->fixed_allocation_size) {
	if ((++(ctx->n))==OPC_DATA_BUFFER_APPROXIMATION_INTERVAL) ctx->max1 = 0;
	else if (ctx->n==OPC_DATA_BUFFER_APPROXIMATION_INTERVAL_15) {
	    ctx->max2 = 0;
	    ctx->n = OPC_DATA_BUFFER_APPROXIMATION_INTERVAL_05;
	}
        
	size = ctx->allocation_size;
	used_space = ctx->save;

	if ((size>OPC_DATA_BUFFER_APPROXIMATION_RISE*ctx->max1)&&(size>OPC_DATA_BUFFER_APPROXIMATION_RISE*ctx->max2)) {
	    if (used_space) {
		if (size>ctx->max1) ctx->max1 = size;
	        if (size>ctx->max2) ctx->max2 = size;
		
		if (used_space>ctx->max1) ctx->max1 = used_space;
	        if (used_space>ctx->max2) ctx->max2 = used_space;
		
		ctx->save = 0;
	    } else ctx->save = size;
	} else {
	    if (size>ctx->max1) ctx->max1=size;
	    if (size>ctx->max2) ctx->max2=size;

	    if (used_space) {
		if (used_space<OPC_DATA_BUFFER_APPROXIMATION_RISE*size) {
		    if (used_space>ctx->max1) ctx->max1 = used_space;
		    if (used_space>ctx->max2) ctx->max2 = used_space;
	        }
		ctx->save=0;
	    }
	}
	ctx->allocation_size = 0;
    }

    ctx->first_chunk = NULL;
    
    /* Preparing for next round.
    ToDo: Find better place? (Manager?) */
    opcDataBufferStart(ctx);
    
    return ctx->allocation;
}

void opcDataBufferCancel(OPCDataBufferContext ctx, OPCAllocation notrequired) {
    OPCBufferContext new_first, first, chunk;
    
    assert(ctx);
    
    first=ctx->first_chunk;
    new_first=opcBufferCancel(first);
    if (new_first!=first) {
	if (ctx->buffers==first) ctx->buffers=new_first;
	else {
	    for (chunk=ctx->buffers;chunk->next!=first;chunk=chunk->next);
	    chunk->next = new_first;
	}
    }
    
    ctx->first_chunk = NULL;
    ctx->preallocated = 0;

    /* Preparing for next round.
    ToDo: Find better place? (Manager?) */
    opcDataBufferStart(ctx);
}


int opcDataBufferCall(OPCDataBufferContext ctx, OPCAllocation allocation, OPCSendFunction func, void *attr, unsigned char align) {
    int err;
    char *ptr, *next;
    unsigned long size;
    
    for (ptr=(void*)allocation;ptr;ptr=next) {
	size = *((unsigned long*)(ptr-HEADER_SIZE));
	next = *((char**)(ptr-sizeof(void*)));
	if ((next)&&(align)) ALIGN(size, opc_binary_configuration->align_allocation);
	err=func(attr,ptr,size);
	if (err) return err;
    }
    return 0;
}

int opcDataBufferSend(OPCUserData user_data, OPCSendFunction func, void *attr) {
    int err;
    char *ptr, *next;
    unsigned long size;
    OPCAllocation allocation = ((OPCUserData)user_data)->data;   


    for (ptr=(void*)allocation;ptr;ptr=next) {
	size = *((unsigned long*)(ptr-HEADER_SIZE));
	next = *((char**)(ptr-sizeof(void*)));
	if (next) ALIGN(size, opc_binary_configuration->align_allocation);
	err=func(attr,ptr,size);
	if (err) return err;
    }
    return 0;
}

unsigned long opcDataBufferSizeof(OPCDataBufferContext ctx, OPCAllocation allocation, unsigned char align) {
    char *ptr = (char*)allocation;
    unsigned long res=0,size;
    
    while (ptr) {
	size = *((unsigned long*)(ptr-HEADER_SIZE));
	ptr = *((char**)(ptr-sizeof(void*)));
	if ((ptr)&&(align)) ALIGN(size, opc_binary_configuration->align_allocation);
	res+=size;
    }
    return res;
}

void* opcDataBufferGet(OPCDataBufferContext ctx, OPCAllocation allocation, unsigned long *size) {
    assert(allocation);
    if (size) *size = *((unsigned long*)((void*)allocation-HEADER_SIZE));
    return (void*)allocation;
}

void* opcDataBufferGetNext(OPCDataBufferContext ctx, OPCAllocation allocation, void *current, unsigned long *size) {
    void *ptr;
    assert(current);
    
    ptr=*((char**)((char*)current-sizeof(void*)));
    if ((ptr)&&(size)) *size=*((unsigned long*)(ptr-HEADER_SIZE));
    return ptr;
}

void opcDataBufferMantain(OPCDataBufferContext ctx) {
    /* ToDo: resize buffers over maximal_size, and just for long time lowly filled
    maximal_size(soft), minimal_size */
    puts("opcDataBufferMantain");
}
