#include <stdio.h>
#include <assert.h>

#include <opcerr.h>
#include <opcmalloc.h>
#include <opcdebug.h>
#include <doublehash.h>

#include <opcholderbucket.h>
#include <opcholder.h>

static OPCHolderContext holder_ctx = NULL;

static void opcHolderZeroContext(OPCHolderContext ctx) {
    assert(ctx);
    ctx->lock=NULL;
    ctx->cond=NULL;
    ctx->double_hash=NULL;
    ctx->max_size=0;
    ctx->timeout=0;
}

OPCHolderContext opcHolderCreateContext(OPCMallocContext mctx) {
    OPCHolderContext ctx;

    opcDebugServer(OPC_DEBUG_NOTICE, "opcHolderCreateContext");    

    ctx = (OPCHolderContext)opcMalloc(mctx, sizeof(struct OPCHolderContextT));
    if (ctx) {
	ctx->malloc = mctx;
	opcHolderZeroContext(ctx);
    }
    return ctx;
}

// ToDo: Hard Quota, Soft Quota
int opcHolderInitContext(OPCHolderContext ctx, unsigned long size) {
    int err;
    unsigned i;

    opcDebugServer(OPC_DEBUG_NOTICE, "opcHolderInitContext");    

    assert(ctx);
    
    opcHolderZeroContext(ctx);
    
    ctx->double_hash = DoubleHashCreateContext(ctx->malloc);
    if (!ctx->double_hash) return ERR_OPC_MEM;
    if (DoubleHashInitContext(ctx->double_hash, size, size+1, 1, NULL)) {
	opcHolderFreeContext(ctx);
	return ERR_OPC_MEM;
    }
    
    ctx->lock = LockCreateContext(ctx->malloc);
    if (!ctx->lock) {
	opcHolderFreeContext(ctx);
	return ERR_OPC_MEM;
    }
    
    err = LockInitContext(ctx->lock);
    if (err) {
	opcHolderFreeContext(ctx);
	return err;
    }
    
    ctx->cond = ConditionCreateContext(ctx->malloc);
    if (!ctx->cond) {
	opcHolderFreeContext(ctx);
	return ERR_OPC_MEM;
    }
    
    err = ConditionInitContext(ctx->cond);
    if (err) {
	opcHolderFreeContext(ctx);
	return err;
    }
    
    return 0;
}

void opcHolderSetMaxSize(OPCHolderContext ctx, unsigned long max_size) {
    assert(ctx);
    
    ctx->max_size = max_size;
    DoubleHashSetMaximumSize(ctx->double_hash, max_size);
}

void opcHolderSetTimeout(OPCHolderContext ctx, time_t timeout) {
    assert(ctx);
    
    ctx->timeout = timeout;
}

void opcHolderFreeContext(OPCHolderContext ctx) {
    opcDebugServer(OPC_DEBUG_NOTICE, "opcHolderFreeContext");    

    assert(ctx);

    if (ctx->cond) ConditionDestroyContext(ctx->cond);
    if (ctx->lock) LockDestroyContext(ctx->lock);
    if (ctx->double_hash) DoubleHashDestroyContext(ctx->double_hash);
    opcHolderZeroContext(ctx);
}

void opcHolderDestroyContext(OPCHolderContext ctx) {
    opcDebugServer(OPC_DEBUG_NOTICE, "opcHolderDestroyContext");    

    assert(ctx);
    opcHolderFreeContext(ctx);
    opcFree(ctx->malloc,ctx);
}

void opcHolderMantain(OPCHolderContext ctx) {
    unsigned long cleaned;
    
    opcDebugServer(OPC_DEBUG_DEBUG, "opcHolderMantain");
    
    assert(ctx);
    
    LockWrite(ctx->lock);
    cleaned = DoubleHashClean(ctx->double_hash, (DoubleHashCleanFunctionT)opcHolderBucketMantain, NULL);
    if (cleaned) ConditionBroadcast(ctx->cond);
        
    // ToDo: reduce size if overloaded
    LockReturn(ctx->lock);
}


unsigned long opcHolderPut(OPCHolderContext ctx, OPCHolderBucketContext bucket) {
    unsigned long pos;

    assert(ctx);
    assert(bucket);

    LockWrite(ctx->lock);
    pos=DoubleHashInsert(ctx->double_hash, bucket, (unsigned long)bucket);
    if (pos==(unsigned long)-2) ConditionPreLock(ctx->cond);
    LockReturn(ctx->lock);
    
    while (pos==(unsigned long)-2) {
	if (ctx->timeout) {
	    pos=ConditionTimedWait(ctx->cond, ctx->timeout);
	    if (pos) {
		opcDebugServer(OPC_DEBUG_ERROR, "opcHolderPut: Timed Out, bucket: %p", bucket);
		return (unsigned long)-1;
	    }
	} else ConditionWait(ctx->cond);
	
	LockWrite(ctx->lock);
	pos=DoubleHashInsert(ctx->double_hash, bucket, (unsigned long)bucket);
	if (pos==(unsigned long)-2) ConditionPreLock(ctx->cond);
	LockReturn(ctx->lock);
    }
    
    opcDebugServer(OPC_DEBUG_DEBUG, "opcHolderPut: %lu, bucket: %p", pos, bucket);    

    return pos;
}


OPCHolderBucketContext opcHolderGet(OPCHolderContext ctx, unsigned long pos, unsigned long magic, unsigned long secret) {
    OPCHolderBucketContext bucket;

    assert(ctx);
    
    LockWrite(ctx->lock);

    bucket=(OPCHolderBucketContext)DoubleHashGetLink(ctx->double_hash, pos);
    if (!opcHolderBucketCheckMagicNumber(bucket, magic)) {
	opcDebugServer(OPC_DEBUG_WARNING, "opcHolderGet: %lu, bucket: %p. Invalid Magic Number!", pos, bucket);
	LockReturn(ctx->lock);
	return NULL;
    }
    if (!opcHolderBucketCheckSecretNumber(bucket, secret)) {
	opcDebugServer(OPC_DEBUG_WARNING, "opcHolderGet: %lu, bucket: %p. Invalid Secret Number!", pos, bucket);
	LockReturn(ctx->lock);
	return NULL;
    }

    opcHolderBucketLock(bucket);
    DoubleHashGet(ctx->double_hash, pos);    
    LockReturn(ctx->lock);
    opcHolderBucketUnlock(bucket);
    
    ConditionSend(ctx->cond);
    
    opcDebugServer(OPC_DEBUG_DEBUG, "opcHolderGet: %lu, bucket: %p", pos, bucket);    

    return bucket;
}

const void* opcHolderGetLink(OPCHolderContext ctx, unsigned long pos, unsigned long magic, unsigned long secret) {
    OPCHolderBucketContext bucket;

    assert(ctx);
    
    LockWrite(ctx->lock);
    bucket=(OPCHolderBucketContext)DoubleHashGetLink(ctx->double_hash, pos);   
    if (!opcHolderBucketCheckMagicNumber(bucket, magic)) {
	opcDebugServer(OPC_DEBUG_WARNING, "opcHolderGetLink: %lu, bucket: %p. Invalid Magic Number!", pos, bucket);
	LockReturn(ctx->lock);
	return NULL;
    }
    if (!opcHolderBucketCheckSecretNumber(bucket, secret)) {
	opcDebugServer(OPC_DEBUG_WARNING, "opcHolderGetLink: %lu, bucket: %p. Invalid Secret Number!", pos, bucket);
	LockReturn(ctx->lock);
	return NULL;
    }
    opcHolderBucketLock(bucket);
    LockReturn(ctx->lock);
    
    return bucket->data;
}

void opcHolderReturnLink(OPCHolderContext ctx, unsigned long pos) {
    OPCHolderBucketContext bucket;
    
    assert(ctx);
    
    bucket=(OPCHolderBucketContext)DoubleHashGetLink(ctx->double_hash, pos);
    assert(bucket);
    
    opcHolderBucketRestartTimer(bucket);
    opcHolderBucketUnlock(bucket);
}
