#include <string.h>
#include <time.h>
#include <assert.h>

#include <opcconfig.h>

#include <opcmalloc.h>
#include <opcdebug.h>
#include <lock.h>

#include <opcholderbucket.h>

OPCHolderBucketContext opcHolderBucketInitContext(OPCMallocContext mctx, void *ptr, time_t lifetime, time_t steptime, DestroyContextFunction destructor) {
    OPCHolderBucketContext ctx;
    
    ctx = (OPCHolderBucketContext)opcMalloc(mctx, sizeof(struct OPCHolderBucketContextT));
    opcDebugServer(OPC_DEBUG_NOTICE, "opcHolderBucketInitContext: %p", ctx);    
    if (!ctx) return NULL;

    ctx->malloc = mctx;

    ctx->lock = LockCreateContext(mctx);
    if (!ctx->lock) {
	opcFree(mctx,ctx);
	return NULL;
    }
    if (LockInitContext(ctx->lock)) {
	LockDestroyContext(ctx->lock);
	opcFree(mctx,ctx);
	return NULL;
    }

    if (lifetime) {
	ctx->destruction_time=time(NULL)+lifetime;
	ctx->step_time=steptime;
    } else ctx->destruction_time=0;
    ctx->destructor=destructor;
    ctx->data=ptr;
    return ctx;
}

void opcHolderBucketDestroyContext(OPCHolderBucketContext ctx) {
    opcDebugServer(OPC_DEBUG_NOTICE, "opcHolderBucketDestroyContext: %p", ctx);    

    assert(ctx);
    
    /* should be unlinked from opcHolder at this point */
    LockWrite(ctx->lock);
    LockReturn(ctx->lock);
    LockDestroyContext(ctx->lock);
    opcFree(ctx->malloc, ctx);
}

void opcHolderBucketLock(OPCHolderBucketContext ctx) {
    assert(ctx);
    
    LockWrite(ctx->lock);
}

void opcHolderBucketUnlock(OPCHolderBucketContext ctx) {
    assert(ctx);
    
    LockReturn(ctx->lock);
}

void opcHolderBucketRestartTimer(OPCHolderBucketContext ctx) {
    time_t destruction_time;

    assert(ctx);
    
    if (ctx->destruction_time) {
	destruction_time = time(NULL) + ctx->step_time;
	if (destruction_time>ctx->destruction_time) ctx->destruction_time=destruction_time;
    }
}


int opcHolderBucketMantain(OPCHolderBucketContext ctx, void *attr) {
    assert(ctx);
    
    if (LockTryWrite(ctx->lock)) return 0;
    LockReturn(ctx->lock); /* Since opcHolderContext is locked nobody can accquire new lock */
    if (ctx->destruction_time > time(NULL)) return 0;
    opcHolderBucketDestroyContext(ctx);
    return 1;
}

int opcHolderBucketSetNumbers(OPCHolderBucketContext ctx, unsigned long magic) {
    assert(ctx);
    
    ctx->magic = magic;
    ctx->secret = (unsigned long)ctx->data;
}

int opcHolderBucketCheckMagicNumber(OPCHolderBucketContext ctx, unsigned long magic) {
    assert(ctx);
    
    return (ctx->magic==magic);
}

int opcHolderBucketCheckSecretNumber(OPCHolderBucketContext ctx, unsigned long secret) {
    assert(ctx);
    
    return (ctx->secret==secret);
}
