#include <assert.h>

#include <opcerr.h>
#include <opcmalloc.h>
#include <opctools.h>

#include <opcconfiguration.h>

void opcConfigurationZeroContext(OPCConfigurationContext ctx) {
    ctx->data_directory = NULL;
    ctx->schema_directory = NULL;
    ctx->opcbinary_schema = NULL;
    ctx->dictionary_directory = NULL;
    ctx->opcbinary_dictionaries = NULL;
    ctx->namespaces.xsi_namespace = NULL;
    ctx->namespaces.opc_namespace = NULL;
    ctx->namespaces.opcbinary_namespace = NULL;
}

OPCConfigurationContext opcConfigurationCreateContext(OPCMallocContext mctx) {
    OPCConfigurationContext ctx;
    ctx = (OPCConfigurationContext)opcMalloc(mctx, sizeof(struct OPCConfigurationContextT));
    if (ctx) {
	ctx->malloc = mctx;
	ctx->namespaces.malloc = mctx;
	opcConfigurationZeroContext(ctx);
    }
    return ctx;
}

int opcConfigurationInitContext(OPCConfigurationContext ctx, unsigned long scalability) {
    opcConfigurationZeroContext(ctx);
    opcConfigurationDefault(ctx, scalability);
    return 0;
}

void opcConfigurationFreeContext(OPCConfigurationContext ctx) {
    assert(ctx);
    if (ctx->data_directory) opcExchangeString(ctx->malloc, &ctx->data_directory, NULL);
    if (ctx->schema_directory) opcFree(ctx->malloc, ctx->schema_directory);
    if (ctx->opcbinary_schema) opcFree(ctx->malloc, ctx->opcbinary_schema);
    if (ctx->dictionary_directory) opcFree(ctx->malloc, ctx->dictionary_directory);
    if (ctx->opcbinary_dictionaries) opcFree(ctx->malloc, ctx->opcbinary_dictionaries);
    if (ctx->namespaces.xsi_namespace) opcFree(ctx->malloc, ctx->namespaces.xsi_namespace);
    if (ctx->namespaces.opc_namespace) opcFree(ctx->malloc, ctx->namespaces.opc_namespace);
    if (ctx->namespaces.opcbinary_namespace) opcFree(ctx->malloc, ctx->namespaces.opcbinary_namespace);

    opcConfigurationZeroContext(ctx);
}

void opcConfigurationDestroyContext(OPCConfigurationContext ctx) {
    assert(ctx);
    opcConfigurationFreeContext(ctx);
    opcFree(ctx->malloc, ctx);
}


void opcConfigurationDefault(OPCConfigurationContext ctx, unsigned long scalability) {
    assert(ctx);
    
    opcConfigurationSetBrowseTimeout(ctx, OPC_CONFIGURATION_BROWSE_TIMEOUT);
    opcConfigurationSetResourceWaitTimeout(ctx, OPC_CONFIGURATION_RESOURCE_TIMEOUT);
    opcConfigurationSetRequestTimeout(ctx, OPC_CONFIGURATION_REQUEST_TIMEOUT);
    opcConfigurationSetHolderSize(ctx, OPC_CONFIGURATION_HOLDER_SIZE, OPC_CONFIGURATION_HOLDER_MAXIMAL_SIZE);
    opcConfigurationSetFormatsHashSize(ctx, OPC_CONFIGURATION_FORMATS_HASH_SIZE);
    opcConfigurationSetPropertiesHashSize(ctx, OPC_CONFIGURATION_PROPERTIES_HASH_SIZE);
    opcConfigurationSetPropertiesIDs(ctx, OPC_CONFIGURATION_PROPERTY_MAXID, OPC_CONFIGURATION_PROPERTY_USERID);
    opcConfigurationSetChunkPoolConfig(ctx, OPC_CONFIGURATION_CHUNK_POOL_BINS);
    opcConfigurationSetBufferPoolConfig(ctx, OPC_CONFIGURATION_BUFFER_POOL_SMALLEST, OPC_CONFIGURATION_BUFFER_POOL_BINS);
    opcConfigurationSetDataDirectory(ctx, OPC_CONFIGURATION_DATA_DIRECTORY);
    opcConfigurationSetSchemaNamespace(ctx, OPC_CONFIGURATION_NAMESPACE_XSI);
}

int opcConfigurationSetBrowseTimeout(OPCConfigurationContext ctx, time_t browse_timeout) {
    assert(ctx);
    ctx->browse_timeout = browse_timeout;
    return 0;
}

int opcConfigurationSetResourceWaitTimeout(OPCConfigurationContext ctx, time_t resource_timeout) {
    assert(ctx);
    ctx->resource_wait_timeout = resource_timeout;
    return 0;
}

int opcConfigurationSetRequestTimeout(OPCConfigurationContext ctx, time_t request_timeout) {
    assert(ctx);
    ctx->request_timeout = request_timeout;
    return 0;
}

int opcConfigurationSetHolderSize(OPCConfigurationContext ctx, unsigned long holder_size, unsigned long holder_maximal_size) {
    assert(ctx);
    ctx->holder_size = holder_size;
    ctx->holder_maximal_size = holder_maximal_size;
    return 0;
}

int opcConfigurationSetFormatsHashSize(OPCConfigurationContext ctx, unsigned long formats_hash_size) {
    assert(ctx);
    ctx->formats_hash_size = formats_hash_size;
    return 0;
}

int opcConfigurationSetPropertiesHashSize(OPCConfigurationContext ctx, unsigned long properties_hash_size) {
    assert(ctx);
    ctx->properties_hash_size = properties_hash_size;
    return 0;
}

int opcConfigurationSetPropertiesIDs(OPCConfigurationContext ctx, unsigned long maxid, unsigned long userid) {
    assert(ctx);
    ctx->property_maxid = maxid;
    ctx->property_userid = userid;
    return 0;
}

int opcConfigurationSetChunkPoolConfig(OPCConfigurationContext ctx, unsigned char bins) {
    assert(ctx);
    assert(bins);
    
    ctx->chunk_pool_bins = bins;
    return 0;
}

int opcConfigurationSetBufferPoolConfig(OPCConfigurationContext ctx, unsigned long smallest, unsigned char bins) {
    assert(ctx);
    assert(smallest>15);
    assert(bins);
    
    ctx->buffer_pool_smallest = smallest;
    ctx->buffer_pool_bins = bins;
    return 0;
}

int opcConfigurationSetDataDirectory(OPCConfigurationContext ctx, const char *data_directory) {
    int length, slength;
    assert(ctx);
    assert(data_directory);
    
    if (opcExchangeString(ctx->malloc, &ctx->data_directory, data_directory)) return ERR_OPC_MEM;
    length = strlen(data_directory);
    slength = length + sizeof(OPC_CONFIGURATION_SCHEMA_DIRECTORY) + 2;
    ctx->schema_directory = (char*)opcMalloc(ctx->malloc, slength*sizeof(char));
    if (!ctx->schema_directory) return ERR_OPC_MEM;
    sprintf(ctx->schema_directory, "%s/%s",data_directory, OPC_CONFIGURATION_SCHEMA_DIRECTORY);
    ctx->opc_schema = (char*)opcMalloc(ctx->malloc, (slength+sizeof(OPC_CONFIGURATION_OPC_SCHEMA)+1)*sizeof(char));
    if (!ctx->opc_schema) return ERR_OPC_MEM;
    sprintf(ctx->opc_schema, "%s/%s",ctx->schema_directory, OPC_CONFIGURATION_OPC_SCHEMA);
    ctx->opcbinary_schema = (char*)opcMalloc(ctx->malloc, (slength+sizeof(OPC_CONFIGURATION_OPCBINARY_SCHEMA)+1)*sizeof(char));
    if (!ctx->opcbinary_schema) return ERR_OPC_MEM;
    sprintf(ctx->opcbinary_schema, "%s/%s",ctx->schema_directory, OPC_CONFIGURATION_OPCBINARY_SCHEMA);

    slength = length + sizeof(OPC_CONFIGURATION_DICTIONARY_DIRECTORY)+2;
    ctx->dictionary_directory = (char*)opcMalloc(ctx->malloc, slength*sizeof(char));
    if (!ctx->dictionary_directory) return ERR_OPC_MEM;
    sprintf(ctx->dictionary_directory, "%s/%s",data_directory,OPC_CONFIGURATION_DICTIONARY_DIRECTORY);
    ctx->opcbinary_dictionaries = (char*)opcMalloc(ctx->malloc, (slength+sizeof(OPC_CONFIGURATION_OPCBINARY_DICTIONARIES_DIRECTORY)+1)*sizeof(char));
    if (!ctx->opcbinary_dictionaries) return ERR_OPC_MEM;
    sprintf(ctx->opcbinary_dictionaries,"%s/%s",ctx->dictionary_directory, OPC_CONFIGURATION_OPCBINARY_DICTIONARIES_DIRECTORY);

    return 0;
}

int opcConfigurationSetSchemaNamespace(OPCConfigurationContext ctx, const char *xsi_namespace) {
    assert(ctx);
    assert(xsi_namespace);
    if (opcExchangeString(ctx->malloc, &ctx->namespaces.xsi_namespace, xsi_namespace)) return ERR_OPC_MEM;
//    ctx->namespaces.xsins = xmlNewNs(NULL,xsi_namespace, "xsi"); 
        
    return 0;
}

int opcNamespacesSetOPCBinaryNamespace(OPCNamespaces ctx, const char *opcbinary_namespace) {
    assert(ctx);
    if (opcExchangeString(ctx->malloc, &ctx->opcbinary_namespace, opcbinary_namespace)) return ERR_OPC_MEM;
    return 0;
}

int opcNamespacesSetOPCNamespace(OPCNamespaces ctx, const char *opc_namespace) {
    assert(ctx);
    assert(opc_namespace);
    if (opcExchangeString(ctx->malloc, &ctx->opc_namespace, opc_namespace)) return ERR_OPC_MEM;
//    ctx->namespaces.opcns = xmlNewNs(NULL,opc_namespace, NULL); 
    return 0;
}

