#ifndef _OPC_DATA_H
#define _OPC_DATA_H

#include <lock.h>
#include <opcmalloc.h>
#include <hashtable.h>

#include <opcformats.h>
#include <opcchunkpool.h>
#include <opcformatnative.h>

#include <opctype.h>
#include <opctime.h>
#include <opcquality.h>
#include <opcformat.h>
#include <opcdatachunk.h>
#include <opcallocator.h>
#include <opcdatamanager.h>
#include <opcdatarepresentation.h>
#include <opcrepresentation.h>

#define OPC_DATA_HASHTABLE_SIZE 16

typedef struct OPCDataContextT *OPCDataContext;
typedef struct OPCDataOptionsT *OPCDataOptions;

struct OPCDataContextT {
    struct OPCDataManagerContextT manager;

    OPCMallocContext malloc;

    LockContext lock;
    LockContext chunk_lock;
    LockContext representation_lock;
    unsigned long refs;
    unsigned long nchunks;

    OPCChunkPoolContext chunk_pool;

    unsigned short int supported_representations;
    unsigned short int allocated_representations;

    OPCRepresentationContext *representation;
    OPCRepresentationInfo *repinfo;
    HashTable rephash;
    
    OPCDataRepresentationContext native_representation;
    OPCDataRepresentationContext opcxml_representation;

    OPCDataChunkContext chunks;
    OPCDataChunkContext current;
    
    OPCFormatsContext formats;
    OPCAllocatorContext binary_allocator;
    OPCBufferPoolContext buffer_pool;
    unsigned long cache_min, cache_max;
    unsigned long binary_minbuf_size, binary_maxbuf_size;
    unsigned long flags;
};

struct OPCDataOptionsT {
    OPCDataManagerAcquireFunction acquire;
    OPCChunkPoolContext chunk_pool;
    OPCFormatsContext formats;
    unsigned long cache_min;
    unsigned long cache_max;
    unsigned long flags;
    OPCAllocatorContext binary_allocator;
    OPCBufferPoolContext buffer_pool;
    unsigned long binary_minbuf_size;
    unsigned long binary_maxbuf_size;
};


OPCDataContext opcDataCreateContext(OPCMallocContext mctx);
int opcDataInitContext(OPCDataContext ctx, OPCDataOptions options);
int opcDataRef(OPCDataContext ctx);
void opcDataUnref(OPCDataContext ctx);

OPCDataOptions opcDataGetDefaultOptions();
int opcDataInitDefaultOptions(OPCDataOptions options);
void opcDataDestroyOptions(OPCDataOptions options);

int opcDataOptionsSetAcquireFunction(OPCDataOptions options, OPCDataManagerAcquireFunction acquire);
int opcDataOptionsSetChunkPool(OPCDataOptions options, OPCChunkPoolContext chunk_pool);
int opcDataOptionsSetCacheSize(OPCDataOptions options, unsigned long min, unsigned long max);
int opcDataOptionsSetDataSourceFlags(OPCDataOptions options, unsigned long flags);

int opcDataOptionsSetFormatsContext(OPCDataOptions options, OPCFormatsContext formats);
int opcDataOptionsSetBufferSize(OPCDataOptions options, unsigned long min, unsigned long max);
int opcDataOptionsSetBinaryAllocator(OPCDataOptions options, OPCAllocatorContext allocator);
int opcDataOptionsSetBufferPool(OPCDataOptions options, OPCBufferPoolContext buffer_pool);


int opcDataSetDataRepresentationOptions(OPCDataContext ctx, OPCDataRepresentationOptions options, OPCFormatContext format);
unsigned short int opcDataRegisterRepresentation(OPCDataContext ctx, OPCRepresentationContext representation);
int opcDataRegisterNativeRepresentation(OPCDataContext ctx, OPCType type, OPCDataRepresentationOptions options);
int opcDataRegisterOPCXMLRepresentation(OPCDataContext ctx, OPCType type, OPCDataRepresentationOptions options);

unsigned long opcDataGetId(OPCDataContext ctx, const char *repname);
int opcDataPushData(OPCDataContext ctx, unsigned long repid, OPCDataInfo info, const void *data, unsigned long size);
int opcDataGetData(OPCDataContext ctx, unsigned long repid, OPCDataInfo *info, OPCRepresentationData *data);
void opcDataReturnData(OPCDataContext ctx, OPCDataInfo info);
void opcDataMantain(OPCDataContext ctx);

OPCAllocatorContext opcDataGetAllocator(OPCDataContext ctx, unsigned long repid, OPCDataInfo info);
void opcDataReturnAllocator(OPCDataContext ctx, unsigned long repid, OPCAllocatorContext allocator);
int opcDataPushAllocation(OPCDataContext ctx, unsigned long repid, OPCAllocatorContext allocator, OPCDataInfo info, OPCAllocation allocation);


#ifdef _OPC_DATA_C
static const struct OPCDataManagerContextT opc_datamanager_data = {
    (OPCDataManagerGetIdFunction)opcDataGetId,			/* GetID */
    NULL,							/* Acquire (Costumized by user) */
    NULL,							/* PushData */
    NULL, 							/* PushAllocation */
    NULL,							/* GetAllocator */
    NULL,							/* ReturnAllocator */
    (OPCDataManagerGetDataFunction)opcDataGetData,		/* GetData */
    NULL,							/* GetDataFirst */
    NULL,							/* GetDataNext */
    (OPCDataManagerReturnDataFunction)opcDataReturnData,	/* ReturnData */
    (OPCDataManagerRefFunction)opcDataRef,			/* Ref */
    (OPCDataManagerUnrefFunction)opcDataUnref,			/* Unref */
    (OPCDataManagerMantainFunction)opcDataMantain		/* Mantain */
};

static const struct OPCDataOptionsT opc_datamanager_data_options = {
    opcAcquireNotApplicable,			/* Acquire Function */
    NULL,					/* ChunkPool (by user) */
    NULL,					/* Formats (by user) */
    1,						/* Minimum items to cache */
    0,						/* Maximum items to cache */
    0,						/* DataSource flags */
    NULL,					/* Default Allocator for Binary Representations */
    NULL,					/* Default BufferPool for DataBuffer allocator */
    0,						/* Minimal buffer size for Binary Representations */
    0						/* Maximal buffer size for Binary Representations */
};

#endif /* _OPC_DATA_C */

#endif /* _OPC_DATA_H */
