#define _OPC_DATA_FAST_C

#include <opcerr.h>
#include <opcdebug.h>
#include <opcassert.h>

#include <opcdatafast.h>

OPCDataFastOptions opcDataFastGetDefaultOptions() {
    int err;
    OPCDataFastOptions options;

    options = (OPCDataFastOptions)malloc(sizeof(struct OPCDataFastOptionsT));
    if (!options) return options;
    
    err = opcDataFastInitDefaultOptions(options);
    if (err) {
	opcDataFastDestroyOptions(options);
	return NULL;
    }
    return options;
}

int opcDataFastInitDefaultOptions(OPCDataFastOptions options) {
    assert(options);
    memcpy(options, &opc_datamanager_fast_options, sizeof(struct OPCDataFastOptionsT));
    opcDataBufferInitDefaultOptions(&options->allocator);
    return 0;
}

int opcDataFastOptionsSetAcquireFunction(OPCDataFastOptions options, OPCDataManagerAcquireFunction acquire) {
    assert(options);
    options->acquire = acquire;
    return 0;
}

int opcDataFastOptionsSetUserizeFunction(OPCDataFastOptions options, OPCRepresentationGetUserDataFunction get) {
    assert(options);
    options->get = get;
    return 0;
}

int opcDataFastOptionsSetBufferSize(OPCDataFastOptions options, unsigned long min, unsigned long max) {
    assert(options);
    options->allocator.minimal_size = min;
    options->allocator.maximal_size = max;
    return 0;
}

int opcDataFastOptionsSetCacheSize(OPCDataFastOptions options, unsigned long min, unsigned long max) {
    assert(options);
    options->buffer_min = min;
    options->buffer_max = max;
    return 0;
}

void opcDataFastDestroyOptions(OPCDataFastOptions options) {
    if (options) free(options);
}


static inline void opcDataFastZeroContext(OPCDataFastContext ctx) {
    assert(ctx);

    ctx->allocator = NULL;
    ctx->chunks = NULL;
    ctx->current = NULL;
    ctx->lock = NULL;
    ctx->chunk_lock = NULL;
    ctx->allocator_lock = NULL;
    ctx->grabed_allocator = 0;
    ctx->nchunks = 0;
}

static void opcDataFastFreeContext(OPCDataFastContext ctx) {
    OPCDataChunkContext chunk, nchunk;
    long i;

    opcDebugServer(OPC_DEBUG_NOTICE, "opcDataFast(%p): FreeContext", ctx);    

    assert(ctx);
    
    for (chunk=ctx->chunks;chunk;chunk=nchunk) {
        nchunk = chunk->next;
	opcDataChunkWaitUnref(chunk, 0);
        opcDataChunkDestroyContext(chunk);
    }

    if (ctx->grabed_allocator) opcDataFastReturnAllocator(ctx);
    if (ctx->allocator) opcDataBufferDestroyContext(ctx->allocator);
    if (ctx->lock) LockDestroyContext(ctx->lock);
    if (ctx->chunk_lock) LockDestroyContext(ctx->chunk_lock);
    opcDataFastZeroContext(ctx);    
}

static void opcDataFastDestroyContext(OPCDataFastContext ctx) {
    opcDebugServer(OPC_DEBUG_NOTICE, "opcDataFast(%p): DestroyContext", ctx);    
    opcDataFastFreeContext(ctx);
    opcFree(ctx->malloc,ctx);
}


OPCDataFastContext opcDataFastCreateContext(OPCMallocContext mctx) {
    OPCDataFastContext ctx;

    ctx = (OPCDataFastContext)opcMalloc(mctx,sizeof(struct OPCDataFastContextT));
    opcDebugServer(OPC_DEBUG_NOTICE, "opcDataFast(%p): CreateContext", ctx);    
    if (ctx) {
	ctx->malloc=mctx;
	ctx->refs = 1;
	opcDataFastZeroContext(ctx);
	memcpy(&ctx->manager, &opc_datamanager_fast, sizeof(struct OPCDataManagerContextT));
    }
    return ctx;
}


int opcDataFastInitContext(OPCDataFastContext ctx, OPCType type, OPCDataFastOptions options) {
    int err;

    opcDebugServer(OPC_DEBUG_NOTICE, "opcDataFast(%p): InitContext (Type: %p)", ctx, type);

    assert(ctx);
    assert(type);
    assert(options);

    if ((!ctx)||(!type)||(!options)) return ERR_OPC_DATAFAST_PARAMS;
    if ((!options->chunk_pool)||(!options->format)) return ERR_OPC_DATAFAST_OPTIONS;
    
    opcDataFastZeroContext(ctx);    

    ctx->type = type;
    ctx->chunk_pool = options->chunk_pool;
    ctx->format = options->format;
    ctx->repinfo.formats = options->formats;
    ctx->manager.Acquire = options->acquire;
    if (options->buffer_min) ctx->buffer_min = options->buffer_min;
    else ctx->buffer_min = 1;
    ctx->buffer_max = options->buffer_max;

    ctx->allocator = opcDataBufferCreateContext(ctx->malloc);
    if (!ctx->allocator) return ERR_OPC_DATAFAST_CREATE_ALLOCATOR;
    
    err = opcDataBufferInitContext(ctx->allocator, &options->allocator);
    if (err) {
	opcDataFastFreeContext(ctx);
	return err;
    }

    opcRepresentationDataDataBuffer((OPCRepresentationInfo)&ctx->repinfo, ctx->allocator);
    if (options->get) ctx->repinfo.repinfo.GetUserData = options->get;
    ctx->repinfo.datafast = ctx;

    ctx->lock = LockCreateContext(ctx->malloc);
    ctx->chunk_lock = LockCreateContext(ctx->malloc);
    ctx->allocator_lock = LockCreateContext(ctx->malloc);
    if ((!ctx->lock)||(!ctx->chunk_lock)||(!ctx->allocator_lock)) {
	opcDataFastFreeContext(ctx);
	return ERR_OPC_DATAFAST_CREATE_LOCK;
    }

    err = LockInitContext(ctx->lock);
    if (!err) err = LockInitContext(ctx->chunk_lock);
    if (!err) err = LockInitContext(ctx->allocator_lock);
    if (err) {
	opcDataFastFreeContext(ctx);
	return err;
    }
    return 0;
}

int opcDataFastSetAcquireFunction(OPCDataFastContext ctx, OPCDataManagerAcquireFunction acquire) {
    assert(ctx);
    
    ctx->manager.Acquire = acquire;
    return 0;
}

int opcDataFastSetUserizeFunction(OPCDataFastContext ctx, OPCRepresentationGetUserDataFunction get) {
    assert(ctx);
    
    ctx->repinfo.repinfo.GetUserData = get;
    return 0;
}

int opcDataFastRef(OPCDataFastContext ctx) {
    assert(ctx);
    
    LockWrite(ctx->lock);
    ctx->refs++;
    LockReturn(ctx->lock);
}

void opcDataFastUnref(OPCDataFastContext ctx) {
    assert(ctx);

    if (!ctx->lock) {
	opcDataFastDestroyContext(ctx);
	return;
    }
    
    LockWrite(ctx->lock);
    if ((--ctx->refs)==0) {
	LockReturn(ctx->lock);
	opcDataFastDestroyContext(ctx);
	return;
    }
    LockReturn(ctx->lock);
}


static inline void opcDataFastClear(OPCDataFastContext ctx) {
    LockContext chunk_lock;
    OPCChunkPoolContext chunk_pool;
    OPCDataChunkContext chunk, nchunk;
    unsigned long buffer_min, buffer_max;
    unsigned long ref = 0;

	/* After chunk_lock locked number of referers can decrease, but not increase! */

    chunk_lock = ctx->chunk_lock;
    chunk_pool = ctx->chunk_pool;
    chunk = ctx->chunks;

    buffer_min = ctx->buffer_min;
    buffer_max = ctx->buffer_max;

    nchunk = chunk;

    if (buffer_max) 
        for (;((ctx->nchunks>buffer_max)&&(!ref));chunk=nchunk) {
            nchunk = chunk->next;
	    ctx->chunks=nchunk;
	    ctx->nchunks--;
	    LockReturn(chunk_lock);
	    opcChunkPoolReturnChunk(chunk_pool, chunk);
	    LockWrite(chunk_lock);
	    ref=opcDataChunkGetRefs(nchunk);
	}
	
    for (;((!ref)&&(ctx->nchunks>buffer_min));chunk=nchunk) {
	nchunk=chunk->next;
	ctx->chunks=nchunk;
	ctx->nchunks--;
	LockReturn(chunk_lock);
	opcChunkPoolReturnChunk(chunk_pool, chunk);
	LockWrite(chunk_lock);
	ref = opcDataChunkGetRefs(nchunk);
	if (!ref) ref = opcDataChunkGetDelayedRefs(nchunk);
    }
    
}


int opcDataFastPushData(OPCDataFastContext ctx, unsigned long repid, OPCDataInfo info, const void *data, unsigned long size) {
    int err;
    OPCRepresentationData repdata;
    OPCRepresentationInfo repinfo;

    OPCDataChunkContext chunk;

    assert(ctx);
    assert(data);

    opcDebugServer(OPC_DEBUG_DEBUG, "opcDataFast(%p): Push", ctx);

    if ((!ctx)||(!data)) return ERR_OPC_DATAFAST_PARAMS;
    if (ctx->grabed_allocator) return ERR_OPC_DATAFAST_ALLOCATOR_BUSY;
    
    repinfo = (OPCRepresentationInfo)&ctx->repinfo;
    chunk = opcChunkPoolGetChunk(ctx->chunk_pool, &repinfo, 1);
    if (!chunk) return ERR_OPC_CHUNKPOOL_GET_CHUNK;
    opcDataChunkCopyInfo(chunk, info);
    
    LockWrite(ctx->allocator_lock);
    repdata=opcDataChunkGetData(chunk, 0);
    err = opcFormatNativeImportData(ctx->format, ctx->type, (OPCAllocatorContext)ctx->allocator, (OPCData)repdata, data, size);
	/* ToDo: prelock datachunk for all subscriptions may need it */
    LockReturn(ctx->allocator_lock);

    if (err) {
	opcChunkPoolReturnChunk(ctx->chunk_pool, chunk);
	return err;
    }
    
    LockWrite(ctx->chunk_lock);
    ctx->nchunks++;
    if (ctx->current) {
	/* What to do with OOB items, with time earlier than current one?
	For Fast items can be assumed only one sequential data source? For items
	with multiple or/and non-sequential data sources OPCDataContext should be used
	if (opcTimeCmp(ctx->current->info.time, info->time)>0) {
	}
	*/
	ctx->current->next = chunk;
	ctx->current=chunk;

	if ((ctx->nchunks>ctx->buffer_min)&&(!opcDataChunkGetRefs(ctx->chunks))) 
	    opcDataFastClear(ctx);
    } else {
	ctx->chunks = chunk;
	ctx->current = chunk;
    }
    LockReturn(ctx->chunk_lock);
    
    return 0;
}

int opcDataFastPushAllocation(OPCDataFastContext ctx, unsigned long repid, OPCDataInfo info, OPCAllocation allocation) {
    int err;
    OPCRepresentationInfo repinfo;
    OPCRepresentationData repdata;
    OPCDataChunkContext chunk;
    
    assert(ctx);
    assert(allocation);

    opcDebugServer(OPC_DEBUG_DEBUG, "opcDataFast(%p): Push", ctx);

    if ((!ctx)||(!allocation)) return ERR_OPC_DATAFAST_PARAMS;

    repinfo = (OPCRepresentationInfo)&ctx->repinfo;
    chunk = opcChunkPoolGetChunk(ctx->chunk_pool, &repinfo, 1);
    if (!chunk) return ERR_OPC_CHUNKPOOL_GET_CHUNK;
    opcDataChunkCopyInfo(chunk, info);
    
    repdata=opcDataChunkGetData(chunk, 0);
    repdata->data = allocation;
    repdata->size = opcDataBufferSizeof(ctx->allocator, allocation, 1);
    repdata->type = OPC_DATA_TYPE_DATABUFFER;
	/* ToDo: prelock datachunk for all subscriptions may need it */

    LockWrite(ctx->chunk_lock);
    ctx->nchunks++;
    if (ctx->current) {
	/* What to do with OOB items, with time earlier than current one?
	For Fast items can be assumed only one sequential data source? For items
	with multiple or/and non-sequential data sources OPCDataContext should be used
	if (opcTimeCmp(ctx->current->info.time, info->time)>0) {
	}
	*/
	ctx->current->next = chunk;
	ctx->current = chunk;

	if ((ctx->nchunks>ctx->buffer_min)&&(!opcDataChunkGetRefs(ctx->chunks))) 
	    opcDataFastClear(ctx);
    } else {
	ctx->chunks = chunk;
	ctx->current = chunk;
    }
    LockReturn(ctx->chunk_lock);

    
    return 0;
}

int opcDataFastGetData(OPCDataFastContext ctx, unsigned long repid, OPCDataInfo *info, OPCRepresentationData *data) {
    int err;
    OPCDataChunkContext current;
    OPCRepresentationData rep;

    assert(ctx);
    assert(ctx->current);

    if (!ctx) return ERR_OPC_DATAFAST_PARAMS;
    if (!ctx->current) return ERR_OPC_DATAFAST_NOTINITIALISED;

    LockRead(ctx->chunk_lock);
    current = ctx->current;
    opcDataChunkRef(current);
    LockReturn(ctx->chunk_lock);

    if (data) *data=opcDataChunkGet(current, 0);
    if (info) *info=opcDataChunkGetInfo(current);
    
    opcDataFastRef(ctx);
    
    return 0;
}

void opcDataFastReturnData(OPCDataFastContext ctx, OPCDataInfo info) {
    assert(ctx);
    if ((!ctx)||(!info)) return;

    LockWrite(ctx->chunk_lock);
    if ((!opcDataChunkUnrefByInfo(info))&&(ctx->chunks == opcDataChunkGetChunkByInfo(info)))
	opcDataFastClear(ctx);
    LockReturn(ctx->chunk_lock);
    
    opcDataFastUnref(ctx);
}


OPCAllocatorContext opcDataFastGetAllocator(OPCDataFastContext ctx) {
    assert(ctx);
    
    LockWrite(ctx->allocator_lock);
    ctx->grabed_allocator = 1;
    return (OPCAllocatorContext)ctx->allocator;
}

void opcDataFastReturnAllocator(OPCDataFastContext ctx) {
    assert(ctx);
    
    if (ctx->grabed_allocator) {
	ctx->grabed_allocator = 0;
	LockReturn(ctx->allocator_lock);
    }
}

void opcDataFastMantain(OPCDataFastContext ctx) {
    assert(ctx);
    if (ctx->allocator) opcDataBufferMantain(ctx->allocator);
}
