#include <assert.h>

#include <opcerr.h>
#include <opcdebug.h>

#include <opcformats.h>
#include <opcformatnative.h>
#include <opcformatopcxml.h>

void opcFormatsZeroContext(OPCFormatsContext ctx) {
    ctx->formats = NULL;
    ctx->binary_format = NULL;
    ctx->native_format = NULL;
    ctx->opcxml_format = NULL;
}

OPCFormatsContext opcFormatsCreateContext(OPCMallocContext mctx) {
    OPCFormatsContext ctx;

    opcDebugServer(OPC_DEBUG_NOTICE, "opcFormatsCreateContext");

    ctx = (OPCFormatsContext)opcMalloc(mctx,sizeof(struct OPCFormatsContextT));
    if (ctx) {
	ctx->malloc=mctx;
	opcFormatsZeroContext(ctx);
    }
    return ctx;
}

int opcFormatsInitContext(OPCFormatsContext ctx, OPCConfigurationContext config) {
    int err;
    unsigned long formats_hash_size;
    OPCFormatContext format;

    assert(ctx);
    assert(config);
    
    if ((!ctx)||(!config)) return ERR_OPC_PARAM;

    formats_hash_size = config->formats_hash_size;
    if (formats_hash_size<4) formats_hash_size=4;
    
    opcDebugServer(OPC_DEBUG_NOTICE, "opcFormatsInitContext");
    opcFormatsZeroContext(ctx);

    ctx->formats = HashTableInit(ctx->malloc, formats_hash_size, HASH_TYPE_POINTER, NULL);
    if (!ctx->formats) {
	opcFormatsFreeContext(ctx);
	return ERR_OPC_MEM;
    }
    HashTableSetDestructor(ctx->formats, (HashTableDestructorFunction)opcFormatDestroyContext);

    ctx->native_format = (OPCFormatNativeContext)opcFormatsRegisterFormat(ctx, opcFormatNative(), NULL);
    if (!ctx->native_format) {
	opcFormatsFreeContext(ctx);
	return ERR_OPC_FORMATS_REGISTER_NATIVE_FORMAT;
    }

    ctx->opcxml_format = opcFormatsRegisterFormat(ctx, opcFormatOPCXML(), NULL);
    if (!ctx->opcxml_format) {
	opcFormatsFreeContext(ctx);
	return ERR_OPC_FORMATS_REGISTER_OPCXML_FORMAT;
    }
  
    return 0;
}

void opcFormatsFreeContext(OPCFormatsContext ctx) {
    assert(ctx);
    opcDebugServer(OPC_DEBUG_NOTICE, "opcFormatsFreeContext");
    
    if (ctx->formats) HashTableDestroy(ctx->formats);
    opcFormatsZeroContext(ctx);
}

void opcFormatsDestroyContext(OPCFormatsContext ctx) {
    opcDebugServer(OPC_DEBUG_NOTICE, "opcFormatsDestroyContext");
    opcFormatsFreeContext(ctx);
    opcFree(ctx->malloc,ctx);
}


OPCFormatContext opcFormatsFindFormat(OPCFormatsContext ctx, const char *format) {
    assert(ctx);
    assert(format);
    
    return (OPCFormatContext)HashTableFind(ctx->formats, (void*)format);
}


OPCFormatContext opcFormatsRegisterFormat(OPCFormatsContext ctx, OPCFormatContext fctx, const void *options) {
    int err;
    OPCFormatContext format;
    
    assert(ctx);
    assert(fctx);
    
    format = opcFormatsFindFormat(ctx, fctx->title);
    if (format) {
//	opcDebugServer()
	return format;
    }
    
    format = opcFormatCreateContext(fctx, ctx->malloc);
    if (!format) {
//	opcDebugServer()
	return NULL;
    }
    
    err = opcFormatInitContext(format, options);
    if (err) {
	opcFormatDestroyContext(format);
//	opcDebugServer()
	return NULL;
    }

    if (HashTableInsert(ctx->formats, format)) {
	opcFormatDestroyContext(format);
//	opcDebugServer()
	return NULL;
    }
    
    return format;
}
