#include <stdio.h>
#include <assert.h>

#include <opcerr.h>
#include <opcconfig.h>
#include <opcdebug.h>

#include <opcmalloc.h>
#include <lchar.h>
#include <hashtable.h>
#include <lock.h>
#include <opcstringlist.h>

#include <opctree.h>
//#include <opcproperty.h>
#include <opcitem.h>

static inline void opcItemZeroDataContext(OPCItemContext ctx) {
    assert(ctx);
    
    ctx->dm = NULL;
//    ctx->pctx=NULL;
}

static inline void opcItemZeroContext(OPCItemContext ctx) {
    assert(ctx);

    ctx->lock = NULL;
    ctx->client_cond = NULL;
    
    ctx->fullname=NULL;
    ctx->name=NULL;
    ctx->title=NULL;
    
    ctx->status=0;
    ctx->manual=2;
    ctx->mark=0;
    ctx->subitems=NULL;
    
    ctx->tree=NULL;
    
    ctx->vendor_filter=NULL;
    
    ctx->clients = 0;
    
    ctx->representation = 0;
    
    opcItemZeroDataContext(ctx);
}

OPCItemContext opcItemCreateContext(OPCMallocContext mctx) {
    OPCMallocContext mymalloc;
    OPCItemContext ctx;

    if (mctx) {
	mymalloc = opcMallocCreateContext(mctx);
	if (!mymalloc) return NULL;
    } else mymalloc = NULL;

    ctx = (OPCItemContext)opcMalloc(mymalloc, sizeof(struct OPCItemContextT));
    if (ctx) {
	ctx->malloc = mymalloc;
	opcItemZeroContext(ctx);
    } else opcMallocDestroyContext(mymalloc);

    opcDebugServer(OPC_DEBUG_NOTICE, "opcItem(%p): CreateContext", ctx);

    return ctx;
}

int opcItemInitContext(OPCItemContext ctx, const char *fullname, const char *title) {
    int err;

    opcDebugServer(OPC_DEBUG_NOTICE, "opcItem(%p): InitContext: \"%s\".", ctx, fullname);    
    
    assert(ctx);
    assert(fullname);
    
    opcItemZeroContext(ctx);
    
    err = opcExchangeString(ctx->malloc, &ctx->fullname, fullname);
    if (err) return ERR_OPC_MEM;

/*
    ctx->name = strrchr(ctx->fullname, '/');
    if (ctx->name) ctx->name++;
    else ctx->name = ctx->fullname;
*/
    
    if (title) err=opcLCharExchangeString(ctx->malloc,&ctx->title, title);
    else err=opcLCharExchangeString(ctx->malloc,&ctx->title, fullname);
    
    if (err) {
	opcItemFreeContext(ctx);
	return ERR_OPC_MEM;
    }
    
    ctx->lock = LockCreateContext(ctx->malloc);
    if (!ctx->lock) {
	opcItemFreeContext(ctx);
	return ERR_OPC_MEM;
    }
    
    err = LockInitContext(ctx->lock);
    if (err) {
	opcItemFreeContext(ctx);
	return err;
    }
    
    ctx->client_cond = ConditionCreateContext(ctx->malloc);
    if (!ctx->client_cond) {
	opcItemFreeContext(ctx);
	return ERR_OPC_MEM;
    }
    
    err = ConditionInitContext(ctx->client_cond);
    if (err) {
	opcItemFreeContext(ctx);
	return err;
    }
    
    ctx->subitems = opcStringListCreate(ctx->malloc);
    if (!ctx->subitems) {
	opcItemFreeContext(ctx);
	return err;
    }

/*  This must be in over place and full destruction should occure only if haven't manual 
    sub(sub?)childs 
    
    ctx->pctx = opcPropertyGetInstance(sctx->pctx);
    if (!ctx->pctx) {
	opcItemFreeContext(ctx);
	return ERR_OPC_PROPERTIES_INSTANCE_INIT;
    }
*/	
    opcDebugServer(OPC_DEBUG_INFO, "opcItem(%p): InitContext Ok", ctx);
    
    return 0;
}

void opcItemFreeDataContext(OPCItemContext ctx) {
    assert(ctx);

    if (ctx->dm) opcDataManagerUnref(ctx->dm);
    opcItemZeroDataContext(ctx);
}

int opcItemFreeContext(OPCItemContext ctx) {
    int err = 0;
    
    opcDebugServer(OPC_DEBUG_NOTICE, "opcItem(%p): FreeContext %s", ctx, (ctx->fullname)?ctx->fullname:NULL);

    assert(ctx);

    
    if ((ctx->tree)&&(opcTreeGetStatus(ctx->tree)!=OPC_TREE_DESTROYING_FLAG)) {
	if ((ctx->tree)&&(ctx->status&OPC_ITEM_LINKED_FLAG)&&(!opcItemIsMarked(ctx))) {
	    err=opcTreeUnregisterItem(ctx->tree, ctx);
	    if (err==ERR_OPC_ITEM_HAVE_MANUAL_CHILDS) {
		opcItemResetManualMode(ctx);
	    } else {
		opcDebugServer(OPC_DEBUG_CRITICAL, "opcItem(%p): Unregister failed: %s", ctx, opcError(err, NULL));
		err = 0;
	    }
	}
    }

    if (ctx->clients) {
	opcDebugServer(OPC_DEBUG_DEBUG, "opcItem(%p): Waiting %lu clients to finish!", ctx, ctx->clients);
	
	ConditionPreLock(ctx->client_cond);
	while (ctx->clients)
	    ConditionWaitLocked(ctx->client_cond);
	ConditionUnLock(ctx->client_cond);
	
	opcDebugServer(OPC_DEBUG_DEBUG, "opcItem(%p): No more clients!", ctx);
    }

    ctx->status&=~OPC_ITEM_CONNECTED_FLAG;
    opcItemFreeDataContext(ctx);
	// ToDo: Destroy properties

    if (err==ERR_OPC_ITEM_HAVE_MANUAL_CHILDS) return err;


    if (ctx->subitems) opcStringListDestroy(ctx->subitems);

    // Unmark item
    if (ctx->client_cond) ConditionDestroyContext(ctx->client_cond);
    if (ctx->lock) LockDestroyContext(ctx->lock);

    if (ctx->fullname) opcExchangeString(ctx->malloc,&ctx->fullname,NULL);
    if (ctx->title) opcLCharExchangeString(ctx->malloc,&ctx->title,NULL);

    opcItemZeroContext(ctx);

    opcDebugServer(OPC_DEBUG_INFO, "opcItem(%p): FreeContext Ok", ctx);    
    return 0;
}

void opcItemDestroyContext(OPCItemContext ctx) {
    OPCMallocContext mymalloc;
    int err;

    opcDebugServer(OPC_DEBUG_NOTICE, "opcItem(%p): DestroyContext", ctx);    

    err=opcItemFreeContext(ctx);
    if (err!=ERR_OPC_ITEM_HAVE_MANUAL_CHILDS) {
	mymalloc = ctx->malloc;
	opcFree(mymalloc,ctx);
	if (mymalloc) opcMallocDestroyContext(mymalloc);
    }
}

int opcItemLock(OPCItemContext ctx) {
    assert(ctx);
    assert(ctx->lock);
    
    return LockWrite(ctx->lock);
}

int opcItemUnlock(OPCItemContext ctx) {
    assert(ctx);
    assert(ctx->lock);
    
    return LockReturn(ctx->lock);
}

int opcItemGet(OPCItemContext ctx) {
    assert(ctx);
    assert(ctx->lock);
    
    LockWrite(ctx->lock);
    ctx->clients++;
    LockReturn(ctx->lock);

    ConditionBroadcast(ctx->client_cond);

    return 0;
}

int opcItemReturn(OPCItemContext ctx) {
    assert(ctx);
    assert(ctx->lock);
    
    LockWrite(ctx->lock);
    ctx->clients--;
    LockReturn(ctx->lock);

    ConditionBroadcast(ctx->client_cond);

    return 0;
}


int opcItemSetAssociation(OPCItemContext ctx) {
    assert(ctx);
    ctx->status|=OPC_ITEM_CONNECTED_FLAG;
    return 0;
}

int opcItemSetManualMode(OPCItemContext ctx) {
    assert(ctx);
    assert(ctx->lock);
    
    LockWrite(ctx->lock);
    ctx->manual=1;
    LockReturn(ctx->lock);

    return 0;
}

int opcItemResetManualMode(OPCItemContext ctx) {
    assert(ctx);
    assert(ctx->lock);
    
    LockWrite(ctx->lock);
    ctx->manual=0;
    LockReturn(ctx->lock);

    return 0;
}

int opcItemSetVendorFilter(OPCItemContext ctx, OPCVendorFilter vendor_filter) {
    assert(ctx);
    
    LockWrite(ctx->lock);
    ctx->vendor_filter = vendor_filter;
    LockReturn(ctx->lock);
    return 0;
}

int opcItemSetTitle(OPCItemContext ctx, const char *title) {
    int err;
    assert(ctx);
    assert(title);
    LockWrite(ctx->lock);
    opcLCharExchangeString(ctx->malloc,&ctx->title,title);
    LockReturn(ctx->lock);

    return err;
}

const char *opcItemGetTitle(OPCItemContext ctx, const char *locale) {
    return (const char*)dsgettext(ctx->title,locale);
}

static char *opcItemGetFullName_internal(OPCItemContext ctx) {
    char *res = NULL;
    assert(ctx->fullname);
    opcExchangeString(ctx->malloc, &res, ctx->fullname);
    return res;
}

char *opcItemGetFullName(OPCItemContext ctx) {
    char *res;
    assert(ctx);
    assert(ctx->lock);

    LockRead(ctx->lock);
    res = opcItemGetFullName_internal(ctx);
    LockReturn(ctx->lock);
    
    return res;
}

int opcItemBrowse(OPCItemContext ctx, OPCItemBrowseFilter browse_filter, const char *name_filter, const char *vendor_filter, OPCStringList *result) {
    OPCItemContext ictx;
    OPCStringList sl, res;
    char *fullname;
    int err;

    assert(ctx);
    assert(result);
    assert(ctx->lock);
    
    *result = NULL;

    LockRead(ctx->lock);

    if ((ctx->subitems)&&(ctx->status&OPC_ITEM_LINKED_FLAG)) {
	res = opcStringListCreate(ctx->malloc?ctx->malloc->parrent:NULL);
	if (!res) err = ERR_OPC_MEM;
	else err = 0;
    } else  err = ERR_OPC_ITEM_NOT_LINKED_TO_TREE;
    
    if (!err) {
	for (sl=ctx->subitems;sl;sl=sl->next) {
	    if (sl->item) {
		ictx = *(OPCItemContext*)sl->item;
		if (ictx) {
		    LockRead(ictx->lock);
		    if (((!browse_filter)||(!opcItemBrowseFilter(ictx, browse_filter)))&&
			((!name_filter)||(!opcItemNameFilter(ictx, name_filter)))&&
			((!vendor_filter)||((ctx->vendor_filter)&&(!ctx->vendor_filter(ictx, vendor_filter))))) {
			    fullname = ictx->fullname;
			    if (fullname) opcStringListAdd(res, fullname, strlen(fullname) + 1);
		    }
		    LockReturn(ictx->lock);
		}
	    }
	}
    }
    
    LockReturn(ctx->lock);
    
    if (err) *result=NULL;
    else *result=res;
    
    return err;
}

/*******************************************************************************
**************************** Non-Locking Functions  ****************************
*******************************************************************************/

int opcItemHaveChilds(OPCItemContext ctx) {
    assert(ctx);
    
    if (ctx->status&OPC_ITEM_LINKED_FLAG) {
	assert(ctx->subitems);
	return (!opcStringListIsEmpty(ctx->subitems));
    } 

    return 0;
}

int opcItemIsConnected(OPCItemContext ctx) {
    assert(ctx);
    
    return ctx->status&OPC_ITEM_CONNECTED_FLAG;
}

/*******************************************************************************
*********************************** Filters ************************************
*******************************************************************************/
int opcItemBrowseFilter(OPCItemContext ctx, OPCItemBrowseFilter filter) {
    switch(filter) {
	case BRANCH:
	    if ((ctx->status&OPC_ITEM_CONNECTED_FLAG)&&(opcItemHaveChilds(ctx))) 
		return ERR_OPC_ITEM_FILTERED;
	case ITEM:
	    if (ctx->status&OPC_ITEM_CONNECTED_FLAG==0) 
		return ERR_OPC_ITEM_FILTERED;
    }
    return 0;
}


int opcItemNameFilter(OPCItemContext ctx, const char *filter) {
    return 0;
}


/*******************************************************************************
************************* Item SubItems Managment ******************************
*******************************************************************************/

int opcItemLink(OPCItemContext ctx, OPCItemContext ictx) {
    int err;
    assert(ctx);
    assert(ctx->subitems);
    assert(ictx);
    assert(ctx->lock);
    assert(ictx->lock);
    
    LockWrite(ctx->lock);
    err=opcStringListAdd(ctx->subitems, (void*)&ictx, sizeof(OPCItemContext*));
    LockReturn(ctx->lock);
    return err;
}

int opcItemUnlink(OPCItemContext ctx, OPCItemContext ictx) {
    assert(ctx);
    assert(ictx);
    assert(ctx->lock);
    assert(ictx->lock);

    LockWrite(ctx->lock);
    opcItemMarkParrentsCallback(ctx, ictx);
    opcTreeCleanMarked(ctx->tree);
    LockReturn(ctx->lock);
}
/*******************************************************************************
******************************* Tree Analysis **********************************
*******************************************************************************/

int opcItemHaveManualSubChilds(OPCItemContext ctx) {
    OPCItemContext ictx;
    OPCStringList sl;
    
    assert(ctx);
    assert(ctx->subitems);
    
    for (sl=ctx->subitems;sl;sl=sl->next) {
	ictx = *(OPCItemContext*)sl->item;
	if (ictx) {
	    if (ictx->manual) return 1;
	    if (opcItemHaveManualSubChilds(ictx)) return 1;
	}
    }
    return 0;
}

int opcItemSetTree(OPCItemContext ctx, OPCTreeContext tree) {
    assert(ctx);
    assert(tree);
    assert(ctx->tree==NULL);

    ctx->tree = tree;
}

/*******************************************************************************
************************* Deletion Marks Managment *****************************
*******************************************************************************/

void opcItemMark(OPCItemContext ctx) {
    assert(ctx);
    assert(ctx->lock);
    
    if (!ctx->mark) {
	LockWrite(ctx->lock);
	ctx->mark=1;
    }
}

void opcItemMarkSubTree(OPCItemContext ctx) {
    OPCStringList sl;
    OPCItemContext ictx;
    
    assert(ctx);
    
    opcItemMark(ctx);
    for (sl=ctx->subitems;sl;sl=sl->next) {
    	ictx = *(OPCItemContext*)sl->item;
	if (ictx) opcItemMarkSubTree(ictx);
    }
}

void opcItemUnMark(OPCItemContext ctx) {
    assert(ctx);
    assert(ctx->lock);
    
    if (ctx->mark) {
	ctx->mark=0;
	LockReturn(ctx->lock);
    }
}

void opcItemUnMarkSubTree(OPCItemContext ctx) {
    OPCStringList sl;
    OPCItemContext ictx;

    assert(ctx);
    
    opcItemUnMark(ctx);
    for (sl=ctx->subitems;sl;sl=sl->next) {
    	ictx = *(OPCItemContext*)sl->item;
	if (ictx) opcItemUnMarkSubTree(ictx);
    }
}

int opcItemIsMarked(OPCItemContext ctx) {
    assert(ctx);
    return ctx->mark;
}

/*******************************************************************************
************************* Tree managment callbacks *****************************
*******************************************************************************/


int opcItemMarkChildsCallback(OPCItemContext ctx, OPCItemContext ictx) {
    OPCItemContext ictx2;
    OPCStringList sl;
    
    if ((ctx==ictx)||((opcItemIsMarked(ctx))&&(!ctx->manual))) return 0;
    
    for (sl=ctx->subitems;sl;sl=sl->next) {
	ictx2 = *(OPCItemContext*)sl->item;
        if ((ictx2)&&(opcItemIsMarked(ictx2))) opcItemUnMarkSubTree(ictx2);
    }

    return 0;
}

int opcItemMarkParrentsCallback(OPCItemContext ctx, OPCItemContext ictx) {
    assert(ctx);
    assert(ctx->subitems);
    assert(ctx->tree);
    assert(ictx);

    if (opcStringListRemove(ctx->subitems, (void*)&ictx, sizeof(OPCItemContext*)))
	opcTreeMarkUnused(ctx->tree, ctx);
    return 0;
}

/*******************************************************************************
*************************** Properties and Data ********************************
*******************************************************************************/

OPCDataManagerContext opcItemRegisterDataManager(OPCItemContext ctx, OPCDataManagerContext dm) {
    assert(ctx);
    
    LockWrite(ctx->lock);
    if (ctx->dm) return NULL;
    ctx->dm = (OPCDataManagerContext)dm;
    LockReturn(ctx->lock);
    
    return dm;
}

OPCDataFastContext opcItemRegisterFastDataManager(OPCItemContext ctx, OPCType type, OPCDataFastOptions options, unsigned long flags) {
    int err;
    OPCDataFastContext dm;
    struct OPCDataFastOptionsT opts;
    OPCContext sctx = NULL;

    assert(ctx);
    assert(type);

    if ((!ctx)||(!type)) return NULL;
    
    if (!options) {
	opcDataFastInitDefaultOptions(&opts);
	options = &opts;
/*	options->flags = flags|OPC_DATA_FLAG_SEQUENTIAL; */
    }
    
    if (!options->formats) {
	if (!sctx) {
	    if (!ctx->tree) return NULL;
	    sctx = ctx->tree->sctx;
	    if (!sctx) return NULL;
	}
	if (!sctx->formats) return NULL;
	options->formats = sctx->formats;
    }
    if (!options->format) {
	if (!options->formats->native_format) return NULL;
	options->format = options->formats->native_format;
    }
    
    if (!options->chunk_pool) {
	if (!sctx) {
	    if (!ctx->tree) return NULL;
	    sctx = ctx->tree->sctx;
	    if (!sctx) return NULL;
	}
	if (!sctx->chunks) return NULL;
	options->chunk_pool = sctx->chunks;
    }
    if (!options->allocator.pool) {
	if (!sctx) {
	    if (!ctx->tree) return NULL;
	    sctx = ctx->tree->sctx;
	    if (!sctx) return NULL;
	}
	if (!sctx->buffers) return NULL;
	options->allocator.pool = sctx->buffers;
    }
    
    if (type->size) options->allocator.fixed_allocation_size = type->size;
    
    LockWrite(ctx->lock);
    if (ctx->dm) return NULL;
    
    dm = opcDataFastCreateContext(ctx->malloc);
    if (dm) {
	err = opcDataFastInitContext(dm, type, options);
	if (err) {
	    opcDataFastUnref(dm);
	    dm = NULL;
	} else {
	    ctx->dm = (OPCDataManagerContext)dm;
	    ctx->representation = OPC_REPRESENTATION_NATIVE;
	}
    }  else err = ERR_OPC_DATA_FAST_CREATE_CONTEXT;
    LockReturn(ctx->lock);
    
    if (err)
	opcDebugServer(OPC_DEBUG_WARNING, "opcItem(%p): RegisterFastDataManager: %s", ctx, opcError(err, NULL));

    return dm;
}

OPCDataContext opcItemRegisterLegacyDataManager(OPCItemContext ctx, OPCType type, OPCDataOptions options, unsigned long flags) {
    int err;
    OPCDataContext dm;
    struct OPCDataOptionsT opts;
    OPCContext sctx = NULL;

    assert(ctx);
    assert(type);

    if ((!ctx)||(!type)) return NULL;
    
    if (!options) {
	opcDataInitDefaultOptions(&opts);
	options = &opts;
	options->flags = flags;
    }
    
    if (!options->formats) {
	if (!sctx) {
	    if (!ctx->tree) return NULL;
	    sctx = ctx->tree->sctx;
	    if (!sctx) return NULL;
	}
	if (!sctx->formats) return NULL;
	options->formats = sctx->formats;
    }
    if ((!options->formats->native_format)||(!options->formats->opcxml_format)) return NULL;
    
    if (!options->chunk_pool) {
	if (!sctx) {
	    if (!ctx->tree) return NULL;
	    sctx = ctx->tree->sctx;
	    if (!sctx) return NULL;
	}
	if (!sctx->chunks) return NULL;
	options->chunk_pool = sctx->chunks;
    }
    
    if (!options->buffer_pool) {
	if (!sctx) {
	    if (!ctx->tree) return NULL;
	    sctx = ctx->tree->sctx;
	    if (!sctx) return NULL;
	}
	if (!sctx->buffers) return NULL;
	options->buffer_pool = sctx->buffers;
    }
    
    LockWrite(ctx->lock);
    if (ctx->dm) return NULL;
    
    dm = opcDataCreateContext(ctx->malloc);
    if (dm) {
	err = opcDataInitContext(dm, options);
	if (!err) err = opcDataRegisterNativeRepresentation(dm, type, NULL);
	if (!err) err = opcDataRegisterOPCXMLRepresentation(dm, type, NULL);

	if (err) {
	    opcDataUnref(dm);
	    dm = NULL;
	} else {
	    ctx->dm = (OPCDataManagerContext)dm;
	    ctx->representation = OPC_REPRESENTATION_OPCXML;
	}
    } else err = ERR_OPC_DATA_CREATE_CONTEXT;
    
    LockReturn(ctx->lock);

    if (err)
	opcDebugServer(OPC_DEBUG_WARNING, "opcItem(%p): RegisterLegacyDataManager: %s", ctx, opcError(err, NULL));
    
    return dm;
}


void opcItemRemoveDataManager(OPCItemContext ctx) {
    OPCDataManagerContext dm;
    assert(ctx);

    opcItemDisableDataManager(ctx);

    LockWrite(ctx->lock);    
    if (ctx->dm) {
	dm = ctx->dm;
	ctx->dm = NULL;
	opcDataManagerUnref(dm);
    }
    LockReturn(ctx->lock);
    
}

void opcItemEnableDataManager(OPCItemContext ctx) {
    LockWrite(ctx->lock);
    if (ctx->dm) ctx->status|=OPC_ITEM_CONNECTED_FLAG;
    LockReturn(ctx->lock);
}

void opcItemDisableDataManager(OPCItemContext ctx) {
    LockWrite(ctx->lock);
    ctx->status&=~OPC_ITEM_CONNECTED_FLAG;
    LockReturn(ctx->lock);
}


OPCDataManagerContext opcItemGetDataManager(OPCItemContext ctx) {
}

void opcItemReturnDataManager(OPCItemContext ctx) {
}



/*
OPCDataXMLContext opcItemRegisterXMLDataManager(OPCItemContext ctx,
*/

/*
*/
/*
int opcItemPushNativeData(OPCItemContext ctx, OPCTime time, OPCQuality quality, void *data) {
    int err;
    assert(ctx);

    if (time) {
	opcTimeCopy(&ctx->timestamp, time);
    } else {
	opcTimeCurrentTime(&ctx->timestamp);
	time = &ctx->timestamp;
    }
    
    err=opcDataPushNative(ctx->data, time, quality, data, 0);

    return err;
}

int opcItemGetOPCXMLData(OPCItemContext ctx, OPCDataInfo *info, OPCData *data) {
    assert(ctx);
    
    puts("Accquiring Data");
    // ToDo:
//    if (ctx->status&OPC_ITEM_CONNECTED_FLAG)
        return opcDataGetOPCXML(ctx->data, info, data);
    return ERR_OPC_ITEM_NOTCONNECTED;
}
*/
/*
int opcItemGetData(OPCItemContext ctx, unsigned short int format, OPCTime time, OPCQuality quality, void *data) {
}

int opcItemGetOPCXMLData(OPCItemContext ctx, OPCTime time, OPCQuality quality, void **data) {
    opcDataGet(ctx->data, time, quality, data);
}

int opcItemReturnOPCXMLData(OPCItemContext ctx, void *data) {
}

*/

int opcItemCheckConnection(OPCItemContext ctx) {
    if (!(ctx->status&OPC_ITEM_CONNECTED_FLAG)) return ERR_OPC_ITEM_NOTCONNECTED;
    assert(ctx->dm);

    return 0;
}

const OPCTime opcItemGetTimestamp(OPCItemContext ctx) {
    assert(ctx);
    
    return &ctx->timestamp;
}

int opcItemCheckAge(OPCItemContext ctx, const OPCTime current_time, const OPCDuration maxage) {
    struct OPCTimeT time;

    assert(ctx);
    assert(maxage);
    
//    if (!(ctx->status&OPC_ITEM_CONNECTED_FLAG)) return ERR_OPC_ITEM_NOTCONNECTED;
    
    return 0;
//    if (NOT_APPLICABLE) return 0;
    
    if (!current_time) opcTimeCurrentTime(&time);
    else opcTimeCopy(&time,current_time);
    opcTimeSub(&time,maxage);
    if (opcTimeCmp(&time, &ctx->timestamp)>0) {
//	if (RELOADING_APPLICABLE) ERR_OPC_ITEM_VALUE_OUTOFDATE_ACQUIRING;
	// Ask item to reload if possible otherwise
	
	return ERR_OPC_ITEM_VALUE_OUTOFDATE;
    }
    return 0;
}


int opcItemGetData(OPCItemContext ctx, OPCDataInfo *info, OPCRepresentationData *data) {
    int err;
    assert(ctx);
    
    LockRead(ctx->lock);
    if (ctx->status|OPC_ITEM_CONNECTED_FLAG) {
	opcDataManagerRef(ctx->dm);
	LockReturn(ctx->lock);
	
	err = opcDataManagerGetData(ctx->dm, ctx->representation, info, data);
	opcDataManagerUnref(ctx->dm);
	return err;
    }
    LockReturn(ctx->lock);
    
    return ERR_OPC_ITEM_NOTCONNECTED;
}

void opcItemReturnData(OPCItemContext ctx, OPCDataInfo info) {
    LockRead(ctx->lock);
    
    if (ctx->status|OPC_ITEM_CONNECTED_FLAG) {
	opcDataManagerRef(ctx->dm);
	LockReturn(ctx->lock);
	opcDataManagerReturnData(ctx->dm, info);
	opcDataManagerUnref(ctx->dm);
    }
    LockReturn(ctx->lock);
}

int opcItemPushData(OPCItemContext ctx, OPCDataInfo info, void *data, unsigned long size) {
    // ToDo: Implement
    return 0;
}

// ToDo: GetFirst, GetNext
