#include <stdio.h>
#include <assert.h>

#include <opcerr.h>
#include <opcconfig.h>
#include <opcdebug.h>

#include <opcmalloc.h>

#include <opc.h>
#include <opclocales.h>

static void opcLocalesZeroContext(OPCLocalesContext ctx) {
    assert(ctx);
    ctx->locales = NULL;
    ctx->lock = NULL;
}

OPCLocalesContext opcLocalesCreateContext(OPCMallocContext mctx) {
    OPCLocalesContext ctx;

    opcDebugServer(OPC_DEBUG_NOTICE, "opcLocalesCreateContext");    

    ctx = (OPCLocalesContext)opcMalloc(mctx, sizeof(struct OPCLocalesContextT));
    if (ctx) {
	ctx->malloc = mctx;
	opcLocalesZeroContext(ctx);
    }
    return ctx;
}

int opcLocalesInitContext(OPCLocalesContext ctx) {
    int err;

    opcDebugServer(OPC_DEBUG_NOTICE, "opcLocalesInitContext");    
    assert(ctx);

    opcLocalesZeroContext(ctx);
    ctx->lock = LockCreateContext(ctx->malloc);
    if (!ctx->lock) return ERR_OPC_MEM;
    err = LockInitContext(ctx->lock);
    if (err) {
	opcLocalesFreeContext(ctx);
	return err;
    }
    ctx->locales = HashTableInit(ctx->malloc,opc_supported_locales_number,HASH_TYPE_ARRAY,NULL);
    if (!ctx->locales) {
	opcLocalesFreeContext(ctx);
	return ERR_OPC_MEM;
    }
    err=HashTableAdd(ctx->locales,opc_default_locale, 0);
    if (err) {
	opcLocalesFreeContext(ctx);
	return err;
    }
    return 0;
}

void opcLocalesFreeContext(OPCLocalesContext ctx) {
    opcDebugServer(OPC_DEBUG_NOTICE, "opcLocalesFreeContext");    

    assert(ctx);
    if (ctx->locales) HashTableDestroy(ctx->locales);
    if (ctx->lock) LockDestroyContext(ctx->lock);
    opcLocalesZeroContext(ctx);
}

void opcLocalesDestroyContext(OPCLocalesContext ctx) {
    opcDebugServer(OPC_DEBUG_NOTICE, "opcLocalesDestroyContext");    

    assert(ctx);
    opcLocalesFreeContext(ctx);
    opcFree(ctx->malloc,ctx);
}


int opcLocalesAddLocale(OPCLocalesContext ctx, const char *newlocale) {
    int err;
    HashTable locales;
    LockContext lock;

    assert(ctx);
    locales = ctx->locales;
    assert(locales);
    lock = ctx->lock;
    assert(lock);
    
    LockWrite(lock);
    err=HashTableAdd(ctx->locales,(void*)newlocale, 0);
    LockReturn(lock);
    return err;
}

int opcLocalesAddLocales(OPCLocalesContext ctx, const char **newlocales) {
    unsigned long i;
    int err = 0;
    
    assert(ctx);
    
    if (newlocales) {
	for (i=0;newlocales[i];i++) {
	    if (!err) opcLocalesAddLocale(ctx,newlocales[i]);
	}
    }
    return err;
}

OPCLocale opcLocalesGetLocale(OPCLocalesContext ctx, char *attr) {
    HashTable locales;
    OPCLocale lctx;
    char *locale = NULL;
    unsigned long i;
    
    assert(ctx);
    locales = ctx->locales;
    assert(locales);
    
    lctx = opcLocaleCreate();
    if (!lctx) return NULL;

    if (attr) {
	i=strlen(attr);
	if (i>5) attr[5]=0;
	if ((i>2)&&(attr[2]=='-')) attr[2]='_';
	locale=(char*)HashTableFind(locales,attr);
	if ((!locale)&&(i>2)) {
	    attr[2]=0;
	    locale=(char*)HashTableFind(locales,attr);
	}
    }

    if (locale) {
	opcLocaleSetLocale(lctx,locale);
	if ((locale)&&(i>5)) opcLocaleSetCharset(lctx,attr+6);
	else opcLocaleSetCharset(lctx,opc_default_charset);
    } else {
	opcLocaleSetLocale(lctx,opc_default_locale);
	opcLocaleSetCharset(lctx,opc_default_charset);
    }
    return lctx;
}

int opcLocalesList(OPCLocalesContext ctx, OPCLocalesListFunctionT list_function, void *attr) {
    int err;
    LockContext lock;
    assert(ctx);
    assert(list_function);
    lock=ctx->lock;
    assert(lock);
    
    LockRead(lock);        
    err=HashTableWalk(ctx->locales, (HashTableWalkT)list_function, attr);
    LockReturn(lock);
}
