#define _OPC_PROPERTIES_C
#include <opcassert.h>

#include <opcerr.h>
#include <opcdebug.h>

#include <opcproperty.h>
#include <opcproperties.h>

void opcPropertiesZeroContext(OPCPropertiesContext ctx) {
	ctx->id=NULL;
	ctx->lock=NULL;
	ctx->properties=NULL;
}

OPCPropertiesContext opcPropertiesCreateContext(OPCMallocContext mctx) {
    OPCPropertiesContext ctx;

    opcDebugServer(OPC_DEBUG_NOTICE, "opcPropertiesCreateContext");    

    ctx = (OPCPropertiesContext)opcMalloc(mctx,sizeof(struct OPCPropertiesContextT));
    if (ctx) {
	ctx->malloc=mctx;
	opcPropertiesZeroContext(ctx);
    }
    return ctx;
}

int opcPropertiesInitContext(OPCPropertiesContext ctx, unsigned long properties_hash_size, unsigned short properties_maxid, unsigned short properties_userid) {
    unsigned short i;
    int err;

    opcDebugServer(OPC_DEBUG_NOTICE, "opcPropertiesInitContext");    

    assert(ctx);

    opcPropertiesZeroContext(ctx);

    ctx->id=HashTableInit(NULL,properties_hash_size,HASH_TYPE_POINTER,NULL);
    if (!ctx->id) {
	opcPropertiesFreeContext(ctx);
	return ERR_OPC_PROPERTIES_HASHTABLE_INIT;
    }

    ctx->lock = LockCreateContext(ctx->malloc);
    if (!ctx->lock) {
	opcPropertiesFreeContext(ctx);
	return ERR_OPC_LOCK_CREATE;
    }
    err = LockInitContext(ctx->lock);
    if (err) {
	opcPropertiesFreeContext(ctx);
	return err;
    }

    ctx->properties = (OPCPropertyContext*)opcMalloc(ctx->malloc, properties_maxid*sizeof(OPCPropertyContext));
    if (!ctx->properties) {
	opcPropertiesFreeContext(ctx);
	return ERR_OPC_MEM;
    }
    
    for (i=0;i<properties_maxid;i++)
	ctx->properties[i]=NULL;
        
    ctx->maxid = properties_maxid;
    ctx->userid = properties_userid;
    ctx->curid = properties_userid;
        
    return opcPropertiesRegisterStandard(ctx);
}

void opcPropertiesFreeContext(OPCPropertiesContext ctx) {
    opcDebugServer(OPC_DEBUG_NOTICE, "opcPropertiesFreeContext");    

    assert(ctx);

    if (ctx->properties) opcFree(ctx->malloc, ctx->properties);
    if (ctx->id) HashTableDestroy(ctx->id);
    if (ctx->lock) LockDestroyContext(ctx->lock);

    opcPropertiesZeroContext(ctx);
}

void opcPropertiesDestroyContext(OPCPropertiesContext ctx) {
    opcDebugServer(OPC_DEBUG_NOTICE, "opcPropertiesDestroyContext");    

    opcPropertiesFreeContext(ctx);
    opcFree(ctx->malloc,ctx);
}



int opcPropertiesRegisterProperty(OPCPropertiesContext ctx, OPCPropertyContext property) {
    int err=0;
    char *name;
    unsigned short id;
    
    assert(ctx);
    assert(property);
    
    name = property->name;
    if (!name) return ERR_OPC_PROPERTIES_INVALID_PROPERTY;
    if (opcPropertiesFindProperty(ctx,name)) return ERR_OPC_PROPERTIES_ALREDY_REGISTERED;
    if (property->id) {
	id = property->id;
	if (ctx->properties[id]) return ERR_OPC_PROPERTIES_ID_ALREDY_USED;
    } else if (ctx->curid<ctx->maxid) id=0;
    else return ERR_OPC_PROPERTIES_LIMIT_EXCEEDED;


    LockWrite(ctx->lock);
    if (HashTableInsert(ctx->id, property)) err=ERR_OPC_PROPERTIES_HASHTABLE_INSERT_FAILED;
    if (!err) {
        if (!id) {
	    id=ctx->curid++;
	    opcPropertySetId(property, id);
	}
        ctx->properties[id]=property;
    }
    LockReturn(ctx->lock);
    
    return err;
}


int opcPropertiesRegisterPropertyByType(OPCPropertiesContext ctx, const char *name, const OPCType type) {
    int err;
    OPCPropertyContext property;
    
    assert(ctx);
    assert(name);
    assert(type);
    
    if (opcPropertiesFindProperty(ctx,name)) return ERR_OPC_PROPERTIES_ALREDY_REGISTERED;
    
    property = opcPropertyCreateContext(ctx->malloc);
    if (!property) return ERR_OPC_PROPERTY_CREATE;
    err = opcPropertyInitContextByType(property, name, type);
    if (err) {
	opcPropertyDestroyContext(property);
	return err;
    }

    err = opcPropertiesRegisterProperty(ctx, property);
    if (err) opcPropertyDestroyContext(property);
    return err;
}

int opcPropertiesRegisterPropertyByDataType(OPCPropertiesContext ctx, const char *name, OPCDataType type, const void *attr) {
    int err;
    OPCPropertyContext property;
    
    assert(ctx);
    assert(name);
    assert(type);
    
    if (opcPropertiesFindProperty(ctx,name)) return ERR_OPC_PROPERTIES_ALREDY_REGISTERED;
    
    property = opcPropertyCreateContext(ctx->malloc);
    if (!property) return ERR_OPC_PROPERTY_CREATE;
    err = opcPropertyInitContext(property, name, type, attr);
    if (err) {
	opcPropertyDestroyContext(property);
	return err;
    }

    err = opcPropertiesRegisterProperty(ctx, property);
    if (err) opcPropertyDestroyContext(property);
    return err;
}

OPCPropertyContext opcPropertiesFindProperty(OPCPropertiesContext ctx, const char *property) {
    OPCPropertyContext pctx;
    
    assert(ctx);
    
    LockRead(ctx->lock);
    pctx=(OPCPropertyContext)HashTableFind(ctx->id, property);
    LockReturn(ctx->lock);
    
    return pctx;
}

unsigned short opcPropertiesFindPropertyId(OPCPropertiesContext ctx, const char *property) {
    OPCPropertyContext pctx;
    assert(ctx);

    LockRead(ctx->lock);
    pctx=(OPCPropertyContext)HashTableFind(ctx->id, property);
    LockReturn(ctx->lock);
    
    if (pctx) return pctx->id;
    return (unsigned short)-1;
}

OPCPropertyContext OpcPropertiesGetProperty(OPCPropertiesContext ctx, unsigned short id) {
    assert(ctx);
    assert(ctx->properties);
    
    return ctx->properties[id];
}
