#include <stdio.h>
#include <errno.h>
#include <assert.h>

#include <opcerr.h>
#include <opcconfig.h>
#include <opcdebug.h>

#include <opcmalloc.h>
#include <lchar.h>
#include <hashtable.h>
#include <lock.h>

#include <opcitem.h>

#include <opctree.h>

void opcTreeZeroContext(OPCTreeContext ctx) {
    assert(ctx);
    ctx->status = 0;    
    ctx->item = NULL;
    ctx->lock = NULL;
}

OPCTreeContext opcTreeCreateContext(OPCMallocContext mctx) {
    OPCTreeContext ctx;

    opcDebugServer(OPC_DEBUG_NOTICE, "opcTreeCreateContext");    

    ctx = (OPCTreeContext)opcMalloc(mctx, sizeof(struct OPCTreeContextT));
    if (ctx) {
	ctx->malloc = mctx;
	opcTreeZeroContext(ctx);
    }
    return ctx;
}

int opcTreeInitContext(OPCTreeContext ctx, OPCContext sctx) {
    int err;
    OPCItemContext ictx;
    
    opcDebugServer(OPC_DEBUG_NOTICE, "opcTreeInitContext");    

    assert(ctx);

    opcTreeZeroContext(ctx);

    ctx->item=HashTableInit(ctx->malloc, opc_max_items, HASH_TYPE_POINTER, NULL);
    HashTableSetDestructor(ctx->item,(void*)&opcItemDestroyContext);
    if (!ctx->item) return ERR_OPC_MEM;
    
    ctx->lock=LockCreateContext(ctx->malloc);
    if (!ctx->lock) {
	opcTreeFreeContext(ctx);
	return ERR_OPC_MEM;
    }
    
    err = LockInitContext(ctx->lock);
    if (err) {
	opcTreeFreeContext(ctx);
	return err;
    }
    
    ictx = opcItemCreateContext(ctx->malloc);
    if (!ictx) {
	opcTreeFreeContext(ctx);
	return ERR_OPC_MEM;
    }
    
    err = opcItemInitContext(ictx,"","");
    
    if (!err) err = opcTreeRegisterItem(ctx,ictx,NULL);
    
    if (err) {
	opcTreeFreeContext(ctx);
	return err;
    }
    
    opcItemSetManualMode(ictx);

    ctx->sctx = sctx;

    opcDebugServer(OPC_DEBUG_INFO, "opcTreeInitContext Ok");    

    return 0;
}

void opcTreeFreeContext(OPCTreeContext ctx) {
    opcDebugServer(OPC_DEBUG_NOTICE, "opcTreeFreeContext");    

    assert(ctx);
    
    ctx->status|=OPC_TREE_DESTROYING_FLAG;
    
    if (ctx->item) {
	/* I don't need to lock here because all threads before this point should be terminated */
	HashTableDestroy(ctx->item);
    }
    if (ctx->lock) LockDestroyContext(ctx->lock);
    
    opcTreeZeroContext(ctx);

    opcDebugServer(OPC_DEBUG_INFO, "opcTreeFreeContext Ok");    
}

void opcTreeDestroyContext(OPCTreeContext ctx) {
    opcDebugServer(OPC_DEBUG_NOTICE, "opcTreeDestroyContext");    

    opcTreeFreeContext(ctx);
    opcFree(ctx->malloc,ctx);
}

int opcTreeGetStatus(OPCTreeContext ctx) {
    assert(ctx);
    return ctx->status;
}

OPCItemContext opcTreeFindItem(OPCTreeContext ctx, const char *fullname) {
    OPCItemContext ictx;

    assert(ctx);
    assert(fullname);
    assert(ctx->lock);
    assert(ctx->item);
    
    LockRead(ctx->lock);
    ictx = HashTableFind(ctx->item,fullname);
    LockReturn(ctx->lock);

    return ictx;
}

OPCItemContext opcTreeGetItem(OPCTreeContext ctx, const char *fullname) {
    int err;
    OPCItemContext ictx;

    assert(ctx);
    assert(fullname);
    assert(ctx->lock);
    assert(ctx->item);
    
    LockRead(ctx->lock);
    ictx = HashTableFind(ctx->item,fullname);
    if (ictx) err = opcItemGet(ictx);
    else err = 0;
    LockReturn(ctx->lock);

    if (err) return NULL;
    return ictx;
}

int opcTreeWalk(OPCTreeContext ctx, OPCTreeWalkT func, void *attr) {
    int err;
    assert(ctx);
    assert(ctx->item);
    assert(ctx->lock);
    
    LockRead(ctx->lock);
    err=HashTableWalk(ctx->item,(HashTableWalkT)func,attr);
    LockReturn(ctx->lock);
    return err;
}

int opcTreeLock(OPCTreeContext ctx) {
    assert(ctx);
    assert(ctx->lock);
    
    return LockWrite(ctx->lock);
}

int opcTreeUnlock(OPCTreeContext ctx) {
    assert(ctx);
    assert(ctx->lock);
    
    return LockReturn(ctx->lock);
}

int opcTreeLinkItems(OPCTreeContext ctx, OPCItemContext ictx, OPCItemContext child) {
    int err;

    opcDebugServer(OPC_DEBUG_NOTICE, "opcTreeLinkItems %p,%p.", ictx, child);    
    
    assert(ctx);
    assert(ctx->lock);
    assert(ictx);
    assert(child);
    
    LockWrite(ctx->lock);
    err = opcItemLink(ictx, child);
    LockReturn(ctx->lock);

    opcDebugServer(OPC_DEBUG_INFO, "opcTreeLinkItems %p,%p. Ok", ictx, child);    

    return err;
}

int opcTreeUnlinkItems(OPCTreeContext ctx, OPCItemContext ictx, OPCItemContext child) {
    int err;

    opcDebugServer(OPC_DEBUG_NOTICE, "opcTreeUnlinkItems %p,%p.", ictx, child);    

    assert(ctx);
    assert(ctx->lock);
    assert(ictx);
    assert(child);

    LockWrite(ctx->lock);
    err = opcItemUnlink(ictx, child);        
    LockWrite(ctx->lock);

    opcDebugServer(OPC_DEBUG_INFO, "opcTreeUnlinkItems %p,%p. Ok", ictx, child);    
    return err;
}

OPCItemContext opcTreeCreatePath(OPCTreeContext ctx, const char *path) {
    OPCItemContext ictx;
    char *ch, *ppath;
    int err;

    opcDebugServer(OPC_DEBUG_NOTICE, "opcTreeCreatePath \"%s\"", path);    
    
    assert(ctx);
    assert(path);
    
    ictx = opcTreeFindItem(ctx,path);
    if (ictx) return ictx;
    
    ictx=opcItemCreateContext(ctx->malloc);
    if (!ictx) return NULL;
	
    ch=strrchr(path,'/');

    if (ch) {
        err = opcItemInitContext(ictx,path,ch+1); 
	if (!err) {
	    err = ch-path;
	    ppath=(char*)malloc((err+1)*sizeof(char));
	    if (ppath) {
		if (err) strncpy(ppath,path,err); 
		ppath[err]=0;
		err=opcTreeRegisterItem(ctx,ictx,ppath);
		free(ppath);
	    } else err=ERR_OPC_MEM;
	}
    } else {
	err = opcItemInitContext(ictx,"","");
	if (!err) err=opcTreeRegisterItem(ctx,ictx,"");
    }
    
    if (err) {
	opcItemDestroyContext(ictx);
	if (err!=ERR_OPC_ITEM_ALREDY_REGISTERED) return NULL;
	ictx = opcTreeFindItem(ctx,path);
    }
    
    opcDebugServer(OPC_DEBUG_INFO, "opcTreeCreatePath \"%s\" Ok", path);    

    return ictx;
}

int opcTreeRegisterItem(OPCTreeContext ctx, OPCItemContext ictx, const char *path) {
    OPCItemContext parrent;
    int err;
    char *ch;

    if (path)
	opcDebugServer(OPC_DEBUG_NOTICE, "opcTreeRegisterItem %p under \"%s\"", ictx, path);    
    else 
	opcDebugServer(OPC_DEBUG_NOTICE, "opcTreeRegisterItem %p", ictx);    

    assert(ctx);
    assert(ictx);
    assert(ctx->lock);
    assert(ictx->fullname);

    opcItemLock(ictx);
    
    LockWrite(ctx->lock);
    if (HashTableFind(ctx->item,ictx->fullname)) return ERR_OPC_ITEM_ALREDY_REGISTERED;
    if (HashTableInsert(ctx->item, (void*)ictx)) return ERR_OPC_MEM;
    opcItemSetTree(ictx,ctx);
    LockReturn(ctx->lock);
    
    if (path) {
	parrent = opcTreeCreatePath(ctx,path);
	if (parrent) {
	    err = opcTreeLinkItems(ctx, parrent, ictx);
	    if (parrent->manual==2) parrent->manual = 0;
	} else err = ERR_OPC_ITEM_CREATE_PATH;
    } else err = 0;

    if (err) {
	LockWrite(ctx->lock);
	if (opcTreeMarkUnused(ctx,parrent)!=ERR_OPC_NOTFOUND) opcTreeCleanMarked(ctx);
	LockReturn(ctx->lock);
    } else ictx->status |= OPC_ITEM_LINKED_FLAG;
    
    opcItemUnlock(ictx);

    if (path)
	opcDebugServer(OPC_DEBUG_INFO, "opcTreeRegisterItem %p under \"%s\". Linked to %p. Ok", ictx, path, parrent);    
    else
	opcDebugServer(OPC_DEBUG_INFO, "opcTreeRegisterItem %p Ok", ictx);    

    return err;
}

int opcTreeUnregisterItem(OPCTreeContext ctx, OPCItemContext ictx) {
    opcDebugServer(OPC_DEBUG_NOTICE, "opcTreeUnregisterItem %p", ictx);    
    
    assert(ctx);
    assert(ictx);
    assert(ctx->item);
    
    LockWrite(ctx->lock);
    
    if (opcItemHaveManualSubChilds(ictx)) return ERR_OPC_ITEM_HAVE_MANUAL_CHILDS;
    
    opcItemMarkSubTree(ictx);
    HashTableWalk(ctx->item, (HashTableWalkT)opcItemMarkParrentsCallback, (void*)ictx);
    HashTableWalk(ctx->item, (HashTableWalkT)opcItemMarkChildsCallback, (void*)ictx);
    ictx->mark = 0;
    opcTreeCleanMarked(ctx);
    HashTableUnlink(ctx->item,ictx);
    ictx->status &= ~OPC_ITEM_LINKED_FLAG;    

    opcItemUnlock(ictx);
    LockReturn(ctx->lock);

    opcDebugServer(OPC_DEBUG_INFO, "opcTreeUnregisterItem %p Ok", ictx);
    return 0;
}

int opcTreeBrowse(OPCTreeContext ctx, const char *path, OPCItemBrowseFilter browse_filter, const char *name_filter, const char *vendor_filter, OPCStringList *result) {
    int err;
    OPCItemContext ictx;
    
    assert(ctx);
    assert(path);
    assert(result);
    assert(ctx->lock);
    
    LockRead(ctx->lock);
    ictx = HashTableFind(ctx->item,path);
    if (ictx) err = opcItemBrowse(ictx,browse_filter,name_filter,vendor_filter,result);
    else {
	*result = NULL;
	err = ERR_OPC_ITEM_NOTFOUND;
    }
    LockReturn(ctx->lock);
    
    return err;
}

/*******************************************************************************
**************************** Internal Functions ********************************
*******************************************************************************/

int opcTreeMarkUnused(OPCTreeContext tree, OPCItemContext ctx) {
    assert(ctx);
    assert(tree);
    assert(tree->item);
    
    if ((opcItemHaveChilds(ctx))||(ctx->manual)||(ctx->status&OPC_ITEM_CONNECTED_FLAG)) return ERR_OPC_NOTFOUND;
    opcItemMark(ctx);
    return HashTableWalk(tree->item, (HashTableWalkT)opcItemMarkParrentsCallback, (void*)ctx);
}

int opcTreeCleanMarked(OPCTreeContext tree) {
    assert(tree);
    assert(tree->item);
    return HashTableClean(tree->item, (HashTableCleanT)opcItemIsMarked);
}

