#include <stdio.h>
#include <assert.h>

#include <libxml/schemasInternals.h>

#include <opcerr.h>
#include <opcdebug.h>

#include <opcfdopcbinarydictionary.h>
#include <opcfdopcbinary.h>

void opcFDOPCBinaryZeroContext(OPCFDOPCBinaryContext ctx) {
    assert(ctx);
    
    ctx->lock = NULL;
    ctx->dictdir = NULL;
    ctx->schema_parser = NULL;
    ctx->schema_rep = NULL;
    ctx->schema_ctx = NULL;
    
}

OPCFDOPCBinaryContext opcFDOPCBinaryCreateContext(OPCMallocContext mctx) {
    OPCFDOPCBinaryContext ctx;

    opcDebugServer(OPC_DEBUG_NOTICE, "opcFormatDescriptionOPCBinaryCreateContext");

    ctx = (OPCFDOPCBinaryContext)opcMalloc(mctx,sizeof(struct OPCFDOPCBinaryContextT));
    if (ctx) {
	ctx->malloc=mctx;
	opcFDOPCBinaryZeroContext(ctx);
    }
    return ctx;
}

int opcFDOPCBinaryInitContext(OPCFDOPCBinaryContext ctx, OPCNamespaces namespaces, const char *schema, const char *dictdir) {
    int err;
    assert(ctx);
    assert(namespaces);
    
    opcDebugServer(OPC_DEBUG_NOTICE, "opcFormatDescriptionOPCBinaryInitContext: %s", schema);

    opcFDOPCBinaryZeroContext(ctx);

    if (opcExchangeString(ctx->malloc, &ctx->dictdir, dictdir)) return ERR_OPC_MEM;
    ctx->dictdir_chars = strlen(ctx->dictdir);

    ctx->schema_parser=xmlSchemaNewParserCtxt(schema);
    if (!ctx->schema_parser) {
	opcFDOPCBinaryFreeContext(ctx);
	return ERR_OPC_FORMAT_DESCRIPTION_OPCBINARY_SCHEMA;
    }
    ctx->schema_rep=xmlSchemaParse(ctx->schema_parser);
    if (!ctx->schema_rep) {
	opcFDOPCBinaryFreeContext(ctx);
	return ERR_OPC_FORMAT_DESCRIPTION_OPCBINARY_SCHEMA;
    }
    ctx->schema_ctx=xmlSchemaNewValidCtxt(ctx->schema_rep);
    if (!ctx->schema_ctx) {
	opcFDOPCBinaryFreeContext(ctx);
	return ERR_OPC_FORMAT_DESCRIPTION_OPCBINARY_SCHEMA;
    }
    
    ctx->lock = LockCreateContext(ctx->malloc);
    if (!ctx->lock) {
	opcFDOPCBinaryFreeContext(ctx);
	return ERR_OPC_MEM;
    }
    err = LockInitContext(ctx->lock);
    if (err) {
	opcFDOPCBinaryFreeContext(ctx);
	return err;
    }

    err=opcNamespacesSetOPCBinaryNamespace(namespaces, ctx->schema_rep->targetNamespace);
    if (err) {
	opcFDOPCBinaryFreeContext(ctx);
	return err;
    }
    ctx->namespaces = namespaces;

    xmlSchemaSetValidErrors(ctx->schema_ctx,NULL,NULL,NULL);

    return 0;        
}

void opcFDOPCBinaryFreeContext(OPCFDOPCBinaryContext ctx) {
    assert(ctx);

    opcDebugServer(OPC_DEBUG_NOTICE, "opcFormatDescriptionOPCBinaryFreeContext");

    if (ctx->lock) LockDestroyContext(ctx->lock);
    if (ctx->dictdir) opcExchangeString(ctx->malloc, &ctx->dictdir, NULL);
    if (ctx->schema_ctx) xmlSchemaFreeValidCtxt(ctx->schema_ctx);
    if (ctx->schema_rep) xmlSchemaFree(ctx->schema_rep);
    if (ctx->schema_parser) xmlSchemaFreeParserCtxt(ctx->schema_parser);
    
    opcFDOPCBinaryZeroContext(ctx);
}

void opcFDOPCBinaryDestroyContext(OPCFDOPCBinaryContext ctx) {
    opcDebugServer(OPC_DEBUG_NOTICE, "opcFormatDescriptionOPCBinaryDestroyContext");
    opcFDOPCBinaryFreeContext(ctx);
    opcFree(ctx->malloc,ctx);
}

int opcFDOPCBinaryValidateXML(OPCFDOPCBinaryContext ctx, xmlDocPtr doc) {
    return 0;
}

int opcFDOPCBinaryValidateDictionary(OPCFDOPCBinaryContext ctx, OPCFDOPCBinaryDictionaryContext dctx) {
    LockWrite(ctx->lock);
    LockReturn(ctx->lock);
    return 0;
}

OPCFDOPCBinaryDictionaryContext opcFDOPCBinaryCreateDictionaryFromMemory(OPCFDOPCBinaryContext ctx, const char *xml, unsigned long length) {
    int err;
    xmlDocPtr doc;
    OPCFDOPCBinaryDictionaryContext dctx;
    
    assert(ctx);
    assert(xml);
    assert(length>0);
    assert(ctx->schema_ctx);

    if (!length) length=xmlStrlen(xml);
    doc = xmlReadMemory(xml,length,"xml",NULL,0);
    if (!doc) return NULL;
    
    LockWrite(ctx->lock);
    if (xmlSchemaValidateDoc(ctx->schema_ctx,doc)) {
	LockReturn(ctx->lock);
	xmlFreeDoc(doc);
	return NULL;
    }
    LockReturn(ctx->lock);
    
    dctx=opcFDOPCBinaryDictionaryCreateContext(ctx->malloc);
    if (!dctx) return dctx;
    
    err = opcFDOPCBinaryDictionaryInitContext(dctx,ctx->namespaces,doc);
    if (err) {
	xmlFreeDoc(doc);
	opcFDOPCBinaryDictionaryDestroyContext(dctx);
	return NULL;
    }

    return dctx;    
}

OPCFDOPCBinaryDictionaryContext opcFDOPCBinaryCreateDictionary(OPCFDOPCBinaryContext ctx, const char *dictionary) {
    int err;
    char *stmp;
    xmlDocPtr doc;
    OPCFDOPCBinaryDictionaryContext dctx;
    
    assert(ctx);
    assert(dictionary);
    assert(ctx->schema_ctx);

    opcDebugServer(OPC_DEBUG_NOTICE, "opcFormatDescriptionOPCBinaryCreateDictionary: %s", dictionary);

    if (dictionary[0]=='/') doc = xmlReadFile(dictionary,NULL,0);
    else {
	stmp = (char*)malloc(ctx->dictdir_chars+strlen(dictionary));
	if (!stmp) return NULL;
	sprintf(stmp,"%s/%s",ctx->dictdir,dictionary);
	doc = xmlReadFile(stmp,NULL,0);
	free(stmp);
    }
    if (!doc) return NULL;
    
    LockWrite(ctx->lock);
    if (xmlSchemaValidateDoc(ctx->schema_ctx,doc)) {
	LockReturn(ctx->lock);
	xmlFreeDoc(doc);
	return NULL;
    }
    LockReturn(ctx->lock);
    
    dctx=opcFDOPCBinaryDictionaryCreateContext(ctx->malloc);
    if (!dctx) return dctx;
    
    err = opcFDOPCBinaryDictionaryInitContext(dctx,ctx->namespaces,doc);
    if (err) {
	xmlFreeDoc(doc);
	opcFDOPCBinaryDictionaryDestroyContext(dctx);
	return NULL;
    }

    opcDebugServer(OPC_DEBUG_NOTICE, "opcFormatDescriptionOPCBinaryCreateDictionary: Ok. Created %p", dctx);

    return dctx;    
}
