#include <opcassert.h>

#include <opcerr.h>
#include <opcdebug.h>
#include <opcchunkpool.h>

static inline void opcChunkPoolZeroContext(OPCChunkPoolContext ctx) {
    assert(ctx);
    ctx->first_chunk = NULL;
    ctx->last_chunk = NULL;
    ctx->lock = NULL;
}

OPCChunkPoolContext opcChunkPoolCreateContext(OPCMallocContext mctx) {
    OPCChunkPoolContext ctx;

    opcDebugServer(OPC_DEBUG_NOTICE, "opcChunkPoolCreateContext");    

    ctx = (OPCChunkPoolContext)opcMalloc(mctx,sizeof(struct OPCChunkPoolContextT));
    if (ctx) {
	ctx->malloc=mctx;
	opcChunkPoolZeroContext(ctx);
    }
    return ctx;
}

int opcChunkPoolInitContext(OPCChunkPoolContext ctx, unsigned char bins) {
    int err;
    unsigned char i;
    OPCDataChunkContext *ptr1, *ptr2;

    opcDebugServer(OPC_DEBUG_NOTICE, "opcChunkPoolInitContext");    

    assert(ctx);
    assert(bins);

    ptr1 = (OPCDataChunkContext*)opcMalloc(ctx->malloc, bins*sizeof(OPCDataChunkContext));
    ptr2 = (OPCDataChunkContext*)opcMalloc(ctx->malloc, bins*sizeof(OPCDataChunkContext));
    if ((!ptr1)||(!ptr2)) {
	if (ptr1) opcFree(ctx->malloc, ptr1);
	if (ptr2) opcFree(ctx->malloc, ptr2);
	return ERR_OPC_MEM;
    }
    
    for (i=0;i<bins;i++) {
	ptr1[i]=NULL;
	ptr2[i]=NULL;
    }

    ctx->lock = LockCreateContext(ctx->malloc);
    if (!ctx->lock) {
	opcChunkPoolFreeContext(ctx);
	return ERR_OPC_MEM;
    }
    err = LockInitContext(ctx->lock);
    if (err) {
	opcChunkPoolFreeContext(ctx);
	return err;
    }

    ctx->bins = bins;
    ctx->first_chunk = ptr1;
    ctx->last_chunk = ptr2;
    
    return 0;
}

void opcChunkPoolFreeContext(OPCChunkPoolContext ctx) {
    unsigned char i;
    OPCDataChunkContext ptr1, ptr2;

    opcDebugServer(OPC_DEBUG_NOTICE, "opcChunkPoolFreeContext");    

    assert(ctx);
    
    if (ctx->lock) LockDestroyContext(ctx->lock);
    
    if (ctx->first_chunk) {
	for (i=0;i<ctx->bins;i++) {
	    for (ptr1=ctx->first_chunk[i];ptr1;ptr1=ptr2) {
		ptr2 = ptr1->next;
		opcDataChunkDestroyContext(ptr1);
	    }
	}
	opcFree(ctx->malloc, ctx->first_chunk);
    }
    if (ctx->last_chunk) opcFree(ctx->malloc, ctx->last_chunk);

    opcChunkPoolZeroContext(ctx);    
}

void opcChunkPoolDestroyContext(OPCChunkPoolContext ctx) {
    opcDebugServer(OPC_DEBUG_NOTICE, "opcChunkPoolDestroyContext");    
    opcChunkPoolFreeContext(ctx);
    opcFree(ctx->malloc,ctx);
}

OPCDataChunkContext opcChunkPoolGetChunk(OPCChunkPoolContext ctx, OPCRepresentationInfo *definfo, unsigned short int formats) {
    int err;
    OPCDataChunkContext chunk;
    unsigned short int i;
    unsigned char j;
    
    assert(ctx);
    assert(formats);

    for (j=0,i=(formats>>1);i;i>>=1) j++;
    if (formats != (1<<j)) j++;
    i = (1<<j);
    
    if (j>=ctx->bins) {
	opcDebugServer(OPC_DEBUG_WARNING, "opcChunkPoolGetChunk: You should increase number of bins.");
	j=ctx->bins-1;
    }

    
    LockWrite(ctx->lock);
    
    chunk = ctx->first_chunk[j];
    if (chunk) {
	if (chunk->next) {
	    ctx->first_chunk[j]=chunk->next;
	} else {
	    ctx->first_chunk[j]=NULL;
	    ctx->last_chunk[j]=NULL;
	}
	opcDataChunkRestoreContext(chunk, definfo);
    } else {
	chunk = opcDataChunkCreateContext(ctx->malloc);
	if (chunk) {
	    err = opcDataChunkInitContext(chunk, definfo, i);
	    if (err) opcDataChunkDestroyContext(chunk);
	}
    }
    LockReturn(ctx->lock);

    return chunk;
}

void opcChunkPoolReturnChunk(OPCChunkPoolContext ctx, OPCDataChunkContext chunk) {
    unsigned short int i;
    unsigned char j;

    assert(ctx);
    assert(chunk);

    opcDataChunkFreeContext(chunk);
    
    for (j=0,i=(chunk->supported_representations>>1);i;i>>=1) j++;
    if (chunk->supported_representations != (1<<j)) j++;
    
    if (j>=ctx->bins) {
	opcDebugServer(OPC_DEBUG_WARNING, "opcChunkPoolReturnChunk: You should increase number of bins.");
	j=ctx->bins-1;
    }

    LockWrite(ctx->lock);
    if (ctx->last_chunk[j]) {
	ctx->last_chunk[j]->next = chunk;
	ctx->last_chunk[j] = chunk;
    } else {
	ctx->first_chunk[j] = chunk;
	ctx->last_chunk[j] = chunk;
    }
    chunk->next = NULL;    
    LockReturn(ctx->lock);
}

