#include <opcassert.h>

#include <opcerr.h>
#include <opcdatachunk.h>

OPCDataChunkContext opcDataChunkCreateContext(OPCMallocContext mctx) {
    int err;
    LockContext lock;
    ConditionContext cond;
    OPCDataChunkContext ctx;
    
    lock = LockCreateContext(mctx);
    if (!lock) return NULL;
    err = LockInitContext(lock);
    if (err) {
	LockDestroyContext(lock);
	return NULL;
    }
    
    cond = ConditionCreateContext(mctx);
    if (!cond) {
	LockDestroyContext(lock);
	return NULL;
    }
    err = ConditionInitContext(cond);
    if (err) {
	LockDestroyContext(lock);
	ConditionDestroyContext(cond);
	return NULL;
    }

    ctx = (OPCDataChunkContext)opcMalloc(mctx,sizeof(struct OPCDataChunkContextT));
    if (ctx) {
	ctx->malloc = mctx;
	ctx->lock = lock;
	ctx->cond = cond;
	ctx->data = NULL;
	ctx->supported_representations = 0;
    } else {
	LockDestroyContext(lock);
	ConditionDestroyContext(cond);
    }
    return ctx;
}

int opcDataChunkInitContext(OPCDataChunkContext ctx, OPCRepresentationInfo *definfo, unsigned short int supported_representations) {
    long i;
    OPCRepresentationData data;

    assert(ctx);
    assert(supported_representations);
    
    if ((!ctx->data)||(ctx->supported_representations!=supported_representations)) {
	data = (OPCRepresentationData)opcRealloc(ctx->malloc, ctx->data, supported_representations*sizeof(struct OPCRepresentationDataT));
	if (!data) return ERR_OPC_MEM;
	for (i=ctx->supported_representations;i<supported_representations;i++) data[i].data = NULL;
	ctx->data = data;
	ctx->supported_representations=supported_representations;
    }
    return opcDataChunkRestoreContext(ctx, definfo);
}

int opcDataChunkRestoreContext(OPCDataChunkContext ctx, OPCRepresentationInfo *definfo) {
    OPCRepresentationData data;
    long i;

    ctx->next = NULL;
    ctx->referers = 0;
    ctx->delayed_referers = 0;

    data = ctx->data;
    for (i=ctx->supported_representations-1;i>=0;i--) {
	data[i].data = NULL;
	data[i].info = definfo[i];
    }
    return 0;
}


void opcDataChunkFreeContext(OPCDataChunkContext ctx) {
    OPCRepresentationInfo info;
    OPCRepresentationData data;
    short i;
    
    assert(ctx);

    data = ctx->data;
    for (i=ctx->supported_representations-1;i>=0;i--) {
        if (data[i].data) {
	    info = data[i].info;
	    switch(info->destructor_type) {
		case OPC_DATA_DESTRUCTOR_RETURN:
		    ((OPCReturnFunction)(info->Destructor))(info->ctx,data[i].data);
		break;
		case OPC_DATA_DESTRUCTOR_DESTROY:
		    ((OPCDestroyContextFunction)(info->Destructor))(data[i].data);
		break;
		case OPC_DATA_DESTRUCTOR_RETURN_DATA:
		    ((OPCReturnDataFunction)(info->Destructor))(info->ctx,(OPCData)(data+i));
		break;
		case OPC_DATA_DESTRUCTOR_DESTROY_DATA:
		    ((OPCDestroyDataContextFunction)(info->Destructor))((OPCData)(data+i));
		break;
		case OPC_DATA_DESTRUCTOR_FREE:
		    opcFree(info->malloc, data[i].data);
		break;
	    }
	}
	data[i].data = NULL;
    }
}

void opcDataChunkDestroyContext(OPCDataChunkContext ctx) {
    opcDataChunkFreeContext(ctx);
    if (ctx->data) opcFree(ctx->malloc, ctx->data);
    if (ctx->lock) LockDestroyContext(ctx->lock);
    if (ctx->cond) ConditionDestroyContext(ctx->cond);
    opcFree(ctx->malloc,ctx);
}

int opcDataChunkSet(OPCDataChunkContext ctx, unsigned short int format, void *data, unsigned long size) {
    assert(ctx);
    assert(ctx->data);
    assert(format<ctx->supported_representations);

    ctx->data[format].data = data;
    ctx->data[format].size = size;
    
    return 0;    
}

OPCRepresentationData opcDataChunkGet(OPCDataChunkContext ctx, unsigned short int format) {
    assert(ctx);
    assert(ctx->data);
    
    if ((format<ctx->supported_representations)&&(ctx->data[format].data)) return ctx->data+format;
    return NULL;
}

OPCRepresentationData opcDataChunkGetData(OPCDataChunkContext ctx, unsigned short int format) {
    assert(ctx);
    assert(ctx->data);
    
    if (format<ctx->supported_representations)
	return ctx->data+format;

    return NULL;
}

unsigned short opcDataChunkFindData(OPCDataChunkContext ctx) {
    OPCRepresentationData data;
    unsigned short i, representations;

    assert(ctx);
    
    data = ctx->data;
    assert(data);
    
    representations = ctx->supported_representations;
    
    for (i=0;i<representations;i++)
	if (data[i].data) return i;
    return (unsigned short)-1;
}

int opcDataChunkSetInfo(OPCDataChunkContext ctx, const OPCTime time, const OPCQuality quality) {
    assert(ctx);
    assert(time);
    assert(quality);
    
    opcTimeCopy(&ctx->info.time, time);
    opcQualityCopy(&ctx->info.quality, quality);
    return 0;
}

int opcDataChunkCopyInfo(OPCDataChunkContext ctx, const OPCDataInfo info) {
    if (info) {
    	if (info->time) opcTimeCopy(&ctx->info.time, info->time);
	else opcTimeCurrentTime(&ctx->info.time);
    
	if (info->quality) opcQualityCopy(&ctx->info.quality, info->quality);
	else opcQualityCopy(&ctx->info.quality,OPCQualityGood);
    } else {
	opcTimeCurrentTime(&ctx->info.time);
	opcQualityCopy(&ctx->info.quality,OPCQualityGood);
    }
    return 0;
}

OPCDataInfo opcDataChunkGetInfo(OPCDataChunkContext ctx) {
    assert(ctx);
    
    return (OPCDataInfo)ctx;
}

OPCDataChunkContext opcDataChunkGetChunkByInfo(OPCDataInfo ctx) {
    assert(ctx);
    
    return (OPCDataChunkContext)ctx;
}


unsigned long opcDataChunkRef(OPCDataChunkContext ctx) {
    unsigned long res;
    LockWrite(((OPCDataChunkContext)ctx)->lock);
    res=++(((OPCDataChunkContext)ctx)->referers);
    LockReturn(((OPCDataChunkContext)ctx)->lock);
    ConditionBroadcast(ctx->cond);
    return res;
}

unsigned long opcDataChunkUnref(OPCDataChunkContext ctx) {
    unsigned long res;
    LockWrite(((OPCDataChunkContext)ctx)->lock);
    res=--(((OPCDataChunkContext)ctx)->referers);
    LockReturn(((OPCDataChunkContext)ctx)->lock);
    ConditionBroadcast(ctx->cond);
    return res;
}

unsigned long opcDataChunkGetRefs(OPCDataChunkContext ctx) {
    unsigned long res;
    LockRead(((OPCDataChunkContext)ctx)->lock);
    res=((OPCDataChunkContext)ctx)->referers;
    LockReturn(((OPCDataChunkContext)ctx)->lock);
    return res;
}

unsigned long opcDataChunkRefByInfo(OPCDataInfo ctx) {
    unsigned long res;
    LockWrite(((OPCDataChunkContext)ctx)->lock);
    res=++(((OPCDataChunkContext)ctx)->referers);
    LockReturn(((OPCDataChunkContext)ctx)->lock);
    ConditionBroadcast(((OPCDataChunkContext)ctx)->cond);
    return res;
}

unsigned long opcDataChunkUnrefByInfo(OPCDataInfo ctx) {
    unsigned long res;
    LockWrite(((OPCDataChunkContext)ctx)->lock);
    res=--(((OPCDataChunkContext)ctx)->referers);
    LockReturn(((OPCDataChunkContext)ctx)->lock);
    ConditionBroadcast(((OPCDataChunkContext)ctx)->cond);
    return res;
}

unsigned long opcDataChunkGetRefsByInfo(OPCDataInfo ctx) {
    unsigned long res;
    LockRead(((OPCDataChunkContext)ctx)->lock);
    res=((OPCDataChunkContext)ctx)->referers;
    LockReturn(((OPCDataChunkContext)ctx)->lock);
    return res;
}

unsigned long opcDataChunkDelayedRef(OPCDataChunkContext ctx) {
    unsigned long res;
    LockWrite(((OPCDataChunkContext)ctx)->lock);
    res=++(((OPCDataChunkContext)ctx)->delayed_referers);
    LockReturn(((OPCDataChunkContext)ctx)->lock);
    ConditionBroadcast(ctx->cond);
    return res;
}

unsigned long opcDataChunkRefDelayed(OPCDataChunkContext ctx) {
    unsigned long res;
    LockWrite(((OPCDataChunkContext)ctx)->lock);
    (((OPCDataChunkContext)ctx)->delayed_referers)--;
    res=++(((OPCDataChunkContext)ctx)->referers);
    LockReturn(((OPCDataChunkContext)ctx)->lock);
    ConditionBroadcast(ctx->cond);
    return res;
}

unsigned long opcDataChunkDelayedUnref(OPCDataChunkContext ctx) {
    unsigned long res;
    LockWrite(((OPCDataChunkContext)ctx)->lock);
    res=--(((OPCDataChunkContext)ctx)->delayed_referers);
    LockReturn(((OPCDataChunkContext)ctx)->lock);
    ConditionBroadcast(ctx->cond);
    return res;
}

unsigned long opcDataChunkGetDelayedRefs(OPCDataChunkContext ctx) {
    unsigned long res;
    LockRead(((OPCDataChunkContext)ctx)->lock);
    res=((OPCDataChunkContext)ctx)->delayed_referers;
    LockReturn(((OPCDataChunkContext)ctx)->lock);
    return res;
}

unsigned long opcDataChunkDelayedRefByInfo(OPCDataInfo ctx) {
    unsigned long res;
    LockWrite(((OPCDataChunkContext)ctx)->lock);
    res=++(((OPCDataChunkContext)ctx)->delayed_referers);
    LockReturn(((OPCDataChunkContext)ctx)->lock);
    ConditionBroadcast(((OPCDataChunkContext)ctx)->cond);
    return res;
}

unsigned long opcDataChunkRefDelayedByInfo(OPCDataInfo ctx) {
    unsigned long res;
    LockWrite(((OPCDataChunkContext)ctx)->lock);
    (((OPCDataChunkContext)ctx)->delayed_referers)--;
    res=++(((OPCDataChunkContext)ctx)->referers);
    LockReturn(((OPCDataChunkContext)ctx)->lock);
    ConditionBroadcast(((OPCDataChunkContext)ctx)->cond);
    return res;
}

unsigned long opcDataChunkDelayedUnrefByInfo(OPCDataInfo ctx) {
    unsigned long res;
    LockWrite(((OPCDataChunkContext)ctx)->lock);
    res=--(((OPCDataChunkContext)ctx)->delayed_referers);
    LockReturn(((OPCDataChunkContext)ctx)->lock);
    ConditionBroadcast(((OPCDataChunkContext)ctx)->cond);
    return res;
}

unsigned long opcDataChunkGetDelayedRefsByInfo(OPCDataInfo ctx) {
    unsigned long res;
    LockRead(((OPCDataChunkContext)ctx)->lock);
    res=((OPCDataChunkContext)ctx)->delayed_referers;
    LockReturn(((OPCDataChunkContext)ctx)->lock);
    return res;
}


void opcDataChunkLock(OPCDataChunkContext ctx) {
    assert(ctx);
    LockWrite(ctx->lock);
}

void opcDataChunkUnlock(OPCDataChunkContext ctx) {
    assert(ctx);
    LockReturn(ctx->lock);
}

int opcDataChunkWaitUnref(OPCDataChunkContext ctx, unsigned char delayed) {
    assert(ctx);
    if (delayed) {
	ConditionPreLock(ctx->cond);
	while ((ctx->referers)||(ctx->delayed_referers)) ConditionWaitLocked(ctx->cond);
	ConditionUnLock(ctx->cond);
    } else {
	ConditionPreLock(ctx->cond);
	while (ctx->referers) ConditionWaitLocked(ctx->cond);
	ConditionUnLock(ctx->cond);
    }
    return 0;
}

void opcDataChunkRun(OPCDataChunkContext ctx, OPCDataChunkRunFunction func, void *attr) {
    long i;

    assert(ctx);

    for (i=ctx->supported_representations-1;i>=0;i--) 
	if (ctx->data[i].data) func(attr, ctx->data+i);
}
