#define _OPC_REQUESTPOOL_C
#include <stdio.h>
#include <assert.h>

#include <libxml/parser.h>
#include <libxml/schemasInternals.h>

#include <hashtable.h>

#include <opcerr.h>
#include <opcconfig.h>
#include <opcdebug.h>

#include <opc.h>
#include <opcconfiguration.h>
#include <opcrequest.h>
#include <opcrequestpool.h>

void opcRequestPoolZeroContext(OPCRequestPoolContext ctx) {
    assert(ctx);
    
    ctx->opc_commands=NULL;
    ctx->schema_parser = NULL;
    ctx->schema_rep = NULL;
    ctx->schema_ctx = NULL;
    
}

OPCRequestPoolContext opcRequestPoolCreateContext(OPCMallocContext mctx) {
    OPCRequestPoolContext ctx;

    opcDebugServer(OPC_DEBUG_NOTICE, "opcRequestPoolCreateContext");    

    ctx = (OPCRequestPoolContext)opcMalloc(mctx,sizeof(struct OPCRequestPoolContextT));
    if (ctx) {
	ctx->malloc = mctx;
	opcRequestPoolZeroContext(ctx);
    }
    return ctx;
}

int opcRequestPoolInitContext(OPCRequestPoolContext ctx, OPCContext sctx) {
    unsigned long i;
    int err=0;

    opcDebugServer(OPC_DEBUG_NOTICE, "opcRequestPoolInitContext");    

    assert(ctx);
    assert(opc_interfaces);

    opcRequestPoolZeroContext(ctx);

	// Setting link on OPC Server Context
    ctx->sctx=sctx;

	// Setting up command hash
    ctx->opc_commands = HashTableInit(ctx->malloc,OPC_COMMANDS,HASH_TYPE_POINTER,NULL);
    if (!ctx->opc_commands) {
	xmlCleanupCharEncodingHandlers();
	xmlCleanupParser();
	return ERR_OPC_FE_COMMANDS_INIT;
    }

    for (i=0;opc_commands_ctx[i].name;i++) {
	err+=HashTableInsert(ctx->opc_commands,(void*)&opc_commands_ctx[i]);
    }

    if (err) {
	opcRequestPoolFreeContext(ctx);
	return ERR_OPC_MEM;
    }

	// Setting supported interface versions
    for (i=0;opc_interfaces[i];i++) {
	err=opcServerInterfaceAdd(sctx, opc_interfaces[i], opc_extensions[i]);
	if (err) {
	    opcRequestPoolFreeContext(ctx);
	    return err;
	}
    }

	// Loading XSD schema
    ctx->schema_parser=xmlSchemaNewParserCtxt(sctx->config->opc_schema);
    if (!ctx->schema_parser) {
	opcRequestPoolFreeContext(ctx);
	return ERR_OPC_FE_SCHEMA_PARSE;
    }
    ctx->schema_rep=xmlSchemaParse(ctx->schema_parser);
    if (!ctx->schema_rep) {
	opcRequestPoolFreeContext(ctx);
	return ERR_OPC_FE_SCHEMA_PARSE;
    }
    ctx->schema_ctx=xmlSchemaNewValidCtxt(ctx->schema_rep);
    if (!ctx->schema_ctx) {
	opcRequestPoolFreeContext(ctx);
	return ERR_OPC_FE_SCHEMA_PARSE;
    }

    err = opcNamespacesSetOPCNamespace(&sctx->config->namespaces, ctx->schema_rep->targetNamespace);
    if (err) {
	opcRequestPoolFreeContext(ctx);
	return err;
    }

    return 0;
}

void opcRequestPoolFreeContext(OPCRequestPoolContext ctx) {
    opcDebugServer(OPC_DEBUG_NOTICE, "opcRequestPoolFreeContext");    

    assert(ctx);

    if (ctx->schema_ctx) xmlSchemaFreeValidCtxt(ctx->schema_ctx);
    if (ctx->schema_rep) xmlSchemaFree(ctx->schema_rep);
    if (ctx->schema_parser) xmlSchemaFreeParserCtxt(ctx->schema_parser);

    if (ctx->opc_commands) HashTableDestroy(ctx->opc_commands);

    opcRequestPoolZeroContext(ctx);
}


void opcRequestPoolDestroyContext(OPCRequestPoolContext ctx) {
    opcDebugServer(OPC_DEBUG_NOTICE, "opcRequestPoolDestroyContext");    

    opcRequestPoolFreeContext(ctx);
    opcFree(ctx->malloc,ctx);
}

OPCRequestContext opcRequestPoolGetRequest(OPCRequestPoolContext ctx) {
    int err;
    OPCRequestContext rctx;
    
    assert(ctx);
    
    rctx = opcRequestCreateContext();
    if (rctx) {
	err=opcRequestInitContext(rctx,ctx->sctx,ctx->opc_commands);
	if (err) {
	    free(rctx);
	    return NULL;
	}
    }
    return rctx;
}

int opcRequestPoolReturnRequest(OPCRequestPoolContext ctx, OPCRequestContext rctx) {
    assert(ctx);
    assert(rctx);
    opcRequestDestroyContext(rctx);
    return 0;
}
