#include <sys/time.h>
#include <assert.h>

#include <opcmalloc.h>
#include <condition.h>

static void ConditionZeroContext(ConditionContext ctx) {
    ctx->initialized = 0;
}

ConditionContext ConditionCreateContext(OPCMallocContext mctx) {
    ConditionContext ctx;
    ctx = (ConditionContext)opcMalloc(mctx, sizeof(struct ConditionContextT));
    if (ctx) {
	ctx->malloc = mctx;
	ConditionZeroContext(ctx);
    }
    return ctx;
}

int ConditionInitContext(ConditionContext ctx) {
    int err;

    assert(ctx);

    ConditionZeroContext(ctx);
    
    pthread_mutex_init(&ctx->mutex,NULL);
    pthread_cond_init(&ctx->cond,NULL);

    ctx->initialized = 1;

    return 0;
}

void ConditionFreeContext(ConditionContext ctx) {
    assert(ctx);
    if (ctx->initialized) {
	pthread_cond_destroy(&ctx->cond);
	pthread_mutex_destroy(&ctx->mutex);
    }
    ConditionZeroContext(ctx);
}

void ConditionDestroyContext(ConditionContext ctx) {
    assert(ctx);
    ConditionFreeContext(ctx);
    opcFree(ctx->malloc, ctx);
}

void ConditionBroadcast(ConditionContext ctx) {
    assert(ctx);
    pthread_mutex_lock(&ctx->mutex);
    pthread_cond_broadcast(&ctx->cond);
    pthread_mutex_unlock(&ctx->mutex);
}

void ConditionSend(ConditionContext ctx) {
    assert(ctx);
    pthread_mutex_lock(&ctx->mutex);
    pthread_cond_signal(&ctx->cond);
    pthread_mutex_unlock(&ctx->mutex);
}

void ConditionPreLock(ConditionContext ctx) {
    assert(ctx);
    pthread_mutex_lock(&ctx->mutex);
}

void ConditionUnLock(ConditionContext ctx) {
    assert(ctx);
    pthread_mutex_unlock(&ctx->mutex);
}

void ConditionWait(ConditionContext ctx) {
    assert(ctx);
    pthread_cond_wait(&ctx->cond,&ctx->mutex);
    pthread_mutex_unlock(&ctx->mutex);
}

void ConditionWaitLocked(ConditionContext ctx) {
    assert(ctx);
    pthread_cond_wait(&ctx->cond,&ctx->mutex);
}

int ConditionTimedWait(ConditionContext ctx, time_t sec) {
    int err;
    struct timeval now;
    struct timespec timeout;
    struct timezone tz;

    assert(ctx);

    gettimeofday(&now,&tz);
    timeout.tv_sec = now.tv_sec;
    timeout.tv_nsec = now.tv_usec * 1000;
    
    err = pthread_cond_timedwait(&ctx->cond,&ctx->mutex,&timeout);
    pthread_mutex_unlock(&ctx->mutex);
    return err;
}

int ConditionTimedWaitLocked(ConditionContext ctx, time_t sec) {
    struct timeval now;
    struct timespec timeout;
    struct timezone tz;

    assert(ctx);

    gettimeofday(&now,&tz);
    timeout.tv_sec = now.tv_sec;
    timeout.tv_nsec = now.tv_usec * 1000;

    return pthread_cond_timedwait(&ctx->cond,&ctx->mutex,&timeout);
}
