#include <stdio.h>
#include <errno.h>
#include <assert.h>

#include <opcmalloc.h>
#include <ilock.h>

static void ILockZeroContext(ILockContext ctx) {
    ctx->initialized = 0;
    ctx->pid = 0;
}

ILockContext ILockCreateContext(OPCMallocContext mctx) {
    ILockContext ctx;
    ctx = (ILockContext)opcMalloc(mctx, sizeof(struct ILockContextT));
    if (ctx) {
	ctx->malloc = mctx;
	ILockZeroContext(ctx);
    }
    return ctx;
}

int ILockInitContext(ILockContext ctx) {
    assert(ctx);
    ILockZeroContext(ctx);
#ifdef OPC_SUPPORT_RWLOCK
    pthread_rwlock_init(&ctx->lock_mutex,NULL);
#else
    pthread_mutex_init(&ctx->lock_mutex,NULL);
#endif /* OPC_SUPPORT_RWLOCK */
    ctx->initialized = 1;
    return 0;
}

void ILockFreeContext(ILockContext ctx) {
    assert(ctx);
#ifdef OPC_SUPPORT_RWLOCK
    if (ctx->initialized) pthread_rwlock_destroy(&ctx->lock_mutex);
#else
    if (ctx->initialized) pthread_mutex_destroy(&ctx->lock_mutex);
#endif /* OPC_SUPPORT_RWLOCK */
    ILockZeroContext(ctx);
}

void ILockDestroyContext(ILockContext ctx) {
    assert(ctx);
    ILockFreeContext(ctx);
    opcFree(ctx->malloc, ctx);
}

int ILockRead(ILockContext ctx) {
    assert(ctx);
    assert(ctx->initialized);
#ifdef OPC_SUPPORT_RWLOCK
    return pthread_rwlock_rdlock(&ctx->lock_mutex);
#else
    return pthread_mutex_lock(&ctx->lock_mutex);
#endif /* OPC_SUPPORT_RWLOCK */
}

int ILockWrite(ILockContext ctx) {
    int err;
    assert(ctx);
    assert(ctx->initialized);

    if (pthread_equal(ctx->pid,pthread_self())) return EDEADLK;
#ifdef OPC_SUPPORT_RWLOCK
    err=pthread_rwlock_wrlock(&ctx->lock_mutex);
#else
    err=pthread_mutex_lock(&ctx->lock_mutex);
#endif /* OPC_SUPPORT_RWLOCK */
    ctx->pid = pthread_self();
    return err;
}

int ILockTryRead(ILockContext ctx) {
    assert(ctx);
    assert(ctx->initialized);

    if (pthread_equal(ctx->pid,pthread_self())) return EDEADLK;
    
#ifdef OPC_SUPPORT_RWLOCK
    return pthread_rwlock_tryrdlock(&ctx->lock_mutex);
#else
    return pthread_mutex_trylock(&ctx->lock_mutex);
#endif /* OPC_SUPPORT_RWLOCK */
}

int ILockTryWrite(ILockContext ctx) {
    int err;
    assert(ctx);
    assert(ctx->initialized);

    if (pthread_equal(ctx->pid,pthread_self())) return EDEADLK;
#ifdef OPC_SUPPORT_RWLOCK
    err=pthread_rwlock_trywrlock(&ctx->lock_mutex);
#else
    err=pthread_mutex_trylock(&ctx->lock_mutex);
#endif /* OPC_SUPPORT_RWLOCK */
    if (!err) ctx->pid = pthread_self();
    return err;
}

int ILockReturn(ILockContext ctx) {
    assert(ctx);
    assert(ctx->initialized);
    ctx->pid = 0;
#ifdef OPC_SUPPORT_RWLOCK
    return pthread_rwlock_unlock(&ctx->lock_mutex);
#else
    return pthread_mutex_unlock(&ctx->lock_mutex);
#endif /* OPC_SUPPORT_RWLOCK */
}
