#include <stdio.h>
#include <stdlib.h>
#include <dlfcn.h>
#include <assert.h>

#include <opcerr.h>
#include <opcdebug.h>

#include <opctools.h>

#include <thread.h>
#include <opcplugin.h>

void opcPluginZeroContext(OPCPluginContext ctx) {
    assert(ctx);
    ThreadZeroContext((ThreadContext)ctx);
    ctx->fullname=NULL;
    ctx->filename=NULL;
    ctx->next=NULL;
    ctx->modification_time=0;
    ctx->mark=0;
}

OPCPluginContext opcPluginCreateContext(OPCMallocContext mctx) {
    OPCPluginContext ctx;

    ctx = (OPCPluginContext)opcMalloc(mctx, sizeof(struct OPCPluginContextT));
    opcDebugServer(OPC_DEBUG_NOTICE, "opcPlugin %p: CreateContext", ctx);
    if (ctx) {
	ctx->malloc = mctx;
	opcPluginZeroContext(ctx);
    }
    return ctx;
}

static int opcPluginWaitReload(OPCPluginContext ctx) {
    assert(ctx);
    
    ThreadSetStatusBits((ThreadContext)ctx, OPC_PLUGIN_ERROR_FLAG);
    ThreadWaitStatusSomeBitsOn((ThreadContext)ctx, OPC_THREAD_TERMINATE_FLAG|OPC_PLUGIN_RELOAD_FLAG);
    if (ThreadCheckStatusBitsOff((ThreadContext)ctx, OPC_THREAD_TERMINATE_FLAG)) return 1;
    return 0;
}

static int *opcPluginStartCallback(OPCPluginContext ctx) {
    int err = 0;
    void *handle;
    PluginInfoContext (*info_function)();
    PluginInfoContext plugin_info;
    plugin_step_function step;
    void *pctx;
    
    assert(ctx);
    assert(ctx->fullname);
    
    opcDebugServer(OPC_DEBUG_NOTICE,"opcPlugin %p: Thread: %8.lx",ctx,pthread_self());
reload:
    opcDebugServer(OPC_DEBUG_INFO,"opcPlugin %p: Reload: %s",ctx, ctx->fullname);
    
    ctx->modification_time = GetFileModificationTime(ctx->fullname);
    
//    handle = dlopen(ctx->name, RTLD_LAZY|RTLD_GLOBAL);
    handle = dlopen(ctx->fullname, RTLD_LAZY);
    if (!handle) {
	if (opcPluginWaitReload(ctx)) goto reload;
	opcDebugServer(OPC_DEBUG_DEBUG,"opcPlugin %p: Terminating", ctx);
	return ThreadReturnInt(NULL,ERR_OPC_PLUGIN_LOAD);
    }
    
    ThreadResetStatusBits((ThreadContext)ctx,OPC_PLUGIN_RELOAD_FLAG);
    
    info_function = dlsym(handle,"PluginInfo");
    if (!info_function) {
	dlclose(handle);
	if (opcPluginWaitReload(ctx)) goto reload;
	opcDebugServer(OPC_DEBUG_DEBUG,"opcPlugin %p: Terminating", ctx);
	return ThreadReturnInt(NULL,ERR_OPC_PLUGIN_INFOFUNCTION);
    }

    plugin_info = info_function();
    if (!plugin_info) {
	dlclose(handle);
	if (opcPluginWaitReload(ctx)) goto reload;
	opcDebugServer(OPC_DEBUG_DEBUG,"opcPlugin %p: Terminating", ctx);
	return ThreadReturnInt(NULL,ERR_OPC_PLUGIN_INVALIDINFO);
    }
    
    if (plugin_info->create) pctx = plugin_info->create();
    else pctx = NULL;
    if (plugin_info->init) {
	err = plugin_info->init(pctx,ctx,ctx->plugin_attribute);
	if (err) {
	    if (plugin_info->destroy) plugin_info->destroy(pctx);
	    dlclose(handle);
	    if (opcPluginWaitReload(ctx)) goto reload;
	    opcDebugServer(OPC_DEBUG_DEBUG,"opcPlugin %p: Terminating", ctx);
	    return ThreadReturnInt(NULL,ERR_OPC_PLUGIN_INIT);
	}
    }
    
    if (plugin_info->step) {
	step=plugin_info->step;
	while(ThreadCheckStatusBitsOff((ThreadContext)ctx, OPC_THREAD_TERMINATE_FLAG|OPC_PLUGIN_RELOAD_FLAG)) {
	    ThreadWaitStatusSomeBitsOn((ThreadContext)ctx, OPC_THREAD_TERMINATE_FLAG|OPC_PLUGIN_RUN_FLAG|OPC_PLUGIN_RELOAD_FLAG);
	    while (ThreadCheckStatusBits((ThreadContext)ctx,OPC_PLUGIN_RUN_FLAG,OPC_THREAD_TERMINATE_FLAG|OPC_PLUGIN_RELOAD_FLAG)) {
		opcDebugServer(OPC_DEBUG_DEBUG,"opcPlugin %p: Step",ctx);
		err=step(pctx);
		opcDebugServer(err?OPC_DEBUG_WARNING:OPC_DEBUG_DEBUG,"opcPlugin %p: Step (%lx)",ctx,err);
		if (err) opcPluginWaitReload(ctx);
	    }
	}
    } else {
	ThreadWaitStatusBitsOn((ThreadContext)ctx, OPC_THREAD_TERMINATE_FLAG);
    }

    if (plugin_info->destroy) plugin_info->destroy(pctx);
    dlclose(handle);

    if (ThreadCheckStatusBits((ThreadContext)ctx,OPC_PLUGIN_RELOAD_FLAG,OPC_THREAD_TERMINATE_FLAG)) goto reload;
    
    opcDebugServer(OPC_DEBUG_DEBUG,"opcPlugin %p: Terminating", ctx);
    return ThreadReturnInt(NULL,err);
}

int opcPluginInitContext(OPCPluginContext ctx, const char *path, const char *filename, void *attr) {
    int err;

    assert(ctx);
    assert(path);
    assert(filename);

    opcPluginZeroContext(ctx);

    opcDebugServer(OPC_DEBUG_NOTICE, "opcPlugin %p: InitContext", ctx);    
    
    ctx->fullname = opcCreateFullName(ctx->malloc,path,filename);
    if (!ctx->fullname) return ERR_OPC_MEM;
    
    err=opcExchangeString(ctx->malloc,&ctx->filename,filename);
    if (err) {
	opcPluginFreeContext(ctx);
	return ERR_OPC_MEM;
    }
    
    err=ThreadInitContext((ThreadContext)ctx);
    if (err) {
	opcPluginFreeContext(ctx);
	return err;
    }
    
    ctx->plugin_attribute = attr;

    err= ThreadRun((ThreadContext)ctx, (ThreadFunctionT)opcPluginStartCallback, (void*)ctx);
    if (err) {
	opcPluginFreeContext(ctx);
	return err;
    }

    return 0;
}

int opcPluginFreeContext(OPCPluginContext ctx) {
    int err;
    opcDebugServer(OPC_DEBUG_NOTICE, "opcPlugin %p: FreeContext",ctx);    

    assert(ctx);

    if (ctx->fullname) opcExchangeString(ctx->malloc,&ctx->fullname,NULL);
    if (ctx->filename) opcExchangeString(ctx->malloc,&ctx->filename,NULL);
    err=ThreadFreeContext((ThreadContext)ctx);
    opcPluginZeroContext(ctx);
    return err;
}

int opcPluginDestroyContext(OPCPluginContext ctx) {
    int err;
    opcDebugServer(OPC_DEBUG_NOTICE, "opcPlugin %p: DestroyContext",ctx);    

    assert(ctx);

    err=opcPluginFreeContext(ctx);
    opcFree(ctx->malloc, ctx);
    return err;
}


void opcPluginRun(OPCPluginContext ctx) {
    ThreadSetStatusBits((ThreadContext)ctx,OPC_PLUGIN_RUN_FLAG);
}

void opcPluginStop(OPCPluginContext ctx) {
    ThreadResetStatusBits((ThreadContext)ctx,OPC_PLUGIN_RUN_FLAG);
}

void opcPluginTerminate(OPCPluginContext ctx) {
    ThreadSetStatusBits((ThreadContext)ctx,OPC_THREAD_TERMINATE_FLAG);
}

void opcPluginReload(OPCPluginContext ctx) {
    ThreadSetStatusBits((ThreadContext)ctx,OPC_PLUGIN_RELOAD_FLAG);
}

int opcPluginContinue(OPCPluginContext ctx) {
    return ThreadCheckStatusBitsOff((ThreadContext)ctx,OPC_PLUGIN_RELOAD_FLAG|OPC_THREAD_TERMINATE_FLAG|OPC_PLUGIN_BREAK_FLAG);
}

void opcPluginMark(OPCPluginContext ctx) {
    assert(ctx);
    ctx->mark = 1;
}

void opcPluginUnMark(OPCPluginContext ctx) {
    assert(ctx);
    ctx->mark = 0;
}

int opcPluginMarked(OPCPluginContext ctx) {
    assert(ctx);
    return ctx->mark;
}

int opcPluginCheckName(const char *name) {
    int length;

    assert(name);

    length = strlen(name)-sizeof(OPC_PLUGIN_EXTENSION);
    if (length++<0) return 2;
    
    return strcmp(name+length,OPC_PLUGIN_EXTENSION);
}
