#include <stdio.h>
#include <stdlib.h>
#include <sys/types.h>
#include <dirent.h>
#include <assert.h>

#include <opcerr.h>

#include <opcdebug.h>

#include <opctools.h>

#include <opcpluginset.h>


OPCPluginSetContext opcPluginSetCreateContext(OPCMallocContext mctx) {
    OPCPluginSetContext ctx;

    ctx = (OPCPluginSetContext)opcMalloc(mctx, sizeof(struct OPCPluginSetContextT));
    opcDebugServer(OPC_DEBUG_NOTICE, "opcPluginSet %p: CreateContext", ctx);    
    if (ctx) {
	ctx->malloc = mctx;
	opcPluginSetZeroContext(ctx);
    }
    return ctx;
}

int opcPluginSetInitContext(OPCPluginSetContext ctx, const char *path, void *attr) {
    assert(ctx);
    assert(path);

    opcDebugServer(OPC_DEBUG_NOTICE, "opcPluginSet %p: InitContext",ctx);    

    opcPluginSetZeroContext(ctx);
    opcExchangeString(ctx->malloc,&ctx->path,path);
    ctx->plugin_attribute = attr;
    return opcPluginSetRescanDir(ctx);
}

void opcPluginSetZeroContext(OPCPluginSetContext ctx) {
    ctx->path = NULL;
    ctx->plugin = NULL;
    ctx->run = 0;
}

void opcPluginSetFreeContext(OPCPluginSetContext ctx) {
    int err;
    OPCPluginContext pctx, tctx;
    assert(ctx);

    opcDebugServer(OPC_DEBUG_NOTICE, "opcPluginSet %p: FreeContext", ctx);    

    if (ctx->path) opcExchangeString(ctx->malloc,&ctx->path, NULL);

    opcPluginSetTerminate(ctx);
    for (pctx=ctx->plugin;pctx;pctx=tctx) {
	tctx = pctx->next;
	err=opcPluginDestroyContext(pctx);
	opcDebugServer(err?OPC_DEBUG_ERROR:OPC_DEBUG_INFO,"opcPluginSet %p: opcPlugin %p: Terminating: %s",ctx,pctx,opcError(err,NULL));
    }

    opcPluginSetZeroContext(ctx);

    opcDebugServer(OPC_DEBUG_NOTICE, "opcPluginSet %p: FreeContext Ok",ctx);    
}

void opcPluginSetDestroyContext(OPCPluginSetContext ctx) {
    assert(ctx);

    opcDebugServer(OPC_DEBUG_NOTICE, "opcPluginSet %p: DestroyContext", ctx);    
    
    opcPluginSetFreeContext(ctx);
    opcFree(ctx->malloc, ctx);
}


int opcPluginSetRescanDir(OPCPluginSetContext ctx) {
    int err = 0;
    OPCPluginContext pctx,last;
    DIR *pdir;
    struct dirent *file;

    assert(ctx);
    assert(ctx->path);
    assert(ctx->run>=0);

    opcDebugServer(OPC_DEBUG_NOTICE, "opcPluginSet %p: RescanDir",ctx);    

    pdir=opendir(ctx->path);
    if (!pdir) return ERR_OPC_PLUGIN_PATH;

    for (pctx=ctx->plugin;pctx;pctx=pctx->next) opcPluginMark(pctx);
    
    while (file=readdir(pdir)) {
	if (opcPluginCheckName(file->d_name)) continue;

	for (pctx=ctx->plugin;pctx;last=pctx,pctx=pctx->next)
	    if (opcPluginMarked(pctx)) {
		if (strcmp(file->d_name,pctx->filename)==0) {
		    if ((pctx->modification_time)&&(GetFileModificationTime(file->d_name)!=pctx->modification_time))
			opcPluginReload(pctx);
		    opcPluginUnMark(pctx);
		    break;
		}
	    }
	
	if (!pctx) {
	    pctx = opcPluginCreateContext(ctx->malloc);
	    if (!pctx) {
		err++;
		continue;
	    }
	    
	    err = opcPluginInitContext(pctx, ctx->path, file->d_name, ctx->plugin_attribute);
	    if (err) {
		opcPluginDestroyContext(pctx);
		err++;
		continue;
	    }
	    
	    if (ctx->run) opcPluginRun(pctx);
		
	    if (ctx->plugin) last->next = pctx;
	    else ctx->plugin = pctx;
	}
    }

    for (pctx=ctx->plugin;pctx;pctx=pctx->next)
	if (opcPluginMarked(pctx)) opcPluginTerminate(pctx);
    
    for (pctx=ctx->plugin,last=NULL;pctx;last=pctx,pctx=pctx->next)
	if (opcPluginMarked(pctx)) {
	    if (last) last->next=pctx->next;
	    else ctx->plugin=pctx->next;
	    opcPluginDestroyContext(pctx);
	}
    close(pdir);

    opcDebugServer(OPC_DEBUG_NOTICE, "opcPluginSet %p: RescanDir Ok",ctx);
    return 0;
}

void opcPluginSetSetStatus(OPCPluginSetContext ctx, dword bits) {
    OPCPluginContext pctx;
    
    assert(ctx);
    
    for (pctx=ctx->plugin;pctx;pctx=pctx->next)
	ThreadSetStatus((ThreadContext)pctx,bits);
}

void opcPluginSetSetStatusBits(OPCPluginSetContext ctx, dword bits) {
    OPCPluginContext pctx;
    
    assert(ctx);
    for (pctx=ctx->plugin;pctx;pctx=pctx->next)
	ThreadSetStatusBits((ThreadContext)pctx,bits);
}

void opcPluginSetResetStatusBits(OPCPluginSetContext ctx, dword bits) {
    OPCPluginContext pctx;
    
    assert(ctx);
    
    for (pctx=ctx->plugin;pctx;pctx=pctx->next)
	ThreadResetStatusBits((ThreadContext)pctx,bits);
}

int opcPluginSetCheckStatusBitsOn(OPCPluginSetContext ctx, dword bits) {
    OPCPluginContext pctx;
    
    assert(ctx);
    
    for (pctx=ctx->plugin;pctx;pctx=pctx->next)
	if (!ThreadCheckStatusBitsOn((ThreadContext)pctx,bits)) return 0;
    return 1;
}

int opcPluginSetCheckStatusBitsOff(OPCPluginSetContext ctx, dword bits) {
    OPCPluginContext pctx;
    
    assert(ctx);
    
    for (pctx=ctx->plugin;pctx;pctx=pctx->next)
	if (!ThreadCheckStatusBitsOff((ThreadContext)pctx,bits)) return 0;
    return 1;
}

int opcPluginSetCheckStatusBits(OPCPluginSetContext ctx, dword onbits, dword offbits) {
    OPCPluginContext pctx;
    
    assert(ctx);
    
    for (pctx=ctx->plugin;pctx;pctx=pctx->next)
	if (!ThreadCheckStatusBits((ThreadContext)pctx,onbits,offbits)) return 0;
    return 1;
}

int opcPluginSetCheckStatusSomeBitsOn(OPCPluginSetContext ctx, dword bits) {
    OPCPluginContext pctx;
    
    assert(ctx);
    
    for (pctx=ctx->plugin;pctx;pctx=pctx->next)
	if (!ThreadCheckStatusSomeBitsOn((ThreadContext)pctx,bits)) return 0;
    return 1;
}

int opcPluginSetCheckStatusSomeBitsOff(OPCPluginSetContext ctx, dword bits) {
    OPCPluginContext pctx;
    
    assert(ctx);
    
    for (pctx=ctx->plugin;pctx;pctx=pctx->next)
	if (!ThreadCheckStatusSomeBitsOff((ThreadContext)pctx,bits)) return 0;
    return 1;
}

int opcPluginSetCheckStatusSomeBits(OPCPluginSetContext ctx, dword onbits, dword offbits) {
    OPCPluginContext pctx;
    
    assert(ctx);
    
    for (pctx=ctx->plugin;pctx;pctx=pctx->next)
	if (!ThreadCheckStatusSomeBits((ThreadContext)pctx,onbits,offbits)) return 0;
    return 1;
}

void opcPluginSetWaitStatusBitsOn(OPCPluginSetContext ctx, dword bits) {
    OPCPluginContext pctx;
    
    assert(ctx);
    
    for (pctx=ctx->plugin;pctx;pctx=pctx->next)
	ThreadWaitStatusBitsOn((ThreadContext)pctx,bits);
}

void opcPluginSetWaitStatusBitsOff(OPCPluginSetContext ctx, dword bits) {
    OPCPluginContext pctx;
    
    assert(ctx);
    
    for (pctx=ctx->plugin;pctx;pctx=pctx->next)
	ThreadWaitStatusBitsOff((ThreadContext)pctx,bits);
}

void opcPluginSetWaitStatusBits(OPCPluginSetContext ctx, dword onbits, dword offbits) {
    OPCPluginContext pctx;
    
    assert(ctx);
    
    for (pctx=ctx->plugin;pctx;pctx=pctx->next)
	ThreadWaitStatusBits((ThreadContext)pctx,onbits,offbits);
}

void opcPluginSetWaitStatusSomeBitsOn(OPCPluginSetContext ctx, dword bits) {
    OPCPluginContext pctx;
    
    assert(ctx);
    
    for (pctx=ctx->plugin;pctx;pctx=pctx->next)
	ThreadWaitStatusSomeBitsOn((ThreadContext)pctx,bits);
}

void opcPluginSetWaitStatusSomeBitsOff(OPCPluginSetContext ctx, dword bits) {
    OPCPluginContext pctx;
    
    assert(ctx);
    
    for (pctx=ctx->plugin;pctx;pctx=pctx->next)
	ThreadWaitStatusSomeBitsOff((ThreadContext)pctx,bits);
}

void opcPluginSetWaitStatusSomeBits(OPCPluginSetContext ctx, dword onbits, dword offbits) {
    OPCPluginContext pctx;
    
    assert(ctx);
    
    for (pctx=ctx->plugin;pctx;pctx=pctx->next)
	ThreadWaitStatusSomeBits((ThreadContext)pctx,onbits,offbits);
}

int opcPluginSetStart(OPCPluginSetContext ctx) {
    assert(ctx);
    ctx->run = 1;
    opcPluginSetSetStatusBits(ctx,OPC_PLUGIN_RUN_FLAG);
}

int opcPluginSetStop(OPCPluginSetContext ctx) {
    assert(ctx);
    ctx->run = 0;
    opcPluginSetSetStatusBits(ctx,OPC_PLUGIN_RUN_FLAG);
}

int opcPluginSetBreak(OPCPluginSetContext ctx) {
    assert(ctx);
    opcPluginSetSetStatusBits(ctx,OPC_PLUGIN_BREAK_FLAG);
}

int opcPluginSetTerminate(OPCPluginSetContext ctx) {
    assert(ctx);
    ctx->run = -1;
    opcPluginSetSetStatusBits(ctx,OPC_THREAD_TERMINATE_FLAG);
}

void opcPluginSetWaitTerminate(OPCPluginSetContext ctx) {
    OPCPluginContext pctx;
    
    assert(ctx);
    
    for (pctx=ctx->plugin;pctx;pctx=pctx->next)
	ThreadWaitTerminate((ThreadContext)pctx);
}

void opcPluginSetTerminateAndWait(OPCPluginSetContext ctx) {
    assert(ctx);
    opcPluginSetTerminate(ctx);
    opcPluginSetWaitTerminate(ctx);
}

int opcPluginSetIsTerminated(OPCPluginSetContext ctx) {
    assert(ctx);
    return opcPluginSetCheckStatusBitsOn(ctx,OPC_THREAD_TERMINATE_FLAG);
}
