#include <signal.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <errno.h>
#include <assert.h>

#include <opcconfig.h>
#include <opcerr.h>
#include <opcdebug.h>

#include <opcsignal.h>
#include <opcmalloc.h>

#include <opcstringlist.h>

static void signal_nop_handler(int signo) {}
static sigset_t signal_sigmask;

int opcSignalInit() {
    struct sigaction act;
    sigset_t set;

    opcDebugServer(OPC_DEBUG_NOTICE, "opcSignalInit");    

    if (sigemptyset(&set)==-1) return ERR_OPC_SIGNAL_SET_INIT;
    if (sigaddset(&set,SIGUSR1)==-1) return ERR_OPC_SIGNAL_SET_ADD;
    if (sigaddset(&set,SIGUSR2)==-1) return ERR_OPC_SIGNAL_SET_ADD;
    if (sigprocmask(SIG_BLOCK,&set,NULL)==-1) return ERR_OPC_SIGNAL_MASK_INIT;
    
    act.sa_handler = signal_nop_handler;
    act.sa_flags = 0;

    if (sigaction(SIGUSR1,&act,NULL)==-1) return ERR_OPC_SIGNAL_ACTION_SET;
    if (sigaction(SIGUSR2,&act,NULL)==-1) return ERR_OPC_SIGNAL_ACTION_SET;

    if (sigemptyset(&signal_sigmask)==-1) return ERR_OPC_SIGNAL_SET_INIT;
    if (sigaddset(&signal_sigmask,SIGUSR1)==-1) return ERR_OPC_SIGNAL_SET_ADD;
    if (sigaddset(&signal_sigmask,SIGUSR2)==-1) return ERR_OPC_SIGNAL_SET_ADD;

    return 0;
}

void opcSignalDestroy() {
    opcDebugServer(OPC_DEBUG_NOTICE, "opcSignalDestroy");    
    
    return ;
}

void opcSignalZeroContext(OPCSignalContext ctx) {
    assert(ctx);
    ctx->lock = NULL;
    ctx->usr1 = NULL;
    ctx->usr2 = NULL;
    ctx->ExitFlag = 0;
}

OPCSignalContext opcSignalCreateContext(OPCMallocContext mctx) {
    OPCSignalContext ctx;

    opcDebugServer(OPC_DEBUG_NOTICE, "opcSignalCreateContext");    

    ctx = (OPCSignalContext)opcMalloc(mctx, sizeof(struct OPCSignalContextT));
    if (ctx) {
	ctx->malloc = mctx;
	opcSignalZeroContext(ctx);
    }
    return ctx;
}

int opcSignalInitContext(OPCSignalContext ctx) {
    int err;

    opcDebugServer(OPC_DEBUG_NOTICE, "opcSignalInitContext");    

    assert(ctx);
    opcSignalZeroContext(ctx);
    
    ctx->lock = LockCreateContext(ctx->malloc);
    if (!ctx->lock) return ERR_OPC_MEM;
    
    err = LockInitContext(ctx->lock);
    if (err) {
	opcSignalFreeContext(ctx);
	return err;
    }
    
    ctx->usr1 = opcStringListCreate(ctx->malloc);
    if (!ctx->usr1) {
	opcSignalFreeContext(ctx);
	return ERR_OPC_MEM;
    }
    
    ctx->usr2 = opcStringListCreate(ctx->malloc);
    if (!ctx->usr2) {
	opcSignalFreeContext(ctx);
	return ERR_OPC_MEM;
    }
    
    return 0;
}

int opcSignalFreeContext(OPCSignalContext ctx) {
    assert(ctx);

    opcDebugServer(OPC_DEBUG_NOTICE, "opcSignalFreeContext");    
    
    if (ctx->lock) LockDestroyContext(ctx->lock);
    if (ctx->usr1) opcStringListDestroy(ctx->usr1);
    if (ctx->usr2) opcStringListDestroy(ctx->usr2);
    
    opcSignalZeroContext(ctx);
}

void opcSignalDestroyContext(OPCSignalContext ctx) {
    opcDebugServer(OPC_DEBUG_NOTICE, "opcSignalDestroyContext");    

    assert(ctx);
    opcSignalFreeContext(ctx);
    opcFree(ctx->malloc, ctx);
}

int opcSignalRegister(OPCSignalContext ctx, OPCSignal sig) {
    int err;
    pthread_t pid;
    
    assert(ctx);

    pid = pthread_self();
    
    LockWrite(ctx->lock);
    switch (sig) {
	case OPC_SIGNAL_USR1:
	    err=opcStringListAdd(ctx->usr1,&pid,sizeof(pthread_t));
	break;
	case OPC_SIGNAL_USR2:
	    err=opcStringListAdd(ctx->usr1,&pid,sizeof(pthread_t));
	break;
	default:
	    err=ERR_OPC_SIGNAL_NOT_SUPPORTED;
    }
    LockReturn(ctx->lock);
    
    return err;
}

int opcSignalCancel(OPCSignalContext ctx, OPCSignal sig) {
    int err;
    pthread_t pid;
    
    assert(ctx);
    
    pid = pthread_self();
    
    LockWrite(ctx->lock);
    switch (sig) {
	case OPC_SIGNAL_USR1:
	    err=opcStringListRemove(ctx->usr1,&pid,sizeof(pthread_t));
	break;
	case OPC_SIGNAL_USR2:
	    err=opcStringListRemove(ctx->usr1,&pid,sizeof(pthread_t));
	break;
	default:
	    err=ERR_OPC_SIGNAL_NOT_SUPPORTED;
    }
    LockReturn(ctx->lock);
    return err;
}

int opcSignalSend(OPCSignalContext ctx, OPCSignal sig) {
    int err;
    OPCStringList sl;
    pthread_t pid;
    int signo;
    
    assert(ctx);

    LockWrite(ctx->lock);
    switch (sig) {
	case OPC_SIGNAL_USR1:
	    sl=ctx->usr1;
	    signo=SIGUSR1;
	break;
	case OPC_SIGNAL_USR2:
	    sl=ctx->usr2;
	    signo=SIGUSR2;
	break;
	default:
	    LockReturn(ctx->lock);
	    return ERR_OPC_SIGNAL_NOT_SUPPORTED;
    }
    
    for (;sl;sl=sl->next) {
        if (sl->item) {
    	    pid = *(pthread_t*)(sl->item);
//    	    pthread_kill(pid, signo);
	}
    }
    
    LockReturn(ctx->lock);
    
    return 0;
}

int opcSignalSendTo(pthread_t thread, OPCSignal sig) {
    int signo;
    
    switch(sig) {
	case OPC_SIGNAL_USR1:
	    signo=SIGUSR1;
	break;
	case OPC_SIGNAL_USR2:
	    signo=SIGUSR2;
	break;
	default:
	    return ERR_OPC_SIGNAL_NOT_SUPPORTED;
    }
    pthread_kill(thread,signo);

    return 0;
}

int opcSignalTerminate(OPCSignalContext ctx) {
    assert(ctx);
    ctx->ExitFlag=1;
    return opcSignalSend(ctx,OPC_SIGNAL_USR1);
}

int opcSignalIsTerminated(OPCSignalContext ctx) {
    assert(ctx);
    return ctx->ExitFlag;
}

int opcSignalAccept(int s, int *newsock) {
    socklen_t socksize=sizeof(struct sockaddrT);
    struct sockaddrT clisock;
    int err;
    fd_set rd;
    

select_again:    
    FD_ZERO(&rd);
    FD_SET(s,&rd);

    err=pselect(s+1,&rd,NULL,NULL,NULL,&signal_sigmask);
    if (err<=0) {
	if (errno==EINTR) return ERR_OPC_SIGNAL_EVENT;
	return ERR_OPC_FE_SELECT;
    }
    
    err=accept(s,(struct sockaddr*)&clisock,&socksize);
    if (err<0) {
	if ((errno==EAGAIN)||(errno==EWOULDBLOCK)) goto select_again;
	return ERR_OPC_FE_ACCEPT;
    }
    *newsock=err;
    return 0;
}

int opcSignalAcceptMulti(unsigned long number, int *s, int *newsock) {
    int i, max=0;
    socklen_t socksize=sizeof(struct sockaddrT);
    struct sockaddrT clisock;
    int err;
    fd_set rd;
    

select_again:    
    FD_ZERO(&rd);
    for (i=0;i<number;i++) {
	if (s[i]>max) max=s[i];
	FD_SET(s[i],&rd);
    }

    err=pselect(max+1,&rd,NULL,NULL,NULL,&signal_sigmask);
    if (err<=0) {
	if (errno==EINTR) return ERR_OPC_SIGNAL_EVENT;
	return ERR_OPC_FE_SELECT;
    }
    
    for (i=0;i<number;i++)
	if (FD_ISSET(s[i],&rd)) {
	    err = s[i];
	    break;
	}
    
    err=accept(err,NULL,0);
    if (err<0) {
	if ((errno==EAGAIN)||(errno==EWOULDBLOCK)) goto select_again;
	return ERR_OPC_FE_ACCEPT;
    }
    *newsock=err;
    return 0;
}


int opcSignalRead(int s, void *buf, size_t len, int flags, unsigned long *rbytes) {
    socklen_t socksize=sizeof(struct sockaddrT);
    struct sockaddrT clisock;
    int err;
    fd_set rd;
    
    FD_ZERO(&rd);
    FD_SET(s,&rd);

select_again:    
    err=pselect(s+1,&rd,NULL,NULL,NULL,&signal_sigmask);
    if (err<=0) {
	if (errno==EINTR) return ERR_OPC_SIGNAL_EVENT;
	return ERR_OPC_FE_SELECT;
    }
    
    err=recv(s,buf,len,flags);
    if (err<0) {
	if ((errno==EAGAIN)||(errno==EWOULDBLOCK)) goto select_again;
	return ERR_OPC_FE_RECV;
    }
    *rbytes=err;
    return 0;
}

int opcSignalWrite(int s, const void *buf, size_t len, int flags, unsigned long *wbytes) {
    unsigned long writed = 0;
    socklen_t socksize=sizeof(struct sockaddrT);
    struct sockaddrT clisock;
    int err;
    fd_set rd;
    
    FD_ZERO(&rd);
    FD_SET(s,&rd);

select_again:    
    err=pselect(s+1,NULL,&rd,NULL,NULL,&signal_sigmask);
    if (err<=0) {
	if (errno==EINTR) {
	    if (wbytes) *wbytes = writed;
	    return ERR_OPC_SIGNAL_EVENT;
	}
	return ERR_OPC_FE_SELECT;
    }
    
    err=send(s,buf+writed,len-writed,flags);
    if (err<0) {
	if ((errno==EAGAIN)||(errno==EWOULDBLOCK)) goto select_again;
	return ERR_OPC_FE_SEND;
    } else writed += err;
    
    if (writed<len) goto select_again;
    
    if (wbytes) *wbytes=writed;
    return 0;
}
