#include <pthread.h>
#include <assert.h>

#include <opcerr.h>
#include <opcmalloc.h>
#include <opcdebug.h>

#include <thread.h>

void ThreadZeroContext(ThreadContext ctx) {
    assert(ctx);
    ctx->lock = NULL;
    ctx->cond = NULL;
    ctx->initialized = 0;
}

ThreadContext ThreadCreateContext(OPCMallocContext mctx) {
    ThreadContext ctx;

    ctx = (ThreadContext)opcMalloc(mctx,sizeof(struct ThreadContextT));
    opcDebugServer(OPC_DEBUG_NOTICE, "Thread %p: CreateContext", ctx);    
    if (ctx) {
	ctx->malloc=mctx;
	ThreadZeroContext(ctx);
    }
    return ctx;
}

int ThreadInitContext(ThreadContext ctx) {
    int err;
    assert(ctx);

    opcDebugServer(OPC_DEBUG_NOTICE, "Thread %p: InitContext", ctx);    
    
    ThreadZeroContext(ctx);
    
    ctx->lock = LockCreateContext(ctx->malloc);
    ctx->cond = ConditionCreateContext(ctx->malloc);
    if ((!ctx->lock)||(!ctx->cond)) {
	ThreadFreeContext(ctx);	
	return -1;
    }
    err = LockInitContext(ctx->lock);
    if (err) {
	ThreadFreeContext(ctx);
	return err;
    }
    
    err = ConditionInitContext(ctx->cond);
    if (err) {
	ThreadFreeContext(ctx);
	return err;
    }
    
    ctx->status = 0;
    ctx->initialized = 1;
    return 0;
}

int ThreadFreeContext(ThreadContext ctx) {
    int err;
    assert(ctx);

    opcDebugServer(OPC_DEBUG_NOTICE, "Thread %p: FreeContext",ctx);    

    if (ctx->initialized) err=ThreadTerminateAndWait(ctx);
    else err=0;
    if (ctx->lock) LockDestroyContext(ctx->lock);
    if (ctx->cond) ConditionDestroyContext(ctx->cond);
    ThreadZeroContext(ctx);
    return err;
}

int ThreadDestroyContext(ThreadContext ctx) {
    int err;
    assert(ctx);

    opcDebugServer(OPC_DEBUG_NOTICE, "Thread %p: DestroyContext",ctx);    

    err=ThreadFreeContext(ctx);
    opcFree(ctx->malloc,ctx);
    return err;
}

void ThreadSetStatus(ThreadContext ctx, dword bits) {
    LockContext lock;

    assert(ctx);

    lock=ctx->lock;

    LockWrite(lock);
    ctx->status=bits;
    LockReturn(lock);
    ConditionBroadcast(ctx->cond);
}

void ThreadSetStatusBits(ThreadContext ctx, dword bits) {
    LockContext lock;

    assert(ctx);

    lock=ctx->lock;

    LockWrite(lock);
    ctx->status|=bits;	    
    LockReturn(lock);
    ConditionBroadcast(ctx->cond);
}

void ThreadResetStatusBits(ThreadContext ctx, dword bits) {
    LockContext lock;
    assert(ctx);

    lock=ctx->lock;

    LockWrite(lock);
    ctx->status&=~bits;	    
    LockReturn(lock);

    ConditionBroadcast(ctx->cond);
}

int ThreadCheckStatusBitsOn(ThreadContext ctx, dword bits) {
    assert(ctx);
    return ((ctx->status&bits)==bits);
}

int ThreadCheckStatusBitsOff(ThreadContext ctx, dword bits) {
    assert(ctx);
    return ((ctx->status&bits)==0);
}

int ThreadCheckStatusBits(ThreadContext ctx, dword onbits, dword offbits) {
    assert(ctx);
    return (((ctx->status&offbits)==0)&&((ctx->status&onbits)==onbits));
}

int ThreadCheckStatusSomeBitsOn(ThreadContext ctx, dword bits) {
    assert(ctx);
    return ((ctx->status&bits)!=0);
}

int ThreadCheckStatusSomeBitsOff(ThreadContext ctx, dword bits) {
    assert(ctx);
    return ((ctx->status&bits)!=bits);
}

int ThreadCheckStatusSomeBits(ThreadContext ctx, dword onbits, dword offbits) {
    return (((ctx->status&offbits)!=offbits)&&((ctx->status&onbits)!=0));
}


void ThreadWaitStatusBitsOn(ThreadContext ctx, dword bits) {
    assert(ctx);
    ConditionPreLock(ctx->cond);
    while (!ThreadCheckStatusBitsOn(ctx,bits)) 
	ConditionWaitLocked(ctx->cond);
    ConditionUnLock(ctx->cond);
}

void ThreadWaitStatusBitsOff(ThreadContext ctx, dword bits) {
    assert(ctx);
    ConditionPreLock(ctx->cond);
    while (!ThreadCheckStatusBitsOff(ctx,bits)) 
	ConditionWaitLocked(ctx->cond);
    ConditionUnLock(ctx->cond);
}

void ThreadWaitStatusBits(ThreadContext ctx, dword onbits, dword offbits) {
    assert(ctx);
    ConditionPreLock(ctx->cond);
    while (!ThreadCheckStatusBits(ctx,onbits,offbits)) 
	ConditionWaitLocked(ctx->cond);
    ConditionUnLock(ctx->cond);
}

void ThreadWaitStatusSomeBitsOn(ThreadContext ctx, dword bits) {
    assert(ctx);
    ConditionPreLock(ctx->cond);
    while (!ThreadCheckStatusSomeBitsOn(ctx,bits)) 
	ConditionWaitLocked(ctx->cond);
    ConditionUnLock(ctx->cond);
}

void ThreadWaitStatusSomeBitsOff(ThreadContext ctx, dword bits) {
    assert(ctx);
    ConditionPreLock(ctx->cond);
    while (!ThreadCheckStatusSomeBitsOff(ctx,bits)) 
	ConditionWaitLocked(ctx->cond);
    ConditionUnLock(ctx->cond);
}

void ThreadWaitStatusSomeBits(ThreadContext ctx, dword onbits, dword offbits) {
    assert(ctx);
    ConditionPreLock(ctx->cond);
    while (!ThreadCheckStatusSomeBits(ctx,onbits,offbits)) 
	ConditionWaitLocked(ctx->cond);
    ConditionUnLock(ctx->cond);
}


int ThreadRun(ThreadContext ctx, ThreadFunctionT thread_function, void *attr) {
    int err;
    
    assert(ctx);
    assert(thread_function);
    assert(ctx->initialized);
    
    err = pthread_create(&ctx->thread,NULL,(void*(*)(void*))thread_function,attr);
    if (!err) {
	opcDebugServer(OPC_DEBUG_INFO,"Thread %p: Run %8.lx",ctx, ctx->thread);
	ThreadSetStatusBits(ctx,OPC_THREAD_RUNNING_FLAG|OPC_THREAD_JOINREQUIRED_FLAG);
    }
    return err;
}

void ThreadTerminate(ThreadContext ctx) {
    assert(ctx);
    ThreadSetStatusBits(ctx, OPC_THREAD_TERMINATE_FLAG);
}

int ThreadWaitTerminate(ThreadContext ctx) {
    int err;
    int *retcode;
    LockContext lock;
    
    assert(ctx);
    assert(ctx->initialized);
    
    lock=ctx->lock;

    LockWrite(ctx->lock);
    if (ctx->status&OPC_THREAD_JOINREQUIRED_FLAG==OPC_THREAD_JOINREQUIRED_FLAG) {
	ctx->status&=~OPC_THREAD_JOINREQUIRED_FLAG;
	LockReturn(lock);
	pthread_join(ctx->thread,(void**)&retcode);
	LockWrite(ctx->lock);
	ctx->status&=~OPC_THREAD_RUNNING_FLAG;
	LockReturn(lock);
	ConditionBroadcast(ctx->cond);

	if (retcode) {
	    if (*((void**)retcode)==PTHREAD_CANCELED) err=ERR_OPC_THREAD_CANCELED;
	    else err=*retcode;
	    free(retcode);
	} else err=ERR_OPC_THREAD_RETURNCODE;
	return err;
    } else {
	LockReturn(ctx->lock);
	ThreadWaitStatusBitsOff(ctx,OPC_THREAD_RUNNING_FLAG);
    }
    return 0;
}

int ThreadTerminateAndWait(ThreadContext ctx) {
    assert(ctx);
    ThreadTerminate(ctx);
    return ThreadWaitTerminate(ctx);
}


int ThreadIsRunning(ThreadContext ctx) {
    assert(ctx);
    return ThreadCheckStatusBitsOn(ctx, OPC_THREAD_RUNNING_FLAG);
}

int ThreadIsTerminated(ThreadContext ctx) {
    assert(ctx);
    return ThreadCheckStatusBitsOn(ctx, OPC_THREAD_TERMINATE_FLAG);
}

