#include <assert.h>

#include <hash-string.h>

#include <opcmalloc.h>

#include "../3rd_party/prime.h"

#include <doublehash.h>

static inline unsigned long DoubleHash1(DoubleHashContext ctx, unsigned long k) {
	return k%ctx->size;
}

static inline unsigned long DoubleHash2(DoubleHashContext ctx, unsigned long k) {
	return 1+k%(ctx->size-1);
}

#define find_prime findUpperPrime

static void DoubleHashZeroContext(DoubleHashContext ctx) {
    assert(ctx);
    ctx->buffer = NULL;
    ctx->max_size = 0;
}

DoubleHashContext DoubleHashCreateContext(OPCMallocContext mctx) {
    DoubleHashContext ctx;
    ctx = (DoubleHashContext)opcMalloc(mctx, sizeof(struct DoubleHashContextT));
    if (ctx) {
	ctx->mctx = mctx;
	DoubleHashZeroContext(ctx);
    }
    return ctx;
}

int DoubleHashInitContext(DoubleHashContext ctx, unsigned long size, unsigned long step_size, unsigned char free_on_clear, void (*cfunc)(void *data)) {
    unsigned long i;
    
    assert(ctx);
    
    DoubleHashZeroContext(ctx);

    ctx->gift = free_on_clear;
    ctx->cfunc = cfunc;

    ctx->size = find_prime(size);
    ctx->step_size = step_size;
    
    ctx->buffer = (void**)opcMalloc(ctx->mctx,ctx->size*sizeof(void*));
    if (!ctx->buffer) return 1;
    for (i=0;i<ctx->size;i++) ctx->buffer[i]=NULL;
    
    return 0;
}

void DoubleHashFreeContext(DoubleHashContext ctx) {
    unsigned long i;
    
    assert(ctx);
    
    if (ctx->buffer) {
	for (i=0;i<ctx->size;i++)
	    if (ctx->buffer[i]) DoubleHashRemove(ctx, i);
	opcFree(ctx->mctx, ctx->buffer);
    }

    DoubleHashZeroContext(ctx);
}

void DoubleHashDestroyContext(DoubleHashContext ctx) {
    assert(ctx);
    DoubleHashFreeContext(ctx);
    opcFree(ctx->mctx, ctx);
}

unsigned long DoubleHashInsert(DoubleHashContext ctx, void *data, unsigned long hash) {
    unsigned long i,pos;
    unsigned long size;
    void **buffer;
    
    assert(ctx);
    assert(ctx->buffer);
    
    for (i=0;i<ctx->size;i++) {
	pos=(DoubleHash1(ctx,hash)+i*DoubleHash2(ctx,hash))%ctx->size;
	if (!ctx->buffer[pos]) break;
    }
    
    if (i==ctx->size) {
	size = find_prime(ctx->size + ctx->step_size);
	if ((ctx->max_size)&&(size > ctx->max_size)) return (unsigned long)-2;
	
	buffer=(void**)opcRealloc(ctx->mctx,ctx->buffer,size*sizeof(void*));
	if (!buffer) {
	    ctx->buffer=buffer;
	    return (unsigned long)-1;
	}
	for (i=ctx->size+1;i<size;i++) ctx->buffer[i]=NULL;

	pos = ctx->size;

	ctx->buffer = buffer;
	ctx->size = size;
    }
    
    ctx->buffer[pos]=data;
    return pos;
}

unsigned long DoubleHashInsertS(DoubleHashContext ctx, void *data, const char *str) {
    unsigned long hash;
    
    assert(str);

    hash = hash_string(str);
    return DoubleHashInsert(ctx,data,hash);
}

void DoubleHashRemove(DoubleHashContext ctx, unsigned long pos) {
    assert(ctx);
    assert(ctx->buffer);
    assert(pos<ctx->size);
    
    if (ctx->gift) {
	if (ctx->cfunc) ctx->cfunc(ctx->buffer[pos]);
	else opcFree(ctx->mctx, ctx->buffer[pos]);
    } 
    ctx->buffer[pos]=NULL;
}

void DoubleHashUnlink(DoubleHashContext ctx, unsigned long pos) {
    assert(ctx);
    assert(ctx->buffer);
    assert(pos<ctx->size);
    
    ctx->buffer[pos]=NULL;    
}

const void *DoubleHashGetBuffer(DoubleHashContext ctx) {
    assert(ctx);

    return ctx->buffer;
}

const void *DoubleHashGetLink(DoubleHashContext ctx, unsigned long pos) {
    assert(ctx);
    assert(ctx->buffer);
    assert(pos<ctx->size);

    return ctx->buffer[pos];
}

void *DoubleHashGet(DoubleHashContext ctx, unsigned long pos) {
    void *item;
    
    assert(ctx);
    assert(ctx->buffer);
    assert(pos<ctx->size);

    item = ctx->buffer[pos];
    ctx->buffer[pos]=NULL;
    
    return item;
}

int DoubleHashSetMaximumSize(DoubleHashContext ctx, unsigned long max_size) {
    assert(ctx);
    
    ctx->max_size=max_size;
    return 0;
}

unsigned long DoubleHashClean(DoubleHashContext ctx, DoubleHashCleanFunctionT clean_func, void *attr) {
    int res;
    unsigned long pos;
    unsigned long size;
    unsigned long cleaned=0;

    assert(ctx);
    
    size = ctx->size;
    for (pos=0;pos<size;pos++) {
	if (ctx->buffer[pos]) {
	    res=clean_func(ctx->buffer[pos], attr);
	    if (res) {
		if ((res>1)&&(ctx->gift)) {
		    if (ctx->cfunc) ctx->cfunc(ctx->buffer[pos]);
		    else opcFree(ctx->mctx, ctx->buffer[pos]);
		} 
		ctx->buffer[pos]=NULL;
		cleaned++;
	    }
	}
    }
    return cleaned;
}
