#include <stdio.h>
#include <string.h>
#include <assert.h>

#include <opcmalloc.h>
#include <hashtable.h>

#include <hash-string.h>

#define mod1(x) ((x) - (long)(x))

static const char *HashTableDefaultPointer(const HashStringPtr ctx) {
    return ctx->str;
}

static const char *HashTableDefaultArray(const HashStringArray ctx) {
    return ctx->str;
}

static unsigned long HashTableDefaultHash(const HashStringHash ctx) {
    return ctx->hash;
}

HashTable HashTableInit(OPCMallocContext mctx, unsigned long size, unsigned char hash_type, HashTableStringT string_func) {
    HashTable ctx;
    if (hash_type>HASH_TYPE_LAST) return NULL;
    if (((hash_type==HASH_TYPE_FUNCTION)||(hash_type==HASH_TYPE_HASH_FUNCTION))&&(!string_func)) return NULL;
    
    ctx = (HashTable)opcMalloc(mctx, sizeof(struct HashTableT));
    if (ctx) {
	ctx->malloc=mctx;
	ctx->size=size;
	ctx->mode=hash_type;
	ctx->gift=0;
	ctx->cfunc=NULL;
	ctx->prefunc=NULL;
	ctx->postfunc=NULL;
	ctx->ppfuncattr=NULL;
	switch(hash_type) {
	    case HASH_TYPE_HASH:
		ctx->string_func=(HashTableStringT)&HashTableDefaultHash;
	    break;
	    case HASH_TYPE_ARRAY:
		ctx->string_func=(HashTableStringT)&HashTableDefaultArray;
	    break;
	    case HASH_TYPE_POINTER:
	    case HASH_TYPE_STATICPOINTER:
	    	ctx->string_func=(HashTableStringT)&HashTableDefaultPointer;
	    break;
	    default:
		ctx->string_func=string_func;
	}
	ctx->bucket=(HashTableBucket*)opcCalloc(mctx,size,sizeof(HashTableBucket));
	if (!ctx->bucket) {
	    opcFree(mctx,ctx);
	    return NULL;
	}
    }
    return ctx;
}

void HashTableSetDestructor(HashTable ctx, HashTableDestructorT cfunc) {
    assert(ctx);
    ctx->gift = 1;
    ctx->cfunc = cfunc;
}

void HashTableSetPPDFunctions(HashTable ctx, HashTablePPDFunctionT prefunc, HashTablePPDFunctionT postfunc, void *attr) {
    assert(ctx);
    
    ctx->prefunc=prefunc;
    ctx->postfunc=postfunc;
    ctx->ppfuncattr=attr;
}

void HashTableFree(HashTable ctx) {
    HashTableBucket bucket;
    HashTableBucket tmpbucket;
    unsigned long i;

    assert(ctx);

    if (ctx->prefunc) ctx->prefunc(ctx,HASHTABLE_DESTRUCTIONMODE_ALL,ctx->ppfuncattr);

    for (i=0;i<ctx->size;i++) {
	for (bucket=ctx->bucket[i],ctx->bucket[i]=NULL;bucket;) {
	    if (ctx->gift) {
		if (ctx->cfunc) {
		    ctx->cfunc(bucket->data);
		} else {
		    if (ctx->mode==HASH_TYPE_POINTER)
			opcFree(ctx->malloc, ((HashStringPtr)(bucket->data))->str);
		    opcFree(ctx->malloc, bucket->data);
		}
	    }    
	    tmpbucket=bucket;
	    bucket=bucket->next;
	    opcFree(ctx->malloc, tmpbucket);
	}
    }
    opcFree(ctx->malloc, ctx->bucket);

    if (ctx->postfunc) ctx->postfunc(ctx,HASHTABLE_DESTRUCTIONMODE_ALL,ctx->ppfuncattr);
}

void HashTableDestroy(HashTable ctx) {
    HashTableFree(ctx);
    opcFree(ctx->malloc, ctx);
}

int HashTableInsert(HashTable ctx, void *data) {
    unsigned long pos;
    HashTableBucket bucket;
    
    assert(ctx);
    assert(data);
    
    bucket = (HashTableBucket)opcMalloc(ctx->malloc, sizeof(struct HashTableBucketT));
    if (!bucket) return 1;

    if (ctx->mode<HASH_TYPE_HASH) bucket->hash=hash_string(ctx->string_func(data));
    else bucket->hash=((HashTableHashT)(ctx->string_func))(data);

    pos = ctx->size * mod1( hash_table_coef * bucket->hash );
    bucket->next = ctx->bucket[pos];
    bucket->data = data;
    ctx->bucket[pos]=bucket;
    return 0;
}

int HashTableAdd(HashTable ctx, void *data, unsigned char report) {
    int err;
    unsigned long hash;
    unsigned long pos;
    HashTableBucket bucket;
    const char *str;
    
    assert(ctx);
    assert(data);

    if (ctx->mode<HASH_TYPE_HASH) {
	str = ctx->string_func(data);
	hash = hash_string(str);
    } else hash=((HashTableHashT)(ctx->string_func))(data);

    pos = ctx->size * mod1( hash_table_coef * hash );
    for (bucket = ctx->bucket[pos];bucket;bucket=bucket->next) {
	if (bucket->hash==hash) {
	    if (ctx->mode<HASH_TYPE_HASH) {
		if (!strcmp(str,ctx->string_func(bucket->data))) err = 1;
		else err = 0;
	    } else err = 1;
		
	    if (err) {
		if (report) return 1;
	        if (ctx->gift) {
	    	    if (ctx->cfunc) ctx->cfunc(data);
		    else {
		        if (ctx->mode==HASH_TYPE_POINTER) opcFree(ctx->malloc, (char*)str);
		        opcFree(ctx->malloc, data);
		    }
		}
		return 0;
	    }
	}
    }

    bucket = (HashTableBucket)opcMalloc(ctx->malloc, sizeof(struct HashTableBucketT));
    if (!bucket) return 1;

    bucket->hash = hash;
    bucket->next = ctx->bucket[pos];
    bucket->data = data;
    ctx->bucket[pos]=bucket;
    
    return 0;
}

static int HashTableDeleteMS(HashTable ctx, const char *string, unsigned char multi) {
    int err,res=0;
    unsigned long hash;
    unsigned long pos;
    HashTableBucket bucket;
    HashTableBucket prebucket = NULL;
    
    assert(ctx);
    assert(string);
    
    hash = hash_string(string);
    pos = ctx->size * mod1( hash_table_coef * hash );
    for (bucket = ctx->bucket[pos];bucket;) {
	if (bucket->hash==hash) {
	    if (ctx->mode<HASH_TYPE_HASH) {
		if (!strcmp(string,ctx->string_func(bucket->data))) err = 1;
		else err = 0;
	    } else err = 1;

	    if (err) {
		if (prebucket) prebucket->next = bucket->next;
		else ctx->bucket[pos]=bucket->next;
		if (ctx->gift) {
		    if (ctx->cfunc) ctx->cfunc(bucket->data);
		    else {
			if (ctx->mode==HASH_TYPE_POINTER) opcFree(ctx->malloc, ((HashStringPtr)(bucket->data))->str);
			opcFree(ctx->malloc, bucket->data);
		    }
		}
		opcFree(ctx->malloc, bucket);
		if (multi) {
		    if (prebucket) bucket=prebucket->next;
		    else bucket=ctx->bucket[pos];
		    res++;
		    continue;
		}
		else return 1;
	    }
	}
	prebucket=bucket;
	bucket=bucket->next;
    }
    return res;
}

int HashTableDeleteMulti(HashTable ctx, const char *string) {
    int res;
    
    assert(ctx);
    
    if (ctx->prefunc) ctx->prefunc(ctx,HASHTABLE_DESTRUCTIONMODE_MULTI,ctx->ppfuncattr);
    res=HashTableDeleteMS(ctx,string,1);
    if (ctx->postfunc) ctx->postfunc(ctx,HASHTABLE_DESTRUCTIONMODE_MULTI,ctx->ppfuncattr);
    return res;
}

int HashTableDelete(HashTable ctx, const char *string) {
    int res;
    
    assert(ctx);
    
    if (ctx->prefunc) ctx->prefunc(ctx,HASHTABLE_DESTRUCTIONMODE_SINGLE,ctx->ppfuncattr);
    res=HashTableDeleteMS(ctx,string,0);
    if (ctx->postfunc) ctx->postfunc(ctx,HASHTABLE_DESTRUCTIONMODE_SINGLE,ctx->ppfuncattr);
    return res;
}

static int HashTableRemoveOrUnlink(HashTable ctx, const void *data, unsigned char  remove) {
    unsigned long hash;
    unsigned long pos;
    HashTableBucket bucket;
    HashTableBucket prebucket = NULL;
    const char *str;

    assert(ctx);
    assert(data);
    
    if (ctx->mode<HASH_TYPE_HASH) {
	str = ctx->string_func(data);
	hash = hash_string(str);
    } else hash=((HashTableHashT)(ctx->string_func))(data);

    pos = ctx->size * mod1( hash_table_coef * hash );

    for (bucket = ctx->bucket[pos];bucket;bucket = bucket->next) {
	if (bucket->data==data) {
	    if (prebucket) prebucket->next = bucket->next;
	    else ctx->bucket[pos]=bucket->next;
	    if ((ctx->gift)&&(remove)) {
		if (ctx->cfunc) {
		    ctx->cfunc(bucket->data);
		} else {
		    if (ctx->mode==HASH_TYPE_POINTER)
			opcFree(ctx->malloc, ((HashStringPtr)(bucket->data))->str);
		    opcFree(ctx->malloc, bucket->data);
		}
	    }    
	    opcFree(ctx->malloc, bucket);
	    return 1;
	}
	prebucket=bucket;
    }
    return 0;
}

int HashTableUnlink(HashTable ctx, const void *data) {
    return HashTableRemoveOrUnlink(ctx,data,0);
}

int HashTableRemove(HashTable ctx, const void *data) {
    int res;
    
    assert(ctx);
    
    if (ctx->prefunc) ctx->prefunc(ctx,HASHTABLE_DESTRUCTIONMODE_SINGLE,ctx->ppfuncattr);
    res=HashTableRemoveOrUnlink(ctx,data,1);
    if (ctx->postfunc) ctx->postfunc(ctx,HASHTABLE_DESTRUCTIONMODE_SINGLE,ctx->ppfuncattr);
    return res;
}

void *HashTableFindNext(HashTable ctx, const char *string, const void *start) {
    unsigned long hash;
    unsigned long pos;
    HashTableBucket bucket;
    
    assert(ctx);
    assert(string);
    assert(ctx->mode<HASH_TYPE_HASH);
    
    hash = hash_string(string);
    pos = ctx->size * mod1( hash_table_coef * hash );
    for (bucket = ctx->bucket[pos];bucket;bucket=bucket->next) {
	if (start) {
	    if (bucket->data!=start) continue;
	    else start=NULL;
	}
	if ((bucket->hash==hash)&&(!strcmp(string,ctx->string_func(bucket->data)))) return bucket->data;
    }
    return NULL;
}

void *HashTableFind(HashTable ctx, const char *string) {
    return HashTableFindNext(ctx,string,NULL);
}

void *HashTableHashFindNext(HashTable ctx, unsigned long hash, const void *start) {
    unsigned long pos;
    HashTableBucket bucket;
    
    assert(ctx);
    
    pos = ctx->size * mod1( hash_table_coef * hash );
    for (bucket = ctx->bucket[pos];bucket;bucket=bucket->next) {
	if (start) {
	    if (bucket->data!=start) continue;
	    else start=NULL;
	}
	if (bucket->hash==hash) return bucket->data;
    }
    return NULL;
}

void *HashTableHashFind(HashTable ctx, unsigned long hash) {
    return HashTableHashFindNext(ctx,hash,NULL);
}

int HashTableWalk(HashTable ctx, HashTableWalkT func, void *attr) {
    HashTableBucket bucket;
    HashTableBucket next;
    unsigned long i;
    int err;

    for (i=0;i<ctx->size;i++) {
	for (bucket=ctx->bucket[i];bucket;) {
	    next=bucket->next;
	    err=func(bucket->data, attr); /* Can remove current element */
	    if (err) return err;
	    bucket=next;
	}
    }
    return 0;
}

int HashTableClean(HashTable ctx, HashTableCleanT func) {
    HashTableBucket prebucket;
    HashTableBucket bucket;
    HashTableBucket next;
    unsigned long i;
    int err;

    for (i=0;i<ctx->size;i++) {
	prebucket = NULL;
	for (bucket=ctx->bucket[i];bucket;) {
	    next=bucket->next;
	    if (func(bucket->data)) {
		if (ctx->gift) {
		    if (ctx->cfunc) ctx->cfunc(bucket->data);
		    else {
			if (ctx->mode==HASH_TYPE_POINTER)
			    opcFree(ctx->malloc, ((HashStringPtr)(bucket->data))->str);
			opcFree(ctx->malloc, bucket->data);
		    }
		}
		opcFree(ctx->malloc, bucket);
		
		if (prebucket) prebucket->next = next;
		else ctx->bucket[i] = next;
	    } else prebucket = bucket;
	    bucket=next;
	}
    }
    return 0;
}
