#include <assert.h>

#include <opcerr.h>
#include <opcenumeration.h>

inline static void opcEnumerationZeroContext(OPCEnumerationContext ctx) {
    ctx->hash = NULL;
    ctx->destroy_bucket = 0;
    ctx->string = NULL;
}

OPCEnumerationContext opcEnumerationCreateContext(OPCMallocContext mctx) {
    OPCEnumerationContext ctx;

    ctx = (OPCEnumerationContext)opcMalloc(mctx, sizeof(struct OPCEnumerationContextT));
    if (ctx) {
	ctx->malloc = mctx;
	opcEnumerationZeroContext(ctx);
    }
    return ctx;
}

int opcEnumerationInitContext(OPCEnumerationContext ctx, OPCEnumerationBucket bucket) {
    int err = 0;
    unsigned long i;

    assert(ctx);
    assert(bucket);

    opcEnumerationZeroContext(ctx);
    
    for (i=0;bucket[i].string;i++);
    ctx->maxid=i;
    
    ctx->hash = HashTableInit(ctx->malloc, 1.5*i, HASH_TYPE_POINTER, NULL);
    if (!ctx->hash) return ERR_OPC_MEM;
    
    for (i=0;bucket[i].string;i++)
	err+=HashTableInsert(ctx->hash, &bucket[i]);

    if (err) {
	opcEnumerationFreeContext(ctx);
	return ERR_OPC_MEM;
    }
    
    return 0;
}

int opcEnumerationInitContextFromArray(OPCEnumerationContext ctx, const char **title_array, unsigned char copy_strings) {
    int err = 0;
    OPCEnumerationBucket bucket;
    unsigned long i;
    unsigned long size;

    assert(ctx);
    assert(title_array);

    opcEnumerationZeroContext(ctx);
    
    for (size=0;title_array[size];size++);
    ctx->maxid=size;
    
    if (copy_strings) 
	ctx->hash = HashTableInit(ctx->malloc, 1.5*size, HASH_TYPE_POINTER, NULL);
    else
	ctx->hash = HashTableInit(ctx->malloc, 1.5*size, HASH_TYPE_STATICPOINTER, NULL);

    if (!ctx->hash) return ERR_OPC_MEM;
    
    bucket = (OPCEnumerationBucket)opcMalloc(ctx->malloc, size*sizeof(struct OPCEnumerationBucketT));
    if (!bucket) {
	opcEnumerationFreeContext(ctx);
	return ERR_OPC_MEM;
    }
    
    ctx->bucket = bucket;
    ctx->destroy_bucket = 1;
    
    for (i=0;i<size;i++) {
	bucket[i].id = i;
	if (copy_strings) {
	    bucket[i].string = NULL;
	    opcExchangeString(ctx->malloc, &bucket[i].string, title_array[i]);
	} else bucket[i].string = (char*)title_array[i];
	err+=HashTableInsert(ctx->hash, &bucket[i]);
    }

    if (err) {
	opcEnumerationFreeContext(ctx);
	return ERR_OPC_MEM;
    }
    
    return 0;    
}

int opcEnumerationInitContextFromString(OPCEnumerationContext ctx, const char *titles, unsigned char copy_string) {
    int err;
    OPCEnumerationBucket bucket;
    unsigned long i,size;
    char *str = NULL;
    
    assert(ctx);
    assert(titles);

    opcEnumerationZeroContext(ctx);
    
    if (copy_string) {
	err = opcExchangeString(ctx->malloc,&str,titles);
	if (err) return ERR_OPC_MEM;
	ctx->string = str; 
    } else {
	str = (char*)titles;
    }
    
    if ((str[0]==',')||(str[0]==0)) {
	opcEnumerationFreeContext(ctx);
	return ERR_OPC_INVALID_ENUMERATION_DESCRIPTION;
    }
    
    for(size=1,i=1;str[i];i++) 
	if (str[i]==',') {
	    if (str[i-1]==',') break;
	    size++;
	} else if ((str[i]=='<')||(str[i]=='&')||(str[i]=='"')) break;
    
    if ((str[i])||(str[i-1]==',')) {
	opcEnumerationFreeContext(ctx);
	return ERR_OPC_INVALID_ENUMERATION_DESCRIPTION;
    }
    
    ctx->maxid = size;


    ctx->hash = HashTableInit(ctx->malloc, 1.5*size, HASH_TYPE_STATICPOINTER, NULL);
    if (!ctx->hash) return ERR_OPC_MEM;
    HashTableSetDestructor(ctx->hash, NULL);

    bucket = (OPCEnumerationBucket)opcMalloc(ctx->malloc, size*sizeof(struct OPCEnumerationBucketT));
    if (!bucket) {
	opcEnumerationFreeContext(ctx);
	return ERR_OPC_MEM;
    }

    ctx->bucket = bucket;
    ctx->destroy_bucket = 1;
    
    bucket[0].string = str;
    bucket[0].id = 0;
    
    for (size=1,i=1;str[i];i++) {
	if (str[i]==',') {
	    bucket[size].id=size;
	    bucket[size++].string=str+i+1;
	    str[i]=0;
	}
    }

    return 0;
}

void opcEnumerationFreeContext(OPCEnumerationContext ctx) {
    assert(ctx);
    if (ctx->hash) HashTableDestroy(ctx->hash);
    if ((ctx->destroy_bucket)&&(ctx->bucket)) opcFree(ctx->malloc, ctx->bucket);
    if (ctx->string) opcExchangeString(ctx->malloc, ctx->string);
    opcEnumerationZeroContext(ctx);
}

void opcEnumerationDestroyContext(OPCEnumerationContext ctx) {
    opcEnumerationFreeContext(ctx);
    opcFree(ctx->malloc,ctx);
}

unsigned long opcEnumerationGetID(OPCEnumerationContext ctx, const char *str) {
    OPCEnumerationBucket bucket;
    
    assert(ctx);
    assert(str);
    assert(ctx->hash);
    
    bucket = HashTableFind(ctx->hash,str);
    if (!bucket) return (unsigned long)-1;
    return bucket->id;
    
}

const char *opcEnumerationGetString(OPCEnumerationContext ctx, unsigned long number) {
    assert(ctx);
    assert(number<ctx->maxid);
    assert(ctx->bucket);
    
    return ctx->bucket[number].string;
}

OPCEnumerationBucket opcEnumerationGetBucketByString(OPCEnumerationContext ctx, const char *str) {
    OPCEnumerationBucket bucket;
    
    assert(ctx);
    assert(str);
    assert(ctx->hash);
    
    return (OPCEnumerationBucket)HashTableFind(ctx->hash,str);
}

OPCEnumerationBucket opcEnumerationGetBucketByID(OPCEnumerationContext ctx, unsigned long number) {
    assert(ctx);
    assert(number<ctx->maxid);
    assert(ctx->bucket);
    
    return ctx->bucket+number;
}
