#define _FRONTEND_C

#include <stdio.h>
#include <sys/types.h>
#include <sys/wait.h>
#include <sys/stat.h>
#include <unistd.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <string.h>
#include <fcntl.h>
#include <pthread.h>
#include <signal.h>
#include <errno.h>
#include <assert.h>

#include <opcerr.h>
#include <opcconfig.h>

#include <opcdebug.h>
#include <lock.h>

#include <opctools.h>

#include <opcrequestpool.h>

#include <frontend.h>
#include <frontendchild.h>
#include <opcsignal.h>

OPCFrontendContext opcFrontendCreateContext() {
    unsigned long i;
    OPCFrontendContext ctx;

    opcDebugServer(OPC_DEBUG_NOTICE, "opcFrontendCreateContext");    

    ctx = (OPCFrontendContext)malloc(sizeof(struct OPCFrontendContextT));
    if (ctx) {
	ctx->socket=-1;
	ctx->isocket=-1;
	ctx->buffpool=NULL;
	ctx->rpctx=NULL;
	ctx->sigctx=NULL;
	ctx->lock=NULL;
	for (i=0;i<max_frontend_childs;i++) ctx->child[i]=NULL;
    }
    return ctx;    
}

void opcFrontendFreeContext(OPCFrontendContext ctx) {
    unsigned long i;
    assert(ctx);

    opcDebugServer(OPC_DEBUG_NOTICE, "opcFrontendFreeContext");    

	// Return all used buffers back to buffer pool (skiped DS)
    ctx->buffpool=NULL;
    
    if (ctx->socket>0) close(ctx->socket);
    if (ctx->isocket>0) close(ctx->isocket);
    unlink(saddr);
    
    ctx->rpctx=NULL;
    ctx->sigctx=NULL;
    
    if (ctx->lock) LockDestroyContext(ctx->lock);
}

void opcFrontendDestroyContext(OPCFrontendContext ctx) {
    assert(ctx);

    opcDebugServer(OPC_DEBUG_NOTICE, "opcFrontendDestroyContext");    

    opcFrontendFreeContext(ctx);
    free(ctx);
}

int opcFrontendInitContext(OPCFrontendContext ctx, OPCRequestPoolContext rpctx, BufferPool bp, OPCSignalContext sigctx) {
    int is;
    int s;
    int flags;
    int err;
    struct sockaddrT mysock;
    struct sockaddr_in isock;

    assert(ctx);
    assert(rpctx);    
    assert(bp);
    assert(sigctx);

    opcDebugServer(OPC_DEBUG_NOTICE, "opcFrontendInitContext");    
    
	// Creating unix domain listening socket
    unlink(saddr);

    s=socket(PF_UNIX,SOCK_STREAM,0);
    if (!s) return ERR_OPC_FE_SOCKET;

    flags=fcntl(s,F_GETFL,0);
    if ((flags<0)||(fcntl(s,F_SETFL,flags|O_NONBLOCK)<0)) {
	close(s);
	return ERR_OPC_FE_FCNTL_NONBLOCK;
    }
    
    mysock.sa_family=AF_UNIX;
    strcpy(mysock.sa_data,saddr);

    if (bind(s,(struct sockaddr*)&mysock,3+sizeof(saddr))==-1) {
	close(s);
	return ERR_OPC_FE_BIND;
    }

    if (listen(s,1)<0) {
	unlink(saddr);
	close(s);
	return ERR_OPC_FE_LISTEN;	
    }

    if (chmod(saddr,00777)<0) {
	unlink(saddr);
	close(s);
	return ERR_OPC_FE_PERMS;
    }


	// Creating TCP/IP listening socket (planed)
#ifdef FRONTEND_NETWORK_PORT    
    is=socket(PF_INET,SOCK_STREAM,IPPROTO_TCP);
    if (!is) {
	unlink(saddr);
	close(s);
	return ERR_OPC_FE_NETWORK_SOCKET;
    }
    
    isock.sin_family=AF_INET;
    isock.sin_port=htons(FRONTEND_NETWORK_PORT);
    isock.sin_addr.s_addr=htonl(INADDR_ANY);

    if (bind(is,(struct sockaddr*)&isock,sizeof(struct sockaddr_in))==-1) {
	unlink(saddr);
	close(s);
	close(is);
	return ERR_OPC_FE_NETWORK_BIND;
    }
    
    if (listen(is,1)<0) {
	unlink(saddr);
	close(s);
	close(is);
	return ERR_OPC_FE_NETWORK_LISTEN;
    }

    ctx->isocket=is;
#endif /* FRONTEND_NETWORK_PORT */

    ctx->socket=s;

	// Initialising Pool of Buffers (Get own, if not supported DS)
    ctx->buffpool=bp;
    ctx->rpctx=rpctx;
    ctx->sigctx=sigctx;

    ctx->lock = LockCreateContext(NULL);
    if (!ctx->lock) {
	opcFrontendFreeContext(ctx);
	return ERR_OPC_MEM;
    }
    
    err = LockInitContext(ctx->lock);
    if (err) {
	opcFrontendFreeContext(ctx);
	return err;
    }

    return 0;
}


int *opcFrontendListen(OPCFrontendContext ctx) {
    int s;
    int n = 0, sock[2];
    OPCFrontendChildContext child_ctx;
    char tmp_buffer[opc_reqresp_buffer_size];
    unsigned long i = 0, j;
    uint32_t answer;
    sigset_t set;
    int err = 0;
    int flags;

    assert(ctx);
    assert(ctx->sigctx);
    assert(ctx->rpctx);
    assert(ctx->buffpool);
    
    if (ctx->socket) sock[n++] = ctx->socket;
    if (ctx->isocket) sock[n++] = ctx->isocket;

    if (!n) return ThreadReturnInt(NULL,ERR_OPC_NOTINITIALIZED);

    opcSignalRegister(ctx->sigctx, OPC_SIGNAL_TERMINATE);    
    
    while (!opcSignalIsTerminated(ctx->sigctx)) {
	    // Accepting connection
	err=opcSignalAcceptMulti(n,sock,&s);
	if (err) {
	    if (err==ERR_OPC_SIGNAL_EVENT) continue;
	    err = ERR_OPC_FE_ACCEPT;
	    break;
	}
//	s=accept(ctx->socket,(struct sockaddr*)&clisock,&socksize);
	
	    // Setting NONBLOCKING mode
	flags=fcntl(s,F_GETFL,0);
	if ((flags<0)||(fcntl(s,F_SETFL,flags|O_NONBLOCK)<0)) {
	    close(s);
	    err = ERR_OPC_FE_FCNTL_NONBLOCK;
	    break;
	}

	LockWrite(ctx->lock);
	
	    // Finding first free child slot
	for (j=++i;i<max_frontend_childs;i++) {
	    if (!ctx->child[i]) { flags = 0; break; }
	    else if (opcFrontendChildIsReady(ctx->child[i])) { flags=1; break; }
	    else if (opcFrontendChildIsForClean(ctx->child[i])) { flags = 2; break; }
	}
	
	if (i==max_frontend_childs) {
	    for (i=0;i<j;i++) {
		if (!ctx->child[i]) { flags = 0; break; }
	        else if (opcFrontendChildIsReady(ctx->child[i])) { flags=1; break; }
		else if (opcFrontendChildIsForClean(ctx->child[i])) { flags = 2; break; }
	    }
	    
	    if (i==j) {
		LockReturn(ctx->lock);
		    /* Wait a bit here */
		j=recv(s,&tmp_buffer,opc_reqresp_buffer_size,0);
		if ((j>3)&&((*(uint32_t*)tmp_buffer)==OPC_FECMD_NEWREQ)) {
		    answer = htonl(OPC_FECMD_BUSY);
		    send(s,&answer,sizeof(uint32_t),0);
		}
		close(s);
		continue;
	    }
	}
	
	switch(flags) {
	    case 0:
		// Creating child ctx
		child_ctx=opcFrontendChildCreateContext(s);
		if (!child_ctx) {
		    LockReturn(ctx->lock);
		    close(s);
		    err = ERR_OPC_FE_MEM;
		    break;
		}
	    break;
	    case 2:
		child_ctx = ctx->child[i];
		opcFrontendChildFreeContext(child_ctx);
	}
	
	err=opcFrontendChildInitContext(child_ctx,ctx,s);
	if (err) {
	    LockReturn(ctx->lock);
	    close(s);
	    opcFrontendChildDestroyContext(child_ctx);
	    break;
	}
	
	    // Spawning 2 processes per child
	if (pthread_create(&child_ctx->reader_pthread,NULL,(pthread_routine)&opcFrontendChildReader,child_ctx)) {
	    LockReturn(ctx->lock);
	    opcFrontendChildDestroyContext(child_ctx); // In other cases will selfdestroy
	    err = ERR_OPC_FE_CHILDSPAWN;
	    break;
	}

	if (pthread_create(&child_ctx->parser_pthread,NULL,(pthread_routine)&opcFrontendChildParser,child_ctx)) {
	    LockReturn(ctx->lock);
	    opcFrontendChildDestroyContext(child_ctx);
	    err = ERR_OPC_FE_CHILDSPAWN;
	    break;
	}
	
	ctx->child[i]=child_ctx;
	LockReturn(ctx->lock);
    }

	// Terminating and cleaning childs

    for (i=0;i<max_frontend_childs;i++) {
	if (ctx->child[i]) 
	    opcFrontendChildTerminate(ctx->child[i]);
    }    

    for (i=0;i<max_frontend_childs;i++) {
	if (ctx->child[i]) {
	    opcFrontendChildDestroyContext(ctx->child[i]);
	    ctx->child[i]=NULL;
	}
    }    

    return ThreadReturnInt(NULL,err);
}


void opcFrontendCleanupChilds(OPCFrontendContext ctx) {
    unsigned long i;
    assert(ctx);
    
    LockWrite(ctx->lock);
    
    for (i=0;i<max_frontend_childs;i++) {
	if ((ctx->child[i])&&(opcFrontendChildIsForClean(ctx->child[i]))) {
	    opcFrontendChildFreeContext(ctx->child[i]);
/*	    opcFrontendChildDestroyContext(ctx->child[i]);
	    ctx->child[i]=NULL;*/
	}
    }    
    
    LockReturn(ctx->lock);
}
