#include <stdio.h>
#include <netinet/in.h>
#include <pthread.h>
#include <assert.h>
#include <errno.h>

#include <opcerr.h>
#include <opcconfig.h>
#include <opcdebug.h>
#include <opcsignal.h>
#include <opctools.h>
#include <opcerror.h>
#include <dime.h>

#include <opcrequestpool.h>
#include <opcrequest.h>

#include <frontendchild.h>
#include <bufferpool.h>

void opcFrontendChildSetStatus(OPCFrontendChildContext ctx, unsigned long bits) {
    assert(ctx);

    pthread_mutex_lock(&ctx->status_mutex);
    ctx->status=bits;
    pthread_mutex_unlock(&ctx->status_mutex);
    pthread_cond_broadcast(&ctx->status_cond);
}

void opcFrontendChildSetStatusBits(OPCFrontendChildContext ctx, unsigned long bits) {
    assert(ctx);

    pthread_mutex_lock(&ctx->status_mutex);
    ctx->status|=bits;
    pthread_mutex_unlock(&ctx->status_mutex);
    pthread_cond_broadcast(&ctx->status_cond);
}

void opcFrontendChildResetStatusBits(OPCFrontendChildContext ctx, unsigned long bits) {
    assert(ctx);

    pthread_mutex_lock(&ctx->status_mutex);
    ctx->status&=~bits;
    pthread_mutex_unlock(&ctx->status_mutex);
    pthread_cond_broadcast(&ctx->status_cond);
}

int opcFrontendChildCheckStatusBitsOn(OPCFrontendChildContext ctx, unsigned long bits) {
    assert(ctx);
    return ((ctx->status&bits)==bits);
}

int opcFrontendChildCheckStatusBitsOff(OPCFrontendChildContext ctx, unsigned long bits) {
    assert(ctx);
    return ((ctx->status&bits)==0);
}

void opcFrontendChildWaitStatusBitsOn(OPCFrontendChildContext ctx, unsigned long bits) {
    assert(ctx);
    while (!opcFrontendChildCheckStatusBitsOn(ctx,bits))
	pthread_cond_wait(&ctx->status_cond,&ctx->status_cond_mutex);
}

void opcFrontendChildWaitStatusBitsOff(OPCFrontendChildContext ctx, unsigned long bits) {
    assert(ctx);
    while (!opcFrontendChildCheckStatusBitsOff(ctx,bits))
	pthread_cond_wait(&ctx->status_cond,&ctx->status_cond_mutex);
}

void opcFrontendChildTerminate(OPCFrontendChildContext ctx) {
    assert(ctx);
    
    opcFrontendChildSetStatusBits(ctx,OPC_FRONTEND_CHILD_STATUS_TERMINATE);

    if (ctx->status&OPC_FRONTEND_CHILD_STATUS_READER_RUNING)
	opcSignalSendTo(ctx->reader_pthread, OPC_SIGNAL_TERMINATE);
    if (ctx->status&OPC_FRONTEND_CHILD_STATUS_PARSER_RUNING)
	opcSignalSendTo(ctx->parser_pthread, OPC_SIGNAL_TERMINATE);

    pthread_cond_broadcast(&ctx->lbp_cond);
}

void opcFrontendChildWaitTerminate(OPCFrontendChildContext ctx) {
    assert(ctx);
    opcFrontendChildWaitStatusBitsOn(ctx,
			OPC_FRONTEND_CHILD_STATUS_READER_TERMINATED|
			OPC_FRONTEND_CHILD_STATUS_PARSER_TERMINATED);
}

int opcFrontendChildIsTerminate(OPCFrontendChildContext ctx) {
    assert(ctx);
    return opcFrontendChildCheckStatusBitsOn(ctx,
			OPC_FRONTEND_CHILD_STATUS_READER_TERMINATED|
			OPC_FRONTEND_CHILD_STATUS_PARSER_TERMINATED);
}

int opcFrontendChildIsForClean(OPCFrontendChildContext ctx) {
    if (opcFrontendChildCheckStatusBitsOn(ctx,OPC_FRONTEND_CHILD_STATUS_READY)) return 0;
    return opcFrontendChildCheckStatusBitsOn(ctx,
			OPC_FRONTEND_CHILD_STATUS_READER_TERMINATED|
			OPC_FRONTEND_CHILD_STATUS_PARSER_TERMINATED|
			OPC_FRONTEND_CHILD_STATUS_READER_RUNING|
			OPC_FRONTEND_CHILD_STATUS_PARSER_RUNING);
}

int opcFrontendChildIsReady(OPCFrontendChildContext ctx) {
    return opcFrontendChildCheckStatusBitsOn(ctx,OPC_FRONTEND_CHILD_STATUS_READY);
}

void opcFrontendChildTerminateAndWait(OPCFrontendChildContext ctx) {
    assert(ctx);
    opcFrontendChildTerminate(ctx);
    opcFrontendChildWaitTerminate(ctx);
}

void opcFrontendChildZeroContext(OPCFrontendChildContext ctx) {
    assert(ctx);
    ctx->socket=-1;
    ctx->hostname=0;
    ctx->username=0;
    ctx->fctx=NULL;
    ctx->buffpool=NULL;
    ctx->rpctx=NULL;
    ctx->rctx=NULL;
    ctx->lbp=NULL;
    ctx->status =	OPC_FRONTEND_CHILD_STATUS_READER_TERMINATED|
			OPC_FRONTEND_CHILD_STATUS_PARSER_TERMINATED|
			OPC_FRONTEND_CHILD_STATUS_READY;
    ctx->err=0;
}

OPCFrontendChildContext opcFrontendChildCreateContext() {
    OPCFrontendChildContext ctx;

    ctx = (OPCFrontendChildContext)malloc(sizeof(struct OPCFrontendChildContextT));
    opcDebugServer(OPC_DEBUG_NOTICE, "FrontendChild %p: CREATE", ctx);    
    if (ctx) {
	opcFrontendChildZeroContext(ctx);
	pthread_mutex_init(&ctx->status_mutex,NULL);
	pthread_cond_init(&ctx->status_cond,NULL);
	pthread_mutex_init(&ctx->status_cond_mutex,NULL);
	pthread_mutex_lock(&ctx->status_cond_mutex);

	pthread_mutex_init(&ctx->error_mutex,NULL);

	pthread_mutex_init(&ctx->lbp_mutex,NULL);
	pthread_cond_init(&ctx->lbp_cond,NULL);
	pthread_mutex_init(&ctx->lbp_cond_mutex,NULL);
	    // Not locked!
    }
    return ctx;
}

int opcFrontendChildInitContext(OPCFrontendChildContext ctx, OPCFrontendContext fctx, int s) {
    opcDebugServer(OPC_DEBUG_NOTICE, "FrontendChild %p: INIT", ctx);    

    assert(ctx);
    assert(fctx);
    assert(fctx->rpctx);
    assert(fctx->buffpool);
    assert(s);

//    opcFrontendChildZeroContext(ctx);
    
    ctx->socket=s;
    ctx->fctx=fctx;
    ctx->buffpool=fctx->buffpool;
    ctx->rpctx=fctx->rpctx;
    ctx->rctx=opcRequestPoolGetRequest(ctx->rpctx);

    opcFrontendChildResetStatusBits(ctx,OPC_FRONTEND_CHILD_STATUS_READY);

    return 0;
}

void opcFrontendChildReturnBuffers(OPCFrontendChildContext ctx) {
    int err;
    OPCFrontendChildBP clbp;	// Local buffer pool

    assert(ctx);
    
    pthread_mutex_lock(&ctx->lbp_mutex);
    while (ctx->lbp) {
	clbp=ctx->lbp;
	err=BufferPoolReturn(ctx->buffpool,clbp->buffer);assert(err==0);
	ctx->lbp=clbp->next;
	free(clbp);
    }
    pthread_mutex_unlock(&ctx->lbp_mutex);
}

void opcFrontendChildFreeContext(OPCFrontendChildContext ctx) {
    int err;
    int *childerr;
    OPCFrontendChildBP clbp;

    opcDebugServer(OPC_DEBUG_NOTICE, "FrontendChild %p: FREE", ctx);    

    assert(ctx);

    opcFrontendChildTerminateAndWait(ctx);
    opcFrontendChildReturnBuffers(ctx);
    
    if (ctx->status&OPC_FRONTEND_CHILD_STATUS_READER_RUNING) {
	err = pthread_join(ctx->reader_pthread,(void**)&childerr);
	if (err) err = ERR_OPC_FRONTEND_CHILD_THREAD_JOIN;
	else {
	    if (*(void**)childerr==PTHREAD_CANCELED) err=ERR_OPC_FRONTEND_CHILD_THREAD_CANCELED;
	    else err=*childerr;
	    free(childerr);
	}
	opcDebugServer(err?OPC_DEBUG_ERROR:OPC_DEBUG_INFO,"Terminating Frontend Child Reader Thread: %s", opcError(err,NULL));
    }
    
    if (ctx->status&OPC_FRONTEND_CHILD_STATUS_PARSER_RUNING) {
	err = pthread_join(ctx->parser_pthread,(void**)&childerr);
	if (err) err = ERR_OPC_FRONTEND_CHILD_THREAD_JOIN;
	else {
	    if (*(void**)childerr==PTHREAD_CANCELED) err=ERR_OPC_FRONTEND_CHILD_THREAD_CANCELED;
	    else err=*childerr;
	    free(childerr);
	}
	opcDebugServer(err?OPC_DEBUG_ERROR:OPC_DEBUG_INFO,"Terminating Frontend Child Parser Thread: %s", opcError(err,NULL));
    }

    if (ctx->hostname) free(ctx->hostname);
    if (ctx->username) free(ctx->username);
    if (ctx->socket>0) close(ctx->socket);
    if (ctx->rctx) {
	err=opcRequestPoolReturnRequest(ctx->rpctx,ctx->rctx);
	assert(err==0);
    }
    
    opcFrontendChildZeroContext(ctx);
    opcFrontendChildSetStatusBits(ctx,OPC_FRONTEND_CHILD_STATUS_READY);
}

void opcFrontendChildDestroyContext(OPCFrontendChildContext ctx) {

    opcDebugServer(OPC_DEBUG_NOTICE, "FrontendChild %p: DESTROY", ctx);    

    assert(ctx);
    opcFrontendChildFreeContext(ctx);

    pthread_mutex_destroy(&ctx->status_mutex);
    pthread_cond_destroy(&ctx->status_cond);
    pthread_mutex_destroy(&ctx->status_cond_mutex);

    pthread_mutex_destroy(&ctx->error_mutex);

    pthread_mutex_destroy(&ctx->lbp_mutex);
    pthread_cond_destroy(&ctx->lbp_cond);
    pthread_mutex_destroy(&ctx->lbp_cond_mutex);

    free(ctx);
}

int opcFrontendChildSendRespCode(OPCFrontendChildContext ctx, unsigned long code) {
    int err;
    uint32_t rcode;
    unsigned long writed;

    assert(ctx);
    
    rcode = htonl(code);
//    if (send(ctx->socket,&code,sizeof(unsigned long),0)<=0) return ERR_OPC_FE_SENDERR;
    err = opcSignalWrite(ctx->socket,&rcode,sizeof(uint32_t),0,&writed);
    if ((err)||(writed!=sizeof(uint32_t))) {
	if (err==ERR_OPC_SIGNAL_EVENT) {
	    // Try to send out... If failed forget...
	    writed = send(ctx->socket,&rcode,sizeof(uint32_t),0);
	    if (writed==sizeof(uint32_t)) err=0;
	} else err=ERR_OPC_FE_SENDERR;
    }
    
    return err;    
}

int *opcFrontendChildReader(OPCFrontendChildContext ctx) {
    int err = 0;
    unsigned long readed;
    unsigned char hnl,unl;
    unsigned char UseBufferFlag=0;
    char *buffer;
    OPCFrontendChildBP clbp;

    opcDebugServer(OPC_DEBUG_NOTICE, "FrontendChild %p: READER (thread %8.lx)",ctx,pthread_self());

    assert(ctx);

    opcFrontendChildResetStatusBits(ctx,OPC_FRONTEND_CHILD_STATUS_READER_TERMINATED);
    opcFrontendChildSetStatusBits(ctx,OPC_FRONTEND_CHILD_STATUS_READER_RUNING);
    
    while (opcFrontendChildCheckStatusBitsOff(ctx,OPC_FRONTEND_CHILD_STATUS_TERMINATE)) {
	if ((!UseBufferFlag)||(!readed)) {
	    if (!UseBufferFlag) buffer=ctx->buffer;

//	    readed=recv(ctx->socket,ctx->buffer,opc_frontend_child_primary_buffer_size,0);
	    err = opcSignalRead(ctx->socket, ctx->buffer, opc_frontend_child_primary_buffer_size, 0, &readed);
	    if ((err)||(readed<1)) {
		if (err==ERR_OPC_SIGNAL_EVENT) continue;
		opcFrontendChildTerminate(ctx);
		break;
	    }
	} 

	if ((readed<sizeof(unsigned long))||((*(uint32_t*)buffer)!=OPC_FECMD_NEWREQ)) {
	    if (opcFrontendChildSendRespCode(ctx,OPC_FECMD_INVALIDREQ)) {
		opcFrontendChildTerminate(ctx);
		break;
	    }
	    if (UseBufferFlag) readed=0;
	    continue;
	}
	
	if (ctx->hostname) free(ctx->hostname);

	hnl=buffer[sizeof(uint32_t)];
	if (hnl) {
	    if ((hnl+2+sizeof(uint32_t))>readed) { /* checking for unl size also */
		ctx->hostname=NULL;
		if (opcFrontendChildSendRespCode(ctx,OPC_FECMD_INVALIDREQ)) {
		    opcFrontendChildTerminate(ctx);
		    break;
		}
		if (UseBufferFlag) readed=0;
		continue;
	    }
	    ctx->hostname=(char*)malloc(hnl+1);
	    if (ctx->hostname) {
		strncpy(ctx->hostname,buffer+sizeof(uint32_t)+1,hnl);
	        ctx->hostname[hnl]=0;
	    } else {
		opcFrontendChildTerminate(ctx);
		break;	    
	    }
	} else ctx->hostname=NULL;
	
	if (ctx->username) free(ctx->username);
	
	unl=buffer[sizeof(uint32_t)+hnl+1];
	if (unl) {
	    if ((unl+hnl+2+sizeof(uint32_t))>readed) {
		ctx->username=NULL;
		if (opcFrontendChildSendRespCode(ctx,OPC_FECMD_INVALIDREQ)) {
		    opcFrontendChildTerminate(ctx);
		    break;
		}
		if (UseBufferFlag) readed=0;
		continue;
	    }
	    ctx->username=(char*)malloc(unl+1);
	    if (ctx->username) {
		strncpy(ctx->username,buffer+hnl+sizeof(uint32_t)+2,unl);
	        ctx->hostname[unl]=0;
	    } else {
		opcFrontendChildTerminate(ctx);
		break;	    
	    }
	} else ctx->username=NULL;

	opcDebugRequest(OPC_DEBUG_NOTICE,"Frontend event (socket %u)",ctx->socket);
	if (hnl) opcDebugRequest(OPC_DEBUG_INFO," client host: %s",ctx->hostname);
	if (unl) opcDebugRequest(OPC_DEBUG_INFO," client user: %s",ctx->username);
	
	    // If RequestContext still uninitialized trying to initialize it
	if (!ctx->rctx) {
	    ctx->rctx=opcRequestPoolGetRequest(ctx->rpctx);
	    if (!ctx->rctx) {
		opcDebugRequest(OPC_DEBUG_CRITICAL, "Internal Error. Request Aborted.");
		opcFrontendChildSendRespCode(ctx,OPC_FECMD_RCERROR);
		opcFrontendChildTerminate(ctx);
		err = ERR_OPC_REQUEST_INIT;
		break;
	    }
	}

	    // Sending ACK
	if (opcFrontendChildSendRespCode(ctx,OPC_FECMD_NEWREQ)) {
	    opcDebugRequest(OPC_DEBUG_CRITICAL, " Error Sending ACK response to client");
	    opcFrontendChildTerminate(ctx);
	    break;
	}

	opcFrontendChildResetStatusBits(ctx,OPC_FRONTEND_CHILD_STATUS_ENDEVENT);
	while (opcFrontendChildCheckStatusBitsOff(ctx,OPC_FRONTEND_CHILD_STATUS_TERMINATE|OPC_FRONTEND_CHILD_STATUS_ENDEVENT)) {
	    if (UseBufferFlag) {
		UseBufferFlag=0;
	    } else {
		buffer=BufferPoolGet(ctx->buffpool);
		if (!buffer) {
		    opcDebugRequest(OPC_DEBUG_CRITICAL, "Internal Error. Request Aborted.");
		    opcFrontendChildTerminate(ctx);
		    err=ERR_OPC_BUFFPOOL_GET;
		    break;
		}
	    }
	    
// 	    readed=recv(ctx->socket,buffer,opc_buffer_size,0);
	    err = opcSignalRead(ctx->socket, buffer, opc_buffer_size, 0, &readed);
	    if ((err)||(readed<1)) {
		if (err==ERR_OPC_SIGNAL_EVENT) {
		    readed=0;
		    UseBufferFlag=1;
		    continue;
		}
		opcFrontendChildTerminate(ctx);
		opcDebugRequest(OPC_DEBUG_CRITICAL, " Error reading xml request from apache module (Errno: %lu)\n",errno);
		break;
	    }

	    pthread_mutex_lock(&ctx->error_mutex);
	    if (opcFrontendChildCheckStatusBitsOn(ctx,OPC_FRONTEND_CHILD_STATUS_ENDEVENT)) {
		    UseBufferFlag=1;
		    pthread_mutex_unlock(&ctx->error_mutex);
		    break;
	    }

	    pthread_mutex_lock(&ctx->lbp_mutex);
	    if (ctx->lbp) {
		clbp=ctx->lbp;
		while(clbp->next) clbp=clbp->next;
		clbp->next=(OPCFrontendChildBP)malloc(sizeof(struct OPCFrontendChildBPT));
		clbp=clbp->next;
	    } else clbp=(OPCFrontendChildBP)malloc(sizeof(struct OPCFrontendChildBPT));

	    if (!clbp) {
		opcDebugRequest(OPC_DEBUG_CRITICAL, "Internal Error. Request Aborted.");
		opcFrontendChildTerminate(ctx);
		pthread_mutex_unlock(&ctx->lbp_mutex);
		pthread_mutex_unlock(&ctx->error_mutex);
		err = ERR_OPC_MEM;
		break;
	    }

	    clbp->buffer=buffer;
	    clbp->size=readed;
	    clbp->next=NULL;
	    if (!ctx->lbp) ctx->lbp=clbp;
	    pthread_mutex_unlock(&ctx->lbp_mutex);
	    pthread_cond_broadcast(&ctx->lbp_cond);
	    pthread_mutex_unlock(&ctx->error_mutex);
	}
    }

    opcFrontendChildSetStatusBits(ctx,OPC_FRONTEND_CHILD_STATUS_READER_TERMINATED);

    opcDebugServer(OPC_DEBUG_INFO, "FrontendChild %p: READER Finished",ctx);

    if (err!=ERR_OPC_SIGNAL_EVENT) return ThreadReturnInt(NULL,err);
    return ThreadReturnInt(NULL,0);
}

void opcFrontendChildReturnBuffer(OPCFrontendChildContext ctx) {
    int err;
    OPCFrontendChildBP clbp;	// Local buffer pool

    pthread_mutex_lock(&ctx->lbp_mutex);
    clbp=ctx->lbp;
    err=BufferPoolReturn(ctx->buffpool,clbp->buffer);assert(err==0);
    ctx->lbp=clbp->next;
    free(clbp);
    pthread_mutex_unlock(&ctx->lbp_mutex);
}

int opcFrontendChildParserError(OPCFrontendChildContext ctx, int error, int code) {
    int err;
    
    assert(ctx);
    
    pthread_mutex_lock(&ctx->error_mutex);
    
	// Setting EndEvent Flag
    opcFrontendChildSetStatusBits(ctx,OPC_FRONTEND_CHILD_STATUS_ENDEVENT);

	// Free Buffers
    opcFrontendChildReturnBuffers(ctx);
    
	// ReInitParser
    err=opcRequestPoolReturnRequest(ctx->rpctx,ctx->rctx); assert(err==0);
    ctx->rctx=opcRequestPoolGetRequest(ctx->rpctx);
    
    pthread_mutex_unlock(&ctx->error_mutex);
    
	// Send message back
    if (opcFrontendChildSendRespCode(ctx,code)) {
	opcFrontendChildSetStatusBits(ctx,OPC_FRONTEND_CHILD_STATUS_PARSER_TERMINATED);
	opcDebugRequest(OPC_DEBUG_CRITICAL, "Request failed, status=Connection Closed, reason=Response Failed");
	return 1;
    }

    opcDebugRequest(OPC_DEBUG_ERROR,"Request served, status=FAILED, reason = %s", opcError(error,NULL));

    return 0;
}

static int opcFrontendSendAttachmentCallback(OPCFrontendChildContext ctx, const char *buffer, unsigned long len) {
    int err;
    unsigned long writed;
    
    err = opcSignalWrite(ctx->socket,buffer,len,0,&writed);
    if ((err)||(writed!=len)) {
	ctx->err = ERR_OPC_FE_SEND;
	if (err==ERR_OPC_SIGNAL_EVENT) {
	    opcDebugRequest(OPC_DEBUG_WARNING, "Request failed, status=Connection Closed, reason=Server Shutdown");
	} else {
	    opcDebugRequest(OPC_DEBUG_CRITICAL, "Request failed, status=Connection Closed, reason=Data Send Failed");
	}
	return -1;
    }

    return 0;
}

static int opcFrontendSendCallback(OPCFrontendChildContext ctx, const char *buffer, int len) {
    int err;
    int size;
    unsigned long writed;
    
    if ((!ctx)||(!buffer)||(len<0)) return -1;
    if (!len) return 0;

    if (ctx->format) {
	if (ctx->format==1) {
	    err = dimeSendHeader(opc_soap_message_id, len, DIME_TYPE_XML, NULL, 0, 0, 1, (OPCSendFunction)opcFrontendSendAttachmentCallback, ctx);
	    ctx->format=2;	    
	} else {
	    err = dimeSendChunk(len, 0, (OPCSendFunction)opcFrontendSendAttachmentCallback, ctx);
	}
	if (err) {
	    ctx->err = ERR_OPC_FE_SEND;
	    opcDebugRequest(OPC_DEBUG_CRITICAL, "Request failed, status=Connection Closed, reason = Response Failed");
	    return -1;
	}

	if (len&3) size = (len&~3)+4; /* Buffer size should be multiple of 4 */
	else size = len;
    } else {
	if (opcFrontendChildSendRespCode(ctx,len)) {
	    ctx->err = ERR_OPC_FE_SEND;
	    opcDebugRequest(OPC_DEBUG_CRITICAL, "Request failed, status=Connection Closed, reason = Response Failed");
	    return -1;
	}
	size = len;
    }

	err = opcSignalWrite(ctx->socket,buffer,size,0,&writed);
    if ((err)||(writed!=size)) {
	ctx->err = ERR_OPC_FE_SEND;
	if (err==ERR_OPC_SIGNAL_EVENT) {
	    opcDebugRequest(OPC_DEBUG_WARNING, "Request failed, status=Connection Closed, reason=Server Shutdown");
	} else {
	    opcDebugRequest(OPC_DEBUG_CRITICAL, "Request failed, status=Connection Closed, reason=Data Send Failed");
	}
	return -1;
    }
    
    return len;
}

static void *opcFrontendOpenMIMEAttachmentCallback(OPCFrontendChildContext ctx, const char *path, const char *name, OPCDataInfo info, OPCUserData data) {
    int err;
    char *buffer;
    unsigned long pos;
    unsigned short len, pathlen, namelen;
    unsigned char c;

/*    if ((data->type!=OPC_DATA_TYPE_DATABUFFER)&&(data->type!=OPC_DATA_TYPE_ALLOCATION)) return NULL; */
    if (!data->size) return NULL;
    
    
    if (path) pathlen = strlen(path);
    if (name) namelen = strlen(name);
    
    if (pathlen) {
	if ((namelen)&&(*name!='/')) c=1;
	else c=0;
    } else {
	if ((namelen)&&(*name=='/')) c=0;
	else c=1;
    }
    len = c + namelen + pathlen;
    
    buffer = malloca(len+2*sizeof(uint32_t)+sizeof(uint16_t)+1);
    if (!buffer) {
	ctx->err = ERR_OPC_MEM;
	return NULL;
    }
    
    *(uint32_t*)buffer = htonl(OPC_FECMD_ATTACHMENT);
    *(uint32_t*)(buffer+sizeof(uint32_t)) = htonl(data->size);
    *(uint16_t*)(buffer+2*sizeof(uint32_t)) = htons(len);
    pos = 2*sizeof(uint32_t)+sizeof(uint16_t);
    
    if (pathlen) {
	strcpy(buffer+pos, path);
	pos+=pathlen;
    }
    if (c) buffer[pos++]='/';
    if (namelen) strcpy(buffer+pos, name);

    len += 2*sizeof(uint32_t)+sizeof(uint16_t);

    err = opcSignalWrite(ctx->socket,buffer,len,0,&pos);
    if ((err)||(pos!=len)) {
	ctx->err = ERR_OPC_FE_SEND;
	if (err==ERR_OPC_SIGNAL_EVENT) {
	    opcDebugRequest(OPC_DEBUG_WARNING, "Request failed, status=Connection Closed, reason=Server Shutdown");
	} else {
	    opcDebugRequest(OPC_DEBUG_CRITICAL, "Request failed, status=Connection Closed, reason=Data Send Failed");
	}
	freea(buffer);
	return NULL;
    }

    freea(buffer);
    return (void*)ctx;
}

static void *opcFrontendOpenDIMEAttachmentCallback(OPCFrontendChildContext ctx, const char *item, OPCDataInfo info, OPCUserData data) {
/*    if ((data->type!=OPC_DATA_TYPE_DATABUFFER)&&(data->type!=OPC_DATA_TYPE_ALLOCATION)) return NULL; */
    if (!data->size) return NULL;
    
    if (ctx->format) {
	if (dimeSendHeader(item,data->size,DIME_TYPE_BINARY,NULL,0,0,0,(OPCSendFunction)opcFrontendSendAttachmentCallback, ctx)) {
	    opcDebugRequest(OPC_DEBUG_CRITICAL, "Request failed, status=Connection Closed, reason = DIME send failed");
	    ctx->err = ERR_OPC_FE_SEND;
	    return NULL;
	}
    } else {
	if (opcFrontendChildSendRespCode(ctx,OPC_FECMD_DIME_RESPONSE)) {
	    opcDebugRequest(OPC_DEBUG_CRITICAL, "Request failed, status=Connection Closed, reason = Start DIME transfer failed");
	    ctx->err = ERR_OPC_FE_SEND;
	    return NULL;
	}
	if (dimeSendHeader(item,data->size,DIME_TYPE_BINARY,NULL,1,0,0,(OPCSendFunction)opcFrontendSendAttachmentCallback, ctx)) {
	    opcDebugRequest(OPC_DEBUG_CRITICAL, "Request failed, status=Connection Closed, reason = DIME send failed");
	    ctx->err = ERR_OPC_FE_SEND;
	    return NULL;
	}
	ctx->format = 1;
    }
	
    return (void*)ctx;
}


int *opcFrontendChildParser(OPCFrontendChildContext ctx) {
    int err;
    int writed;
    unsigned long len=0;	// Length of following message block
    unsigned long size;		// Byte length of buffer
    char *buffer;		// Position in buffer
    unsigned char ExitFlag;
//    struct timespec abstime;
    
    opcDebugServer(OPC_DEBUG_NOTICE, "FrontendChild %p: PARSER (thread %8.lx)",ctx,pthread_self());

    assert(ctx);

    opcFrontendChildResetStatusBits(ctx,OPC_FRONTEND_CHILD_STATUS_PARSER_TERMINATED);
    opcFrontendChildSetStatusBits(ctx,OPC_FRONTEND_CHILD_STATUS_PARSER_RUNING);
    
    while (opcFrontendChildCheckStatusBitsOff(ctx,OPC_FRONTEND_CHILD_STATUS_TERMINATE)) {
	pthread_mutex_lock(&ctx->lbp_mutex);
	if (ctx->lbp) pthread_mutex_unlock(&ctx->lbp_mutex);
	else {
//	    abstime.tv_sec=time(NULL)+opc_frontend_child_waittime;
	    pthread_mutex_lock(&ctx->lbp_cond_mutex);
	    pthread_mutex_unlock(&ctx->lbp_mutex);
	    err=pthread_cond_wait(&ctx->lbp_cond,&ctx->lbp_cond_mutex);
//	    err=pthread_cond_timedwait(&ctx->lbp_cond,&ctx->lbp_cond_mutex,&abstime);
	    pthread_mutex_unlock(&ctx->lbp_cond_mutex);
//	    if (err==ETIMEDOUT) continue;
	    if (opcFrontendChildCheckStatusBitsOn(ctx,OPC_FRONTEND_CHILD_STATUS_TERMINATE)) break;
	}
	opcDebugRequest(OPC_DEBUG_INFO," Got a buffer, length=%lu",ctx->lbp->size);

	buffer=ctx->lbp->buffer;
	size=ctx->lbp->size;//-sizeof(unsigned long);
	
	for (ExitFlag=0;!ExitFlag;) {
	    opcDebugRequest(OPC_DEBUG_DEBUG," %i %u",size,len);
	    if (!len) {
		if ((signed long)(size-len)<0) {
		    opcDebugRequest(OPC_DEBUG_DEBUG,"%i %u...\n",size,len);
		    opcFrontendChildParserError(ctx,ERR_OPC_FE_CLIREQUEST,OPC_FECMD_INVALIDREQ);
		    len=0;
		    ExitFlag=1;
		    break;
		}
		len=ntohl(*(unsigned long*)buffer);
		buffer+=sizeof(unsigned long);
		size-=sizeof(unsigned long); // Error
	    } 
	    opcDebugRequest(OPC_DEBUG_DEBUG,"r: %i %i %lx\n",size,len,len);

	    switch(len) {
		case OPC_FECMD_NEWREQ:
			// Sending Retransmit Response to Server
		    if (opcFrontendChildSendRespCode(ctx,OPC_FECMD_REPEAT)) {
			opcFrontendChildTerminate(ctx);
			opcDebugRequest(OPC_DEBUG_ERROR, " ERROR Sending REPEAT response to client");
		    }
		case OPC_FECMD_CANCEL:
		    opcDebugRequest(OPC_DEBUG_NOTICE," Request canceled");

		    // Interrupt old request, start new one
		    opcFrontendChildReturnBuffer(ctx);
		    opcFrontendChildSetStatusBits(ctx,OPC_FRONTEND_CHILD_STATUS_ENDEVENT);

			// Free Buffers (all unused buffers)
		    pthread_mutex_lock(&ctx->error_mutex);
		    opcFrontendChildReturnBuffers(ctx);
		    pthread_mutex_unlock(&ctx->error_mutex);
		    
			// Reinitialising Request Context
		    opcRequestPoolReturnRequest(ctx->rpctx,ctx->rctx);
		    ctx->rctx=opcRequestPoolGetRequest(ctx->rpctx);
		    ExitFlag=1;
		    len=0;
		break;
		case OPC_FECMD_XMLEND: 
		    // Check if no parts left (Multipart MIME)
		    opcDebugRequest(OPC_DEBUG_NOTICE," END of XML Request");
		    opcFrontendChildReturnBuffer(ctx);

		    ctx->format = 0;
//		    err=opcRequestProcess(ctx->rctx, NULL, NULL, NULL, NULL);
		    err=opcRequestProcess(ctx->rctx, (OPCRequestOpenAttachmentFunction)&opcFrontendOpenDIMEAttachmentCallback, (OPCSendFunction)&opcFrontendSendAttachmentCallback, NULL, ctx);
		    if (err) {
		    	opcFrontendChildParserError(ctx,err,OPC_FECMD_OPC_INVALIDREQ);
			len=0;
			ExitFlag=1;
			break;
		    }

		    opcFrontendChildSetStatusBits(ctx,OPC_FRONTEND_CHILD_STATUS_ENDEVENT);

		    err=opcRequestSend(ctx->rctx, (xmlOutputWriteCallback)opcFrontendSendCallback, NULL, ctx);
		    if (err) {
			if ((ctx->err)||(ctx->format)) {
			    opcDebugRequest(OPC_DEBUG_CRITICAL, "Request failed, status=Connection Closed, reason = Response Serialization Failed");
			    opcFrontendChildTerminate(ctx);
			} else {
		    	    opcFrontendChildParserError(ctx,err,OPC_FECMD_INTERNAL_ERROR);
			    len=0;
			}
			ExitFlag=1;
			break;
		    }
		    
		    if (ctx->format) {
			if (dimeSendCloseRecord((OPCSendFunction)opcFrontendSendAttachmentCallback, ctx)) {
			    opcDebugRequest(OPC_DEBUG_CRITICAL, "Request failed, status=Connection Closed, reason = Final DIME Record Send Failed");
			    opcFrontendChildTerminate(ctx);
			    ExitFlag = 1;
			}
		    } else {
			if (opcFrontendChildSendRespCode(ctx,OPC_FECMD_XMLEND)) {
			    opcDebugRequest(OPC_DEBUG_CRITICAL, "Request failed, status=Connection Closed, reason = EOF send failed");
			    opcFrontendChildTerminate(ctx);
			    ExitFlag=1;
			    break;
			}			    
		    }

/*		    

		    if (opcFrontendChildSendRespCode(ctx,ctx->rctx->answer_size)) {
			opcFrontendChildTerminate(ctx);
			opcDebugRequest(OPC_DEBUG_CRITICAL, "Request failed, status=Connection Closed, reason = Response Failed");
			ExitFlag=1;
			break;
		    }
		    
		    err = opcSignalWrite(ctx->socket,ctx->rctx->answer,ctx->rctx->answer_size,0,&writed);
		    if ((err)||(writed!=ctx->rctx->answer_size)) {
			if (err==ERR_OPC_SIGNAL_EVENT) {
			    // Trying send it fast ones again (ToDo: check if something alredy sent)
			    writed += send(ctx->socket,ctx->rctx->answer+writed,ctx->rctx->answer_size-writed,0);
			    if (writed!=ctx->rctx->answer_size) {
				opcDebugRequest(OPC_DEBUG_WARNING, "Request failed, status=Connection Closed, reason=Server Shutdown");
				ExitFlag=1;
				break;
			    }
			} else {
			    printf("Errno: %lx %lx (%lu %lu)\n",err, errno, ctx->rctx->answer_size, writed);
			    opcFrontendChildTerminate(ctx);
			    opcDebugRequest(OPC_DEBUG_CRITICAL, "Request failed, status=Connection Closed, reason=Data Send Failed");
			    ExitFlag=1;
			    break;
			}
			
		    }
*/		    
		    opcDebugRequest(OPC_DEBUG_NOTICE,"Request served, status=OK");

			// Free Buffers (all unused buffers)
/* DS: Error?
		    pthread_mutex_lock(&ctx->error_mutex);
		    opcFrontendChildReturnBuffers(ctx);
		    pthread_mutex_unlock(&ctx->error_mutex);
*/		    
			// Reinitialising Request Context
		    opcRequestPoolReturnRequest(ctx->rpctx,ctx->rctx);
		    ctx->rctx=opcRequestPoolGetRequest(ctx->rpctx);
		    ExitFlag=1;
		    len=0;
		break;
		default:
		    if (size<=len) { /* We have one or even part of message in buffer */
			if (!size) {
			    ExitFlag=1;
			    break;
			}
			opcDebugRequest(OPC_DEBUG_INFO," Message, length=%lu",len);

			err=opcRequestXMLData(ctx->rctx,buffer,size);
			if (err) {
		    	    opcFrontendChildParserError(ctx,err,OPC_FECMD_OPC_INVALIDREQ);
			    len=0;
			    ExitFlag=1;
			    break;
			}
			opcFrontendChildReturnBuffer(ctx);
			ExitFlag=1;
			len-=size;
		    } else { /* We have several messages in buffer */
			opcDebugRequest(OPC_DEBUG_INFO," Multipart Message, length=%lu",len);
			err=opcRequestXMLData(ctx->rctx,buffer,len);
			if (err) {
			    opcFrontendChildParserError(ctx,err,OPC_FECMD_OPC_INVALIDREQ);
			    len=0;
			    ExitFlag=1;
			    break;
			}	
			opcDebugRequest(OPC_DEBUG_DEBUG, "%i %u!",size,len);		    
			buffer+=len;
			size-=len;
			len=0;
		    }
	    }
	    opcDebugRequest(OPC_DEBUG_DEBUG, "EF: %i",ExitFlag);
	}
    }
    
    opcFrontendChildSetStatusBits(ctx,OPC_FRONTEND_CHILD_STATUS_PARSER_TERMINATED);

    opcDebugServer(OPC_DEBUG_INFO, "FrontendChild %p: PARSER Finished",ctx);
    return ThreadReturnInt(NULL,0);
}
