#include <stdio.h>
#include <pthread.h>
#include <signal.h>
#include <assert.h>

#include <opcerr.h>
#include <opcconfig.h>
#include <opcdebug.h>

#include <makepath.h>

#include <opc.h>
#include <opcsignal.h>
#include <opcrequestpool.h>

#include <bufferpool.h>
#include <frontend.h>
#include <backend.h>
#include <opcserver.h>

static void usr1handler(int signo) {}

int opcServerInit() {
    return opcInit();
}

void opcServerDestroy() {
    opcDestroy();
}

void opcServerZeroContext(OPCServerContext ctx) {
    assert(ctx);

    ctx->malloc=NULL;    
    ctx->ctx=NULL;
    ctx->sigctx=NULL;
    ctx->rpctx=NULL;
    ctx->fctx=NULL;
    ctx->bctx=NULL;
    ctx->buffpool=NULL;
    ctx->status=0;
}


OPCServerContext opcServerCreateContext() {
    OPCServerContext ctx;
    
    opcDebugServer(OPC_DEBUG_NOTICE,"opcServerCreateContext");
    
    ctx = (OPCServerContext)malloc(sizeof(struct OPCServerContextT));
    if (ctx) opcServerZeroContext(ctx);
    return ctx;
}

int opcServerInitContext(OPCServerContext ctx) {
    int err;
    
    opcDebugServer(OPC_DEBUG_NOTICE,"opcServerInitContext");
    
    assert(ctx);
    
    ctx->malloc=opcMallocCreateContext(NULL);
    if (!ctx->malloc) return ERR_OPC_MALLOC_INIT;

    ctx->sigctx=opcSignalCreateContext(ctx->malloc);
    if (!ctx->sigctx) {
	opcServerFreeContext(ctx);
	return ERR_OPC_SIGNAL_INIT;
    }

    err=opcSignalInitContext(ctx->sigctx);
    if (err) {
	opcServerFreeContext(ctx);
	return err;
    }

    ctx->buffpool=BufferPoolCreate(opc_buffer_size);
    if (!ctx->buffpool) {
	opcServerFreeContext(ctx);
	return ERR_OPC_BUFFPOOL_INIT;
    }

    ctx->ctx=opcCreateContext(ctx->malloc);
    if (!ctx->ctx) return ERR_OPC_LIBRARY_INIT;

    err=opcInitContext(ctx->ctx, NULL, OPC_OPTION_SCALE_FACTOR);
    if (err) {
	opcServerFreeContext(ctx);
	return err;
    }
    
    ctx->status|=OPC_SERVER_CONTEXT_INITIALIZED;

    ctx->rpctx=opcRequestPoolCreateContext(ctx->malloc);
    if (!ctx->rpctx) {
	opcServerFreeContext(ctx);
	return ERR_OPC_REQUESTPOOL_INIT;
    }

    err=opcRequestPoolInitContext(ctx->rpctx,ctx->ctx);
    if (err) {
	opcServerFreeContext(ctx);
	return err;
    }
    
    
    err = make_path (opc_sockets_dir, -1, -1, -1, -1, 1);
    if (err) {
	opcServerFreeContext(ctx);
	return ERR_OPC_PATH;
    }
    
    ctx->bctx=opcBackendCreateContext(ctx->malloc);
    if (!ctx->bctx) {
	opcServerFreeContext(ctx);
	return ERR_OPC_BACKEND_INIT;
    }
    
    err = opcBackendInitContext(ctx->bctx, ctx->ctx, ctx->sigctx);
    if (err) {
	opcServerFreeContext(ctx);
	return err;
    }
    
    err = opcBackendInitSocket(ctx->bctx, backend_socket_name);
    if (err) {
	opcServerFreeContext(ctx);
	return err;
    }
    
    ctx->fctx=opcFrontendCreateContext();
    if (!ctx->fctx) {
	opcServerFreeContext(ctx);
	return ERR_OPC_FRONTEND_INIT;
    }
    
    err=opcFrontendInitContext(ctx->fctx,ctx->rpctx,ctx->buffpool,ctx->sigctx);
    if (err) {
	opcServerFreeContext(ctx);
	return err;
    }

    return 0;
}

    // All spawned childs must been destroyed prior to this call
void opcServerFreeContext(OPCServerContext ctx) {
    int err=0;
    int *childerr;

    assert(ctx);

    opcDebugServer(OPC_DEBUG_NOTICE, "opcServerFreeContext");

    if (ctx->status&OPC_SERVER_CONTEXT_INITIALIZED) opcServerTerminate(ctx);

    if ((ctx->status&OPC_SERVER_FRONTEND_INITIALIZED)) {
	err = pthread_join(ctx->frontend, (void**)&childerr);
	if (err) err = ERR_OPC_FRONTEND_THREAD_JOIN;
	else {
	    if (childerr) {
		if (*(void**)childerr==PTHREAD_CANCELED) err=ERR_OPC_FRONTEND_THREAD_CANCELED;
		else err=*childerr;
		free(childerr);
	    } else err=ERR_OPC_FRONTEND_THREAD_RETURNCODE;
	}
	opcDebugServer(err?OPC_DEBUG_ERROR:OPC_DEBUG_INFO,"Terminating Frontend Thread: %s", opcError(err,NULL));
    }
    
    if ((ctx->status&OPC_SERVER_BACKEND_INITIALIZED)) {
	err = pthread_join(ctx->backend, (void**)&childerr);
	if (err) err = ERR_OPC_BACKEND_THREAD_JOIN;
	else {
	    if (childerr) {
		if (*(void**)childerr==PTHREAD_CANCELED) err=ERR_OPC_BACKEND_THREAD_CANCELED;
		else err=*childerr;
		free(childerr);
	    } else err=ERR_OPC_BACKEND_THREAD_RETURNCODE;
	}
	opcDebugServer(err?OPC_DEBUG_ERROR:OPC_DEBUG_INFO,"Terminating Backend Thread: %s", opcError(err,NULL));
    }
    
    if (ctx->fctx) opcFrontendDestroyContext(ctx->fctx);
    if (ctx->bctx) opcBackendDestroyContext(ctx->bctx);
    if (ctx->rpctx) opcRequestPoolDestroyContext(ctx->rpctx);
    if (ctx->ctx) opcDestroyContext(ctx->ctx);
    if (ctx->buffpool) BufferPoolDestroyForce(ctx->buffpool);
    if (ctx->sigctx) opcSignalDestroyContext(ctx->sigctx);
    if (ctx->malloc) opcMallocDestroyContext(ctx->malloc);
    opcServerZeroContext(ctx);
}

void opcServerDestroyContext(OPCServerContext ctx) {
    opcDebugServer(OPC_DEBUG_NOTICE, "opcServerDestroyContext");
    
    assert(ctx);
    opcServerFreeContext(ctx);
    free(ctx);
}

int opcServerRun(OPCServerContext ctx) {
    int err;

    opcDebugServer(OPC_DEBUG_NOTICE, "opcServerRun");
    
    assert(ctx);

    err=pthread_create(&ctx->backend, NULL, (void*(*)(void*))opcBackendListen, (void*)ctx->bctx);
    if (err) return ERR_OPC_BACKEND_THREAD_RUN;
    else ctx->status|=OPC_SERVER_BACKEND_INITIALIZED;
    
    err=pthread_create(&ctx->frontend, NULL, (void*(*)(void*))opcFrontendListen, (void*)ctx->fctx);
    if (err) {
	opcServerTerminate(ctx);
	return ERR_OPC_FRONTEND_THREAD_RUN;
    } else ctx->status|=OPC_SERVER_FRONTEND_INITIALIZED;
    
    return 0;
}

int opcServerTerminate(OPCServerContext ctx) {
    assert(ctx);
    
    ctx->status|=OPC_SERVER_TERMINATE_FLAG;
    
    if (ctx->sigctx) opcSignalTerminate(ctx->sigctx);
}

void opcServerMantain(OPCServerContext ctx) {
    opcDebugServer(OPC_DEBUG_INFO, "opcServerMantain");
    assert(ctx);

    opcFrontendCleanupChilds(ctx->fctx);
    opcMantain(ctx->ctx);
}
